Package: ZygosityPredictor
Type: Package
Title: Package for prediction of zygosity for variants/genes in NGS
        data
Version: 1.13.0
Date: 2025-07-28
Authors@R: c(
  person("Marco", "Rheinnecker", 
    email="marco.rheinnecker@dkfz-heidelberg.de", 
    role = c("aut", "cre"),
    comment=c(ORCID="0009-0009-7181-3977")), 
  person("Marc", "Ruebsam", 
    role = c("aut")),
  person("Daniel", "Huebschmann", 
    role = c("aut")),  
  person("Martina", "Froehlich", 
    role = c("aut")),
  person("Barbara", "Hutter", 
    role = c("aut"))
    )
Imports: GenomicAlignments, GenomicRanges, Rsamtools, IRanges,
        VariantAnnotation, DelayedArray, dplyr, stringr, purrr, tibble,
        methods, knitr, igraph, readr, stats, magrittr, rlang
License: GPL-2
Description: The ZygosityPredictor allows to predict how many copies of a gene are affected by small variants. In addition to the basic calculations of the affected copy number of a variant, the Zygosity-Predictor can integrate the influence of several variants on a gene and ultimately make a statement if and how many wild-type copies of the gene are left. This information proves to be of particular use in the context of translational medicine. For example, in cancer genomes, the Zygosity-Predictor can address whether unmutated copies of tumor-suppressor genes are present. Beyond this, it is possible to make this statement for all genes of an organism. The Zygosity-Predictor was primarily developed to handle SNVs and INDELs (later addressed as small-variants) of somatic and germline origin. In order not to overlook severe effects outside of the small-variant context, it has been extended with the assessment of large scale deletions, which cause losses of whole genes or parts of them.
RoxygenNote: 7.2.3
Encoding: UTF-8
biocViews: BiomedicalInformatics, FunctionalPrediction,
        SomaticMutation, GenePrediction
Depends: R (>= 4.3.0)
LazyData: false
Suggests: rmarkdown, testthat, BiocStyle
VignetteBuilder: knitr
git_url: https://git.bioconductor.org/packages/ZygosityPredictor
git_branch: devel
git_last_commit: 207b374
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-29 01:53:23 UTC; biocbuild
Author: Marco Rheinnecker [aut, cre] (ORCID:
    <https://orcid.org/0009-0009-7181-3977>),
  Marc Ruebsam [aut],
  Daniel Huebschmann [aut],
  Martina Froehlich [aut],
  Barbara Hutter [aut]
Maintainer: Marco Rheinnecker <marco.rheinnecker@dkfz-heidelberg.de>

# Generated by roxygen2: do not edit by hand

export(ZP_ov)
export(aff_germ_copies)
export(aff_som_copies)
export(gene_ov)
export(predict_per_variant)
export(predict_zygosity)
importFrom(DelayedArray,rowRanges)
importFrom(GenomicAlignments,cigarRangesAlongQuerySpace)
importFrom(GenomicAlignments,cigarRangesAlongReferenceSpace)
importFrom(GenomicAlignments,end)
importFrom(GenomicAlignments,first)
importFrom(GenomicAlignments,granges)
importFrom(GenomicAlignments,last)
importFrom(GenomicAlignments,readGAlignmentPairs)
importFrom(GenomicAlignments,readGAlignments)
importFrom(GenomicAlignments,start)
importFrom(GenomicAlignments,width)
importFrom(GenomicRanges,"elementMetadata<-")
importFrom(GenomicRanges,GRanges)
importFrom(GenomicRanges,elementMetadata)
importFrom(GenomicRanges,end)
importFrom(GenomicRanges,seqnames)
importFrom(GenomicRanges,start)
importFrom(IRanges,IRanges)
importFrom(IRanges,mergeByOverlaps)
importFrom(IRanges,subsetByOverlaps)
importFrom(Rsamtools,ScanBamParam)
importFrom(Rsamtools,TabixFile)
importFrom(Rsamtools,scanBamFlag)
importFrom(Rsamtools,seqnamesTabix)
importFrom(VariantAnnotation,geno)
importFrom(VariantAnnotation,info)
importFrom(VariantAnnotation,readVcf)
importFrom(dplyr,all_of)
importFrom(dplyr,arrange)
importFrom(dplyr,as_tibble)
importFrom(dplyr,between)
importFrom(dplyr,bind_rows)
importFrom(dplyr,case_when)
importFrom(dplyr,desc)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,mutate_all)
importFrom(dplyr,mutate_at)
importFrom(dplyr,na_if)
importFrom(dplyr,nth)
importFrom(dplyr,pull)
importFrom(dplyr,relocate)
importFrom(dplyr,rowwise)
importFrom(dplyr,select)
importFrom(dplyr,summarize)
importFrom(dplyr,tally)
importFrom(dplyr,tibble)
importFrom(dplyr,ungroup)
importFrom(igraph,all_shortest_paths)
importFrom(igraph,graph_from_data_frame)
importFrom(knitr,kable)
importFrom(magrittr,"%>%")
importFrom(methods,is)
importFrom(purrr,compact)
importFrom(purrr,set_names)
importFrom(readr,write_tsv)
importFrom(rlang,ensym)
importFrom(stats,chisq.test)
importFrom(stats,na.omit)
importFrom(stringr,"%>%")
importFrom(stringr,str_detect)
importFrom(stringr,str_match)
importFrom(stringr,str_replace)
importFrom(stringr,str_replace_all)
importFrom(stringr,str_split)
importFrom(stringr,str_sub)
importFrom(tibble,as_tibble)
importFrom(tibble,column_to_rownames)
importFrom(tibble,tibble)

## ----eval=FALSE---------------------------------------------------------------
# if (!require("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# 
# BiocManager::install("ZygosityPredictor")

## -----------------------------------------------------------------------------
library(ZygosityPredictor)
library(dplyr)
library(stringr)
library(GenomicRanges)

# file to sequence alignment 
FILE_BAM <- system.file("extdata", "ZP_example.bam", 
                        package = "ZygosityPredictor")
VCF <- system.file("extdata", "ZP_example_chr7.vcf.gz", 
                        package = "ZygosityPredictor")
# meta information of the sample
PURITY <- 0.98
PLOIDY <- 1.57
SEX <- "female"
# variants
data("GR_SCNA")
data("GR_GERM_SMALL_VARS")
data("GR_SOM_SMALL_VARS")
# used gene model
data("GR_GENE_MODEL")
data("GR_HAPLOBLOCKS")


## -----------------------------------------------------------------------------
## as an example we take the first variant of our prepared input data and 
## extract the required information from different input data layer
## the allele frequency and the chromosome can be taken from the GRanges object

AF = elementMetadata(GR_GERM_SMALL_VARS[1])[["af"]]
CHR = seqnames(GR_GERM_SMALL_VARS[1]) %>%
  as.character()

## the total copy number (tcn) can be extracted from the CNV object by selecting
## the CNV from the position of the variant

TCN = elementMetadata(
  subsetByOverlaps(GR_SCNA, GR_GERM_SMALL_VARS[1])
  )[["tcn"]]

## purity and sex can be taken from the global variables of the sample
## with this function call the affected copies are calculated for the variant
aff_germ_copies(af=AF,
                tcn=TCN,
                purity=PURITY,
                chr=CHR,
                sex=SEX)

## -----------------------------------------------------------------------------
## the function for somatic variants works the same way as the germline function

AF = elementMetadata(GR_SOM_SMALL_VARS[1])[["af"]]
CHR = seqnames(GR_SOM_SMALL_VARS[1]) %>%
  as.character()
TCN = elementMetadata(
  subsetByOverlaps(GR_SCNA, GR_SOM_SMALL_VARS[1])
  )[["tcn"]]

aff_som_copies(af=AF,
               chr=CHR,
               tcn=TCN,
               purity=PURITY,
               sex=SEX)

## -----------------------------------------------------------------------------
## as an example we calculate the affected copies for the somatic variants:
GR_SOM_SMALL_VARS %>%
  ## cnv information for every variant is required.. merge with cnv object
  IRanges::mergeByOverlaps(GR_SCNA) %>% 
  as_tibble() %>%
  ## select relevant columns
  select(chr=1, pos=2, gene, af, tcn) %>%
  mutate_at(.vars=c("tcn", "af"), .funs=as.numeric) %>%
  rowwise() %>%
  mutate(
    aff_copies = aff_som_copies(chr, af, tcn, PURITY, SEX),
    wt_copies = tcn-aff_copies
  )

## -----------------------------------------------------------------------------
predict_per_variant(purity=PURITY, 
                    sex=SEX,
                    somCna=GR_SCNA,
                    somSmallVars=GR_SOM_SMALL_VARS)


## ----results = FALSE----------------------------------------------------------
fp <- predict_zygosity(
  purity = PURITY, 
  ploidy = PLOIDY,
  sex = SEX,
  somCna = GR_SCNA, 
  somSmallVars = GR_SOM_SMALL_VARS, 
  germSmallVars = GR_GERM_SMALL_VARS, 
  geneModel = GR_GENE_MODEL,
  bamDna = FILE_BAM,
  vcf=VCF,
  haploBlocks = GR_HAPLOBLOCKS
)

## -----------------------------------------------------------------------------
ZP_ov(fp)

## -----------------------------------------------------------------------------
gene_ov(fp, OR2J1)

## -----------------------------------------------------------------------------
sessionInfo()

func_end <- function(ZP_env){
  vm("end", -1, ZP_env)
}
func_start <- function(ZP_env){
  mes <- paste0(as.character(sys.call(-1)[1]))
  vm(mes, 1, ZP_env)
}
increment_log_depth <- function(ZP_env) {
  ZP_env$global_ZygosityPredictor_variable_debug_log_depth <- ZP_env$global_ZygosityPredictor_variable_debug_log_depth + 1
}
decrement_log_depth <- function(ZP_env) {
  ZP_env$global_ZygosityPredictor_variable_debug_log_depth <- ZP_env$global_ZygosityPredictor_variable_debug_log_depth - 1
}
increment_call_depth <- function(ZP_env) {
  ZP_env$global_ZygosityPredictor_variable_debug_call_depth <- ZP_env$global_ZygosityPredictor_variable_debug_call_depth + 1
}
decrement_call_depth <- function(ZP_env) {
  ZP_env$global_ZygosityPredictor_variable_debug_call_depth <- ZP_env$global_ZygosityPredictor_variable_debug_call_depth - 1
}
#' @importFrom readr write_tsv
store_log <- function(geneDir, obj, file){
  if(!is.null(geneDir)){
    if(!is.null(obj)){
      write_tsv(obj, 
                file=file.path(geneDir, 
                               file))      
    }
  }
}
#' @keywords internal
#' @importFrom stringr %>%
#' @importFrom GenomicRanges GRanges elementMetadata
#' @importFrom purrr compact
#' @importFrom dplyr summarize pull
insert_missing_cnv_regions <- function(somCna, geneModel, sex, ploidy, ZP_env){
  vm(as.character(sys.call()[1]), 1, ZP_env)
  seqnames <- start <- end <- tcn <- . <- NULL
  if(sex=="male"){
    allowed_chr <- allowed_inputs("chrom_names")
  } else {
    allowed_chr <- allowed_inputs("chrom_names") %>% 
      .[which(!str_detect(., "Y"))]
  }
  tbl_refgen <-
    as_tibble(geneModel) %>%
    group_by(seqnames) %>%
    summarize(mn=min(start),
              mx=max(end)) %>%
    filter(seqnames %in% allowed_chr)
  tbl_cnv <- 
    as_tibble(somCna) %>%
    mutate(tcn=as.numeric(tcn))
  new_somCna <- lapply(as.character(tbl_refgen$seqnames), function(CHR){
    all_chr <- tbl_cnv %>%
      filter(seqnames==CHR) %>%
      arrange(start)
    mn <- tbl_refgen %>% 
      filter(seqnames==CHR) %>% pull(mn)
    mx <- tbl_refgen %>% 
      filter(seqnames==CHR) %>% pull(mx)    
    
    if(nrow(all_chr)==0){
      tbl_refgen %>% 
        filter(seqnames==CHR) %>%
        select(seqnames, start=mn, end=mx) %>% 
        mutate(tcn=ploidy, cna_type="no CNA type annotated") %>%
        GRanges() %>%
        return()
    } else {
      if(all_chr[1,]$start<mn){mn <- all_chr[1,]$start}
      if(all_chr[nrow(all_chr),]$end>mx){mx <- all_chr[nrow(all_chr),]$end}
      poslist <- 
        apply(all_chr, 1, function(ROW){
          return(c(start=as.numeric(ROW[["start"]]), 
                   end=as.numeric(ROW[["end"]])))
        }, simplify=FALSE) %>%
        append(list(c(start=NA, end=mn)),.) %>%
        append(.,list(c(start=mx, end=NA)))
      
      to_insert <- lapply(c(2:length(poslist)), function(I){
        diff <- poslist[[I]][["start"]]-poslist[[I-1]][["end"]]
        if(diff>1){
          return(c(start=poslist[[I-1]][["end"]]+1, 
                   end=poslist[[I-1]][["end"]]+diff-1))
        }
      })  %>%
        compact()
      
      if(length(to_insert)==0){
        GRanges(all_chr) %>% return()
      } else {
        return(c(GRanges(all_chr),  bind_rows(to_insert) %>%
                   mutate(chr=CHR, tcn=ploidy, 
                          cna_type="no CNA type annotated") %>%
                   GRanges()))        
      }
    }
  })
  comb_somCna <- Reduce(function(x,y)c(x,y),new_somCna)
  func_end(ZP_env)
  return(comb_somCna)
}

#' @keywords internal
#' @importFrom IRanges mergeByOverlaps
#' @importFrom dplyr mutate select mutate_at rowwise
#' @importFrom stringr str_detect
merge_sCNAs <- function(obj, somCna, ZP_env){
  func_start(ZP_env)
  alt <- af <- tcn <- tcn_assumed <- cna_type <- gene <- ref <- NULL
  merged <- obj %>%
    mergeByOverlaps(somCna) %>% as_tibble() %>%
    mutate(cna_type=ifelse(str_detect(cna_type, "LOH"), "LOH", "HZ")) %>%
    select(chr=1, pos=2, gene, ref, alt, af, tcn, cna_type, all_imb, gt_cna, 
           seg_id, tcn_assumed) %>%
    mutate_at(.vars = c("af", "tcn"), .funs=as.numeric) %>%
    rowwise() #%>%
  func_end(ZP_env)
  return(merged)
}
#' @keywords internal
#' description follows
#' @importFrom stringr %>% str_detect
#' @importFrom GenomicRanges GRanges elementMetadata
#' @importFrom IRanges mergeByOverlaps
#' @importFrom dplyr mutate filter
prepare_somatic_variant_table <- function(somSmallVars, templateGenes, 
                                          somCna, purity, sex, ZP_env){
  func_start(ZP_env)
  cna_type <- gene <- ref <- alt <- af <- tcn <- cna_type <- chr <- 
    aff_cp <- wt_cp <- . <- tcn_assumed <- NULL
  if(!is.null(somSmallVars)){
    reduced_to_templateGenes <- 
      somSmallVars[which(somSmallVars$gene %in% templateGenes)]
    if(length(reduced_to_templateGenes)>0){
      tbl_prepared_variants <- reduced_to_templateGenes %>%
        merge_sCNAs(., somCna, ZP_env) %>%
        
        ################################
        mutate(
          origin="somatic",
          class=define_class(ref, alt),
          aff_cp = aff_som_copies(chr, af, tcn, purity, sex),
          wt_cp = tcn-aff_cp,
          vn_status=case_when(
            (isTRUE(str_detect(cna_type,'LOH'))|chr=="X"&sex=="male")&
              wt_cp<=0.5 ~ 2,
            TRUE ~ 1
          ),
          pre_info = 
            ifelse(isTRUE(str_detect(cna_type,'LOH')),
              ifelse(wt_cp<=0.5,
                paste(
                    'somatic-variant -> LOH -> left wt copies',round(wt_cp,2),
                    '-> all copies affected'),
                paste(
                    'somatic-variant -> LOH -> ',
                    "left wt copies",round(wt_cp,2),"-> wt copies left")),
              ifelse(chr=="X"&sex=="male",
                ifelse(wt_cp<=0.5,
                  paste(
                    "somatic-variant -> chrX & male -> ",
                    "left wt copies",round(wt_cp,2),"-> all copies affected"),
                  paste(
                    'somatic-variant -> chrX & male -> left wt copies ',
                    round(wt_cp,2),
                    '-> wt copies left')),
                  'somatic-variant -> no LOH -> wt copies left')
          )
        )      
    } else {
      tbl_prepared_variants <- NULL
    }
  } else {
    tbl_prepared_variants <- NULL
  }
  func_end(ZP_env)
  return(tbl_prepared_variants)
}
#' @keywords internal
#' description follows
#' @importFrom stringr %>% str_detect
#' @importFrom GenomicRanges GRanges
#' @importFrom IRanges subsetByOverlaps
#' @importFrom purrr compact
#' @importFrom dplyr filter as_tibble select bind_rows mutate
extract_all_dels_of_sample <- function(somCna, geneModel, DEL_TYPE, 
                                       byTcn, sex, ploidy, include_dels, ZP_env){
  #print("jou")
  func_start(ZP_env)
  tcn <- cna_type <- seqnames <- gene <- . <- NULL
  if(include_dels==TRUE){
    TBL_CNV <- somCna %>%
      as_tibble() %>%
      mutate(tcn=as.numeric(tcn))
    if(DEL_TYPE=="homdel"){
      if(byTcn==TRUE){
        relevant_CNAs <- TBL_CNV %>%
          filter(round(tcn)==0)
      } else {
        relevant_CNAs <- TBL_CNV %>%
          filter(str_detect(cna_type, allowed_inputs("cna_homdel_annotation"))) 
      }
      pre_info <- 'homdel -> all copies affected'
    } else {
      if(byTcn==TRUE){
        if(sex=="female"){
          relevant_CNAs <- TBL_CNV %>%
            filter(round(tcn)<ploidy)
        } else {
          relevant_CNAs <- TBL_CNV %>%
            filter(!str_detect(seqnames, "X|Y")&round(tcn)<ploidy)
        }
      } else {
        relevant_CNAs <- TBL_CNV %>%
          filter(!str_detect(cna_type, 
                             allowed_inputs("cna_homdel_annotation"))) %>%
          filter(str_detect(cna_type, 
                            allowed_inputs("cna_incompletedel_annotation")))
      }
      pre_info <- "incompletedel -> wt-copies left"
    } 
    if(nrow(relevant_CNAs)==0){
      #return(NULL)
      full_CNAs <- NULL
    } else {
      merged_CNAs <- apply(relevant_CNAs, 1, function(CNV){
        GR_single_CNV <- GRanges(as_tibble(t(CNV)))
        suppressWarnings(
        rel <- subsetByOverlaps(geneModel, GR_single_CNV) %>%
          as_tibble()
        )
        if(nrow(rel)==0){
          #return(NULL)
          full_CNAs <- NULL
        } else {
          rel %>%
            select(gene) %>%
            mutate(chr=CNV[["seqnames"]],
                   pos=as.numeric(CNV[["start"]]),
                   tcn=as.numeric(CNV[["tcn"]]),
                   tcn_assumed=FALSE,
                   af=NA,
                   origin="somatic",
                   class=DEL_TYPE,
                   ref= NA,
                   alt= NA,
                   cna_type=DEL_TYPE,
                   all_imb=NA,
                   gt_cna=NA,
                   seg_id=NA,
                   aff_cp=ploidy-as.numeric(CNV[["tcn"]]),
                   wt_cp=as.numeric(CNV[["tcn"]]),
                   pre_info=pre_info,
                   vn_status=ifelse(DEL_TYPE=="homdel", 2, 1)
            ) %>%
            return()
        }
      })%>% compact() %>% bind_rows() %>% unique()
      if(nrow(merged_CNAs)==0){
        full_CNAs <- NULL
        #return(NULL)
      } else {
        full_CNAs <- merged_CNAs
      }
    }
  } else {
    #return(NULL)
    full_CNAs <- NULL
  }
  func_end(ZP_env)
  return(full_CNAs)
}
#' @keywords internal
#' description follows
#' @importFrom stringr %>%
#' @importFrom dplyr mutate select
prepare_germline_variants <- function(germSmallVars, somCna, purity, sex, ZP_env){
  cna_type <- gene <- ref <- alt <- af <- tcn <- cna_type <- chr <- aff_cp <- 
    origin <- pos <- wt_cp <- pre_info <- . <- tcn_assumed <- NULL
  func_start(ZP_env)
  if(is.null(germSmallVars)){
    df_germ <- NULL
  } else {
    df_germ <- germSmallVars %>% 
      merge_sCNAs(., somCna, ZP_env) %>%
      mutate(
        origin="germline",
        class=define_class(ref, alt),
        aff_cp = aff_germ_copies(chr, af, tcn, purity, sex, 1),
        wt_cp = tcn-aff_cp,
        vn_status=case_when(
          isFALSE(cna_type=='LOH') ~ 1,
          af>=0.5 ~ 2,
          TRUE ~ -1 
        ),
        pre_info = 
          ifelse(isTRUE(cna_type=='LOH'),
            ifelse(af>=0.5,
                 paste(
                   'germline-variant -> LOH ->',
                   'AF >= 0.5 -> all copies affected'),
                 paste(
                   'germline-variant -> LOH ->',
                   ' AF < 0.5 -> variant lost in tumor')),
            'germline-variant -> no LOH -> wt copies left')
      ) %>%
      select(gene,
             origin,
             class,
             chr,
             pos,
             ref,
             alt,
             af,
             tcn, 
             tcn_assumed,
             cna_type,
             all_imb,
             gt_cna,
             seg_id,
             aff_cp,
             wt_cp,
             pre_info,
             vn_status
      ) 
  }
  func_end(ZP_env)
  return(df_germ)
}

ascii_to_dec <- function(ascii_encoded){
  if(is.na(ascii_encoded)){
    return(NA)
  } else {
    exp <- (as.integer(charToRaw(ascii_encoded))-33)/(-10)
    if(length(exp)>1){
      exp <- mean(exp)
    }
    # Convert the QUAL string to raw hexadecimal
    return(10^exp)    
  }
}

#' @importFrom knitr kable
print_tibble <- function(tbl_in){
  # res <- paste(as.character(knitr::kable(tbl_in)), collapse="\n")
  res <- paste(as.character(knitr::kable(tbl_in)), collapse="\n")
  return(res)
}
#' @importFrom GenomicRanges GRanges seqnames start end
#' @importFrom IRanges IRanges
split_genomic_range <- function(gr, exclude_positions) {
  # Sort the exclude_positions and ensure they are unique
  exclude_positions <- sort(unique(exclude_positions))
  # Initialize the list to store subranges
  subranges <- list()
  # Add the first subrange from the beginning of the range to the first exclude position
  if (exclude_positions[1] > start(gr)) {
    subranges[[1]] <- GRanges(
      seqnames = seqnames(gr),
      ranges = IRanges(start = start(gr), end = exclude_positions[1] - 1)
    )
  }
  # Add subranges between the exclude positions
  for (i in seq_along(exclude_positions)[-1]) {
    subranges[[length(subranges) + 1]] <- GRanges(
      seqnames = seqnames(gr),
      ranges = IRanges(start = exclude_positions[i - 1] + 1, end = exclude_positions[i] - 1)
    )
  }
  # Add the last subrange from the last exclude position to the end of the range
  if (exclude_positions[length(exclude_positions)] < end(gr)) {
    subranges[[length(subranges) + 1]] <- GRanges(
      seqnames = seqnames(gr),
      ranges = IRanges(start = exclude_positions[length(exclude_positions)] + 1, end = end(gr))
    )
  }
  # Combine the subranges into a single GRanges object
  split_gr <- do.call(c, subranges)
  return(split_gr)
}
#' @importFrom stringr str_detect %>%
#' @importFrom dplyr filter pull
find_unplaus_issues <- function(unplaus){
  if(nrow(unplaus)>0){
    if(any(str_detect(unplaus$unplausible, "1"))){
      issues <- unplaus %>%
        filter(str_detect(unplausible, "1")) %>%
        pull(comb) %>%
        paste(collapse = ", ") %>%
        paste("unplausible phasing combs:",.)
    } else {
      issues <- "none"
    }    
  } else {
    issues <- "none"
  }
}
#' @importFrom dplyr arrange case_when desc mutate pull select
#' @importFrom stringr %>% str_detect 
eval_phasing_new <- function(all_comb, df_gene, printLog, ZP_env){
  func_start(ZP_env)
  rare_case <- ""
  most_relevant_comb <- all_comb %>%
    arrange(desc(score), wt_cp, desc(conf)) %>%
    .[1,] 
  if(most_relevant_comb$score==2){
    ## take most relevant combination
    gene_clear <- TRUE
    wt_cp_clear <- TRUE
    case <- "allaff"
    conf <- most_relevant_comb$conf
    conf_case <- "mostrel"
    eval <- most_relevant_comb$phasing
  } else {
    if(nrow(all_comb)==1){
      ## take the only combination
      eval <- most_relevant_comb$phasing
      if(most_relevant_comb$score==0){
        gene_clear <- FALSE
        wt_cp_clear <- FALSE
        case <- "onecomb_unphased"
        conf <- 0
        conf_case <- "poss_allaff_unphased"
      } else {
        gene_clear <- TRUE
        if(min(most_relevant_comb$min_poss_wt_cp, na.rm=TRUE)<0.5){
          conf <- most_relevant_comb$conf
          conf_case <- "poss_allaff_but_same"
        } else {
          conf <- 1
          conf_case <- "impos_allaff"
        }
        if(is.na(most_relevant_comb$wt_cp)){
          wt_cp_clear <- FALSE
          case <- "onecomb_wtNA"
        } else {
          wt_cp_clear <- TRUE
          case <- "onecomb_wtclear"
        }
      }
    } else {
      rare_case <- eval_rare_case(all_comb)
      ## evaluate all combinatioons
      if(0 %in% all_comb$score){
        ## potentially reachable all copies affectd by unphased combinations
        gene_clear <- wt_cp_clear <- FALSE
        case <- "morecomb_unphased"
        conf <- 0
        conf_case <- "poss_allaff_unphased"
        eval <- NA
      } else {
        ## each combination could be solved
        gene_clear <- TRUE
        if(any(!is.na(all_comb$wt_cp))){
          if(sum(!is.na(all_comb$wt_cp))==nrow(all_comb)){
            wt_cp_clear <- TRUE
            case <- "morecomb_allsolved"
          } else {
            if(min(all_comb$wt_cp, na.rm=TRUE)<=min(all_comb$min_poss_wt_cp, na.rm=TRUE)){
              ## if potential min wt copies are higher than clear ones
              wt_cp_clear <- TRUE
              case <- "morecomb_wtclear"
            } else {
              wt_cp_clear <- FALSE
              case <- "morecomb_wtNA"
            }            
          }
        } else {
          ## no wt copies annoatted 
          wt_cp_clear <- FALSE
          case <- "morecomb_onlyexc"
        }
        
        if(min(all_comb$min_poss_wt_cp, na.rm=TRUE)<0.5){
            ## if this is the case, a possible diff would lead to all copies affected but it was phased as same
            ## for confidence measurement, we here need to take the phasing confidence which indicates this
            conf <- all_comb %>%
              mutate(p_comb=case_when(
                min_poss_wt_cp<0.5 ~ conf,
                TRUE ~ 1
              )) %>%
              pull(p_comb) %>%
              prod()
            rare_case <- " confidence adjusted"
            conf_case <- "morecomb_poss_allaff_but_same"
            all_phasing_annotations <- all_comb[which(!is.na(all_comb$phasing)),]
            if(nrow(all_phasing_annotations)>0){
              merged <- paste(all_phasing_annotations$phasing, collapse = ", ")
              eval <- case_when(
                str_detect(merged, "imbalance") ~ "imbalance-phasing",
                str_detect(merged, "haploblock") ~ "haploblock-phasing",
                str_detect(merged, "indirect") ~ "indirect-phasing",
                TRUE ~ "direct-phasing"
              )
            } else {
              eval <- "aff_cp"
            }
        } else {
            conf <- 1
            conf_case <- "morecomb_imposs_allaff"
            eval <- "aff_cp"
        }
      } ## each combination was solved
    } ## more than one comb
  } ## max score = 1
  if(gene_clear){
    gene_result <- most_relevant_comb %>%
      mutate(info=paste("most relevant phasing combination:",comb)) %>%
      select(gene, score, phasing, wt_cp, info) 
  } else {
    gene_result <- most_relevant_comb %>%
      select(gene) %>%
      mutate(info="unsolvable phasing combination", phasing="unsolved", score=0, wt_cp=NA)
  }
  wt_cp_range <- paste(round(min(all_comb$min_poss_wt_cp, na.rm=TRUE),2), "-",
        round(max(all_comb$max_poss_wt_cp, na.rm=TRUE),2))
  if(wt_cp_clear){
    wt_cp_result <- gene_result 
  } else {
    wt_cp_result <- gene_result %>%
      select(-wt_cp) %>%
      mutate(wt_cp=NA)
  }
  unplaus <- all_comb[which(!is.na(all_comb$unplausible)),]
  issues <- find_unplaus_issues(unplaus)
  annotated_result <- wt_cp_result %>%
    mutate(conf=conf, 
           gene_clear=gene_clear, wt_cp_clear=wt_cp_clear, 
           case=paste0(case, rare_case), n_mut=nrow(df_gene),
           conf_case=conf_case,
           wt_cp_range=wt_cp_range,
           warning=issues,
           eval_by=eval
           )
  func_end(ZP_env)
  return(annotated_result)
}
#' @keywords internal
#' desicion tree if one variant already affects all copies in pre evaluation
#' @importFrom dplyr mutate select filter
#' @importFrom stringr str_detect
eval_one_mut_affects_all <- function(df_gene, printLog, ZP_env){
  func_start(ZP_env)
  concern_info <- df_gene %>% 
    filter(vn_status==2) %>%
    .[1,] %>%
    mutate(n_mut=nrow(df_gene),
           info=paste("all copies affected by variant:", mut_id),
           conf=1,
           eval_by=ifelse(str_detect(class, "homdel"), "homdel","aff_cp")) %>%
    select(gene, n_mut, score=vn_status, conf, info, eval_by, wt_cp) %>%
    mutate(warning="none", 
           wt_cp_range=paste(round(wt_cp,2), 
                             round(wt_cp,2), 
                             sep=" - "), 
           phasing="none")
  append_loglist(concern_info$info, ZP_env=ZP_env)
  func_end(ZP_env)
  return(concern_info)
}
#' @keywords internal
#' if no variant affecta all copies
#' @importFrom dplyr mutate select
eval_one_mut <- function(df_gene, printLog, ZP_env){
  func_start(ZP_env)
  concern_info <- df_gene %>%
    mutate(n_mut=1,
           conf=1,
           eval_by="aff_cp") %>%
    select(gene, n_mut, score=vn_status, conf, 
           info=pre_info, 
           wt_cp,
           eval_by) %>%
    mutate(warning="none", 
           wt_cp_range=paste(round(wt_cp,2), 
                             round(wt_cp,2), 
                             sep=" - "), 
           phasing="none")
  append_loglist(concern_info$info, ZP_env=ZP_env)
  func_end(ZP_env)
  return(concern_info)
}
eval_lost_in_tumor <- function(df_gene, printLog, ZP_env){
  func_start(ZP_env)
  concern_info <- df_gene %>%
    group_by(gene) %>%
    summarize(
      n_mut=length(gene),
      info=paste(unique(pre_info), collapse = ", "),
      wt_cp=min(wt_cp),
    ) %>%
    mutate(conf=1,
           eval_by="allele-freq",
           score=1,
           warning="variant lost in tumor", 
           wt_cp_range=paste(round(wt_cp, 2), round(wt_cp, 2), sep=" - "), 
           phasing="none")
  append_loglist(concern_info$info, ZP_env=ZP_env)
  func_end(ZP_env)
  return(concern_info)
}
#' @importFrom dplyr mutate select filter left_join bind_rows
remove_duplicated_variants <- function(df_gene_raw, ZP_env){
  vm(as.character(sys.call()[1]), 1, ZP_env)
  n_vars_per_pos <- df_gene_raw %>%
    group_by(chr, pos) %>%
    tally() %>%
    filter(n>1)
  if(nrow(n_vars_per_pos)!=0){
    var_to_remove <- apply(n_vars_per_pos, 1, function(VAR){
      selected <- df_gene_raw %>% 
        filter(chr==VAR[["chr"]],
               pos==VAR[["pos"]]) %>%
        #filter(wt_cp==min(.$wt_cp)) %>%
        ## if both vars have the same impac we just select the first
        .[1,]
      warning(
        "more than one variant detected at position:", 
        VAR[["chr"]], ":", VAR[["pos"]],
        "\n  maybe an indel with shifted annotation of position?",
        "\nSelecting variant:",
        selected$mut_id)
      to_remove <- df_gene_raw %>% 
        filter(chr==VAR[["chr"]],
               pos==VAR[["pos"]]) %>%
        filter(mut_id!=selected$mut_id) %>%
        select(chr, pos, mut_id) %>%
        mutate(remove=TRUE)
      return(to_remove)
    }) %>%
      bind_rows()
    df_gene <- df_gene_raw %>%
      left_join(var_to_remove, 
                by=c("chr"="chr", "pos"="pos", "mut_id"="mut_id")) %>%
      filter(is.na(remove)) %>%
      select(-remove)
  } else {
    df_gene <- df_gene_raw
  }
  func_end(ZP_env)
  return(df_gene)
}
#' @keywords internal
#' description follows
#' @importFrom stringr %>%
#' @importFrom dplyr filter mutate as_tibble case_when
predict_zygosity_genewise <- function(GENE, 
                                      evaluation_per_variant, 
                                      bamDna, 
                                      bamRna, 
                                      showReadDetail,
                                      printLog, 
                                      purity, 
                                      sex, 
                                      haploBlocks,
                                      vcf, 
                                      distCutOff, 
                                      logDir, 
                                      somCna, 
                                      snpQualityCutOff, 
                                      phasingMode,
                                      AllelicImbalancePhasing, 
                                      ZP_env){
  #func_start(ZP_env)
  vm("predict_zygosity_genewise", 1, ZP_env) ## do not replcae!!!
  start_gene_eval <- Sys.time()
  increment_loglist(ZP_env)
  append_loglist(GENE, ZP_env=ZP_env)
  gene <- pre_info <-  chr <-  pos <- wt_cp <-  mut_id <- status <- . <- 
    score <- comb <- dist <- tcn <- info <- n <- all_comb <- 
    read_level_phasing_info <- copy_number_phasing_info <- mat_phased_gene <- 
  mat_info_gene <- NULL 
  #start_gene_eval <- Sys.time()
  pre_df_gene <- evaluation_per_variant %>% filter(gene==GENE)
  ## all germline variants which are lost in the tumor are excluded, 
  ## vn_status = -1
  ## check if variants at the same position are present
  df_gene_raw <- remove_duplicated_variants(pre_df_gene, ZP_env=ZP_env)
  df_gene <- df_gene_raw %>%
    filter(vn_status>=0)
 #print(10)
 #print(df_gene_raw)
 #print(df_gene)
  ## check if variants remain in that gene
  if(nrow(df_gene)==0){
    eval_for_gene <- eval_lost_in_tumor(df_gene_raw, printLog, ZP_env=ZP_env)
   #print(eval_for_gene)
      ## all variants lost in tumor
      append_loglist("all variants lost in tumor", ZP_env=ZP_env)
  } else if(any(df_gene$vn_status==2)){
      append_loglist("one of", nrow(df_gene), "variants affects all copies", ZP_env=ZP_env)
      eval_for_gene <- eval_one_mut_affects_all(df_gene, printLog, ZP_env)
    ## (2): if only one variant is present  
  } else if(nrow(df_gene)==1){
      append_loglist("one variant that does not affect all copies", ZP_env=ZP_env)
      eval_for_gene <- eval_one_mut(df_gene, printLog, ZP_env)
    ## (3): more than one variant present in gene   
  } else {  
      append_loglist(nrow(df_gene),
            "heterozygous variants detected: Initializing haplotype phasing", ZP_env=ZP_env)
      if(!is.null(logDir)){
        geneDir <- file.path(logDir, GENE)
        dir.create(geneDir)
      } else {
        geneDir <- NULL
      }
      ## returns two tibbles: first one direct phasing combinations
      ## second one indirect phasing combinations
      full_phasing_result <- phase(df_gene, somCna, bamDna, purity, sex, bamRna, 
                                   haploBlocks, vcf, distCutOff, printLog,
                                   geneDir, showReadDetail, 
                                   snpQualityCutOff, phasingMode, 
                                   AllelicImbalancePhasing, ZP_env)
     #print(full_phasing_result)
      all_comb <- full_phasing_result[[1]] %>%
        mutate(gene=GENE)
      eval_for_gene <- eval_phasing_new(all_comb, df_gene,  
                                    printLog, ZP_env)
      #print(full_phasing_result)
      read_level_phasing_info <- full_phasing_result[[2]] %>%
        mutate(gene=GENE)
      copy_number_phasing_info <- full_phasing_result[[5]] %>%
        mutate(gene=GENE)
      if(nrow(copy_number_phasing_info)==0){
        copy_number_phasing_info <- NULL
      }
      store_log(geneDir, read_level_phasing_info, "all_phasing_combinations.tsv")
      store_log(geneDir, copy_number_phasing_info, "all_phasing_combinations.tsv")
      mat_phased_gene <- list()
      mat_info_gene <- list()
      mat_phased_gene[[GENE]] <- full_phasing_result[[3]]
      mat_info_gene[[GENE]] <- full_phasing_result[[4]]
  } 
  end_gene_eval <- Sys.time()
  df_reduced <- eval_for_gene %>% as_tibble() %>%
      mutate(
        status=case_when(
          score==2 ~ "all_copies_affected",
          score==1 ~ "wt_copies_left",
          TRUE ~ "undefined"
        ),  
        eval_time_s=pt(start_gene_eval, end_gene_eval)
        ) 
  zygosity_gene <- list(pre_df_gene, 
                          list(df_reduced, 
                               all_comb,
                               read_level_phasing_info,
                               mat_phased_gene,
                               mat_info_gene,
                               copy_number_phasing_info)
                          )
  log_message <- unlist(ZP_env$global_ZygosityPredictor_variable_debug_loglist) %>% paste(collapse = "\n")
  func_end(ZP_env)
  return(append(zygosity_gene, log_message))
}
#' @keywords internal
pt <- function(start_gene_eval, end_gene_eval){
  round(
    as.numeric(
      difftime(
        end_gene_eval, 
        start_gene_eval,
        units="secs")
    ),3)
}
#' @keywords internal
#' description follows 
#' @importFrom dplyr case_when
get_classification <- function(data){
  case_when(
    data$class=="homdel" ~ "homdel",
    nchar(data$alt)==nchar(data$ref) ~ "snv",
    nchar(data$alt)>nchar(data$ref) ~ "ins",
    nchar(data$alt)<nchar(data$ref) ~ "del",
  ) %>%
    return()
}
#' @keywords internal
#' description follows
#' @importFrom dplyr case_when
define_class <- function(ref, alt){
  case_when(
    nchar(alt)==nchar(ref) ~ "snv",
    nchar(alt)>nchar(ref) ~ "ins",
    nchar(alt)<nchar(ref) ~ "del",
  ) %>%
    return()
}
#' @keywords internal
#' @importFrom stringr %>%
#' @importFrom dplyr as_tibble filter select mutate mutate_all
combine_uncovered_input_variants <- function(somSmallVars, germSmallVars,
                                             som_covered, germ_covered,
                                             templateGenes, ZP_env){
  func_start(ZP_env)
  mid <- seqnames <- start <- ref <- alt <- gene <- uncovered_som <- 
    uncovered_germ <- NULL
  if(!is.null(somSmallVars)){
    uncovered_som <-  as_tibble(somSmallVars) %>%
      filter(!mid %in% som_covered) %>%
      select(chr=seqnames, pos=start,ref, alt, gene) %>%
      mutate(origin="somatic") %>%
      mutate_all(.funs = as.character)
  } 
  if(!is.null(germSmallVars)){
    uncovered_germ <- as_tibble(germSmallVars) %>%
      filter(!mid %in% germ_covered) %>%
      select(chr=seqnames, pos=start,ref, alt, gene) %>%
      mutate(origin="germline") %>%
      mutate_all(.funs = as.character)
  }
  if(nrow(bind_rows(uncovered_som, uncovered_germ))==0){
    combined_uncovered <- NULL
  } else {
    combined_uncovered <- bind_rows(uncovered_som, uncovered_germ) %>%
      filter(gene %in% templateGenes)
  }
  func_end(ZP_env)
  return(combined_uncovered)
}
#' @keywords internal
#' @importFrom stringr %>%
#' @importFrom dplyr as_tibble group_by mutate ungroup filter
#' @importFrom purrr compact
combine_main_variant_tables <- function(df_germ, df_som, df_homdels,
                                        templateGenes, purity, ZP_env){
  func_start(ZP_env)
  gene <- . <- NULL
  df_all_mutations_unfiltered <- list(df_germ, df_som, df_homdels) %>% 
    compact() %>%
    Reduce(function(x,y)rbind(x,y),.) %>% 
    as_tibble() %>% 
    group_by(gene) %>% 
    mutate(mut_id=paste0("m", seq_len(length(gene))))%>% 
    ungroup() %>%
    mutate(purity=purity) 
  df_all_mutations <- df_all_mutations_unfiltered %>%
    filter(gene %in% templateGenes)
  if(nrow(df_all_mutations_unfiltered)!=nrow(df_all_mutations)){
    warning(
      abs(nrow(df_all_mutations_unfiltered)-nrow(df_all_mutations)),
      "variants in genes not found in reference genome annotation",
      "(geneModel). They are removed from the analysis.",
      "Please provide an annotation that contains every gene from the",
      "variant inputs"
    )
  }
  func_end(ZP_env)
  return(df_all_mutations)
}
#' @keywords internal
#' @importFrom stringr %>%
#' @importFrom dplyr bind_rows group_by mutate relocate select ungroup relocate
bind_incdel_to_pre_eval <- function(df_incompletedels, df_all_mutations, ZP_env){
  purity <- gene <- mut_id <- NULL
  if(is.null(df_all_mutations)&is.null(df_incompletedels)){
    full_df_all_mutations <- NULL
  } else {
    if(is.null(df_incompletedels)){
      full_df_all_mutations <- df_all_mutations %>%
        select(-purity) %>%
        relocate(gene, mut_id, 2:13)
    } else {
      full_df_all_mutations <- bind_rows(
        df_all_mutations,
        df_incompletedels %>% 
          unique() %>% 
          mutate(mut_id=NA)
      ) %>% group_by(gene) %>% 
        mutate(mut_id=paste0("m", seq_len(length(gene))))%>% 
        ungroup()%>%
        select(-purity) %>%
        relocate(gene, mut_id, 2:13)       
    }    
  }
  return(full_df_all_mutations)
}
#' @keywords internal
#' @importFrom stringr %>%
#' @importFrom dplyr filter bind_rows mutate select
bind_incdel_to_final_eval <- function(df_incompletedels, final_output, ZP_env){
  func_start(ZP_env)
  gene <- 0
  if(is.null(final_output)&is.null(df_incompletedels)){
    full_output <- NULL
  } else {
    if(is.null(df_incompletedels)){
      full_output_pre <- final_output
    } else {
      full_output_pre <- bind_rows(
        final_output,
        df_incompletedels %>%
          filter(!gene %in% final_output$gene) %>%
          select(gene) %>%
          unique() %>%
          mutate(status="wt_copies_left",
                 info="somatic-incompletedel")
      )       
    }
    full_output <- full_output_pre %>%
      mutate(status=factor(status, 
                           levels=c("all_copies_affected", "wt_copies_left", 
                                    "undefined")))
  }
  func_end(ZP_env)
  return(full_output)
}

set_global_variables <- function(debug, verbose, printLog, ZP_env){
  ZP_env$global_ZygosityPredictor_variable_debug_call_depth <- 0
  ZP_env$global_ZygosityPredictor_variable_debug_log_depth <- 0
  ZP_env$global_ZygosityPredictor_variable_debug_timelist  <- list()
  ZP_env$global_ZygosityPredictor_variable_debug_verbose <- verbose
  ZP_env$global_ZygosityPredictor_variable_debug_printLog <- printLog
}
#' @keywords internal
catt <- function(printLog=FALSE, level, text){
  if(printLog==TRUE){
    message(rep("  ", level), text)
  }
}
increment_loglist <- function(ZP_env){
  ZP_env$global_ZygosityPredictor_variable_debug_loglist <- list()
}
#' @importFrom stringr str_replace_all %>%
append_loglist <- function(..., ZP_env){
  filler <- paste(rep("  ", ZP_env$global_ZygosityPredictor_variable_debug_call_depth), collapse = "")
  appendix <- paste(..., collapse = " ") %>%
    paste0(filler,.) %>%
    str_replace_all("\n", paste0("\n",filler))
  ZP_env$global_ZygosityPredictor_variable_debug_loglist <- append(ZP_env$global_ZygosityPredictor_variable_debug_loglist, appendix)
  if(ZP_env$global_ZygosityPredictor_variable_debug_printLog==TRUE){
    message(appendix)
  }
}
add_timestamp <- function(ZP_env){
  ZP_env$global_ZygosityPredictor_variable_debug_timelist[[ZP_env$global_ZygosityPredictor_variable_debug_call_depth]] <- Sys.time()
}
remove_timestamp <- function(ZP_env){
  ZP_env$global_ZygosityPredictor_variable_debug_timelist <- 
    ZP_env$global_ZygosityPredictor_variable_debug_timelist[c(1:ZP_env$global_ZygosityPredictor_variable_debug_call_depth)]
}
#' @importFrom magrittr %>%
vm <- function(mes, depth=0, ZP_env){
  if(depth>0){
    to_add <- "|-"
  }  
  if(depth<0){
    to_add <- "+-"
  }
  if(ZP_env$global_ZygosityPredictor_variable_debug_verbose==TRUE){
    to_print <- paste0(paste(rep("| ", ZP_env$global_ZygosityPredictor_variable_debug_call_depth),collapse=""), to_add,mes)
  } 
  if(depth>0){
    increment_call_depth(ZP_env)
    add_timestamp(ZP_env)
    timestamp <- ""
  }  
  if(depth<0){
    timestamp <- pt(ZP_env$global_ZygosityPredictor_variable_debug_timelist[[ZP_env$global_ZygosityPredictor_variable_debug_call_depth]], 
                    Sys.time()) %>%
      paste("  ~",.)
    decrement_call_depth(ZP_env)
    remove_timestamp(ZP_env)
  }
  if(ZP_env$global_ZygosityPredictor_variable_debug_verbose==TRUE){
    message(to_print, timestamp)
  }
}

#' accesor for gene predictions printing detailed info about how a gene status 
#' was assigned
#' 
#' @param fp full prediction (output of predict_zygoisty())
#' @param inp_gene name of gene that should be printed with detailed information
#' @param n max number of rows to print, as some gene status depend on loads of
#' phasing results#' 
#' @return prints overview about run from function predict_zygoisty() with specific information about provided gene
#' @export
#' @importFrom rlang ensym
#' @examples
#' cnvs  = GenomicRanges::GRanges(
#'   dplyr::tibble(
#'     chr = "chr17",
#'     start = c(170060, 34520990),
#'     end = c(34520990, 83198614),
#'     tcn = c(2, 1),
#'     cna_type = c("neutral", "LOH")
#'   )
#' )
#' somatic_vars = GenomicRanges::GRanges(
#'   dplyr::tibble(
#'     chr="chr17",
#'     start = 7675088,
#'     end = 7675088,
#'     ref = "C",
#'     alt = "T",
#'     af = 0.65,
#'     gene = "TP53" 
#'   )
#' )
#' germline_vars = GenomicRanges::GRanges(
#'   dplyr::tibble(
#'     chr="chr17",
#'     start = 41771694,
#'     end = 41771694,
#'     ref = "GTGT",
#'     alt = "G",
#'     af = 0.95,
#'     gene = "JUP" 
#'   )
#' )
#' reference = GenomicRanges::GRanges(
#'   dplyr::tibble(
#'     chr = "chr17",
#'     start = c(7661778, 41754603),
#'     end = c(7687538, 41786931),
#'     gene = c("TP53", "JUP")
#'   )
#' )
#' sex = "female"
#' purity = 0.9
#' bamfile <- system.file("extdata", "ZP_example.bam", 
#'   package = "ZygosityPredictor")
#' fp <- predict_zygosity(purity = purity, sex = sex, 
#'   somCna = cnvs,
#'   somSmallVars = somatic_vars,
#'   germSmallVars = germline_vars,
#'   geneModel = reference,
#'   bamDna = bamfile
#' )
#' gene_ov(fp, TP53)
gene_ov <- function(fp, inp_gene, n=20){
  #inp_gene <- as.character(rlang::ensym(inp_gene))
  inp_gene <- as.character(ensym(inp_gene))
  eval_per_gene <- fp$eval_per_gene %>% filter(gene==inp_gene) 
  eval_per_var <- fp$eval_per_variant %>% filter(gene==inp_gene) %>%
    mutate_at(.vars=c("af", "tcn", "wt_cp", "aff_cp"),
              .funs=round, 2) %>%
    select(-vn_status, -pre_info, -purity)

  message("Top level: Gene status\n")
  message(print_tibble(eval_per_gene))
  message("\n\nSub level: Evaluation per variant\n")
  message(print_tibble(eval_per_var))
  
  if(!is.null(fp$phasing_info)){
   
    phasing_info <- fp$phasing_info %>% 
      filter(gene==inp_gene) %>%
      mutate_at(.vars=c("conf", "wt_cp", "min_poss_wt_cp", "max_poss_wt_cp"),
                .funs=round, 2) %>% 
      arrange(desc(nconst), desc(conf))
    if(!is.null(fp$detailed_RLP_info)){
      detailed_RLP <- fp$detailed_RLP_info %>% 
        filter(gene==inp_gene) %>%
        select(-xsq_diff, -xsq_same, -v_same, -v_diff, -phasing) %>%
        mutate_at(.vars=c("mut2", "mut1", "both", "p_same",    "p_diff"),
                  .funs=round, 2) %>%
        arrange(desc(nconst))   
      if(nrow(detailed_RLP)<n){
        n_RLP <- nrow(detailed_RLP)
      } else {
        n_RLP <- n
      }  
      RLP_mes <- paste(c("\n\nSub level: All read-level-phasing combinations, including SNPs\n", "Showing ", n_RLP, " of ", nrow(detailed_RLP), " phasing attempts\n\n",
          print_tibble(detailed_RLP[1:n_RLP,]), "\n"), collapse="")
    } else {
      RLP_mes <- ""
    }
    if(!is.null(fp$detailed_AIP_info)){
      detailed_AIP <- fp$detailed_AIP_info %>% 
        filter(gene==inp_gene) %>%
        #select(-conf, -xsq_diff, -xsq_same, -v_same, -v_diff, -phasing) %>%
        # mutate_at(.vars=c("p_same",    "p_diff"),
        #           .funs=round, 2) %>%
        arrange(desc(nconst))
  
      if(nrow(detailed_AIP)<n){
        n_AIP <- nrow(detailed_AIP)
      } else {
        n_AIP <- n
      }
      AIP_mes <- paste(c("\n\nSub level: All allelic-imbalance-phasing combinations, including SNPs\n", "Showing ", n_AIP, " of ", nrow(detailed_AIP), " phasing attempts\n\n",
                         print_tibble(detailed_AIP[1:n_AIP,]), "\n"), collapse="")
    } else {
      AIP_mes <- ""
    }
    if(nrow(phasing_info)>0){
      message("\n\nSub level: Main phasing combinations\n")
      message(print_tibble(phasing_info))
      message(RLP_mes)
      message(AIP_mes)
      # message("\n\nSub level: All read-level-phasing combinations, including SNPs\n", "Showing ", n, " of ", nrow(detailed_RLP), " phasing attempts\n")
      # message(print_tibble(detailed_RLP[1:n,]))
      # message("\n\nSub level: All allelic-imbalance-phasing combinations, including SNPs\n", "Showing ", n, " of ", nrow(detailed_AIP), " phasing attempts\n")
      # message(print_tibble(detailed_AIP[1:n,]))
    }
  }

  
}
#' accesor for ZygoistyPredictor runs. Prints an overview about the run
#' 
#' @param fp full prediction (output of predict_zygoisty())
#' @return prints overview about run from function predict_zygoisty()
#' @export
#' @examples
#' cnvs  = GenomicRanges::GRanges(
#'   dplyr::tibble(
#'     chr = "chr17",
#'     start = c(170060, 34520990),
#'     end = c(34520990, 83198614),
#'     tcn = c(2, 1),
#'     cna_type = c("neutral", "LOH")
#'   )
#' )
#' somatic_vars = GenomicRanges::GRanges(
#'   dplyr::tibble(
#'     chr="chr17",
#'     start = 7675088,
#'     end = 7675088,
#'     ref = "C",
#'     alt = "T",
#'     af = 0.65,
#'     gene = "TP53" 
#'   )
#' )
#' germline_vars = GenomicRanges::GRanges(
#'   dplyr::tibble(
#'     chr="chr17",
#'     start = 41771694,
#'     end = 41771694,
#'     ref = "GTGT",
#'     alt = "G",
#'     af = 0.95,
#'     gene = "JUP" 
#'   )
#' )
#' reference = GenomicRanges::GRanges(
#'   dplyr::tibble(
#'     chr = "chr17",
#'     start = c(7661778, 41754603),
#'     end = c(7687538, 41786931),
#'     gene = c("TP53", "JUP")
#'   )
#' )
#' sex = "female"
#' purity = 0.9
#' bamfile <- system.file("extdata", "ZP_example.bam", 
#'   package = "ZygosityPredictor")
#' fp <- predict_zygosity(purity = purity, sex = sex, 
#'   somCna = cnvs,
#'   somSmallVars = somatic_vars,
#'   germSmallVars = germline_vars,
#'   geneModel = reference,
#'   bamDna = bamfile
#' )
#' ZP_ov(fp)
ZP_ov <- function(fp){
  
  if(!is.null(fp$eval_per_variant)){
    n_suc_vars <- nrow(fp$eval_per_variant)
    if(!is.null(fp$uncovered_input)){
      n_uncovered_vars <- nrow(fp$uncovered_input)
    } else {
      n_uncovered_vars <- 0
    }
    n_vars <- n_suc_vars+n_uncovered_vars
    if(!is.null(fp$eval_per_gene)){
      n_pred_genes <- nrow(fp$eval_per_gene)
      overview_pred_genes <- fp$eval_per_gene %>%
        group_by(status, eval_by, .drop=FALSE) %>%
        tally() %>%
        print_tibble()
    } else {
      n_pred_genes <- 0
    }
    if(!is.null(fp$phasing_info)){
      phased_genes <- fp$phasing_info %>%
        filter(nconst>0) %>% pull(gene) %>%
      
        unique() %>% sort %>% paste(collapse = ", ")
    } else {
      phased_genes <- "none"
    }
    message("Zygosity Prediction of ", n_vars, " input variants")
    message(n_uncovered_vars, " are uncovered by sCNA input and were not evaluated")
    message(n_pred_genes, " genes analyzed:")
    message(overview_pred_genes)
    message("phased genes:\n", phased_genes)
  } else {
    message("No input variants provided to ZygosityPredictor")
  }
  
}

#' calculates how many copies are affected by a germnline small variant
#'
#' @param af Allele-frequency of the variant (numeric value between 0 and 1)
#' @param tcn total-copynumber at position of the variant (numeric value >0)
#' @param purity purity of the sample (numeric value between 0 and 1 indicating 
#' the fraction of relevant sample with control/unrelevant tissue)
#' @param chr chromosome of the variant (either format 1,2,..,X,Y or 
#' chr1,..,chrX)
#' @param sex sex of the sample (character: "male", "female", "m", "f")
#' @param c_normal expected copy number at position of the variant in normal 
#' tissue, 1 for gonosomes in male samples, and 2 for male autosomes and all 
#' chromosomes in female samples. (The function can also assess the c_normal 
#' parameter by itself, but then the following two inputs must be provided: 
#' chr and sex)
#' @param af_normal Allele-frequency in normal tissue (numeric value between 0 
#' and 1) 0.5 represents heterozygous variants in diploid genome, 
#' 1 would be homozygous. Could be relevant if germline CNVs are present at the 
#' position. Then also the c_normal parameter would have to be adjusted.
#' @return A numeric value indicating the affecting copies for the variant
#' @export
#' @examples 
#' library(dplyr)
#' library(purrr)
#' library(stringr)
#' aff_germ_copies(af=0.67, tcn=2, purity=0.9, chr="chrX", sex="female")
aff_germ_copies <- function(chr, af, tcn, purity, sex, 
                            c_normal=NULL, af_normal=0.5){
  cv <- formula_checks(chr, af, tcn, purity, sex, c_normal, af_normal)
  aff_cp <- cv$af*cv$tcn+(cv$af-cv$af_normal)*cv$c_normal*((1/cv$purity)-1)
  return(aff_cp) 
  # alternative: (af*(purity*tcn+c_normal*(1-purity))-cc*(1-purity))/purity 
}
#' calculates how many copies are affected by a somatic small variant
#'
#' @param af Allele-frequency of the variant (numeric value between 0 and 1)
#' @param tcn total-copynumber at position of the variant (numeric value >0)
#' @param purity purity of the sample (numeric value between 0 and 1 indicating 
#' the fraction of relevant sample with control/unrelevant tissue)
#' @param chr chromosome of the variant (either format 1,2,..,X,Y or 
#' chr1,..,chrX)
#' @param sex sex of the sample (character: "male", "female", "m", "f")
#' @param c_normal expected copy number at the position of the variant in 
#' normal tissue, 1 for gonosomes in male samples, and 2 for male autosomes and 
#' all chromosomes in female samples. (The function can also assess the 
#' c_normal parameter by itself, but then the following two inputs must be 
#' provided: chr and sex)
#' @return A numeric value indicating the affecting copies for the variant
#' @examples 
#' library(dplyr)
#' library(purrr)
#' library(stringr)
#' aff_som_copies(chr="chrX", af=0.67, tcn=2, purity=0.9, sex="female")
#' @export
aff_som_copies <- function(chr, af, tcn, purity, sex, c_normal=NULL){
  cv <- formula_checks(chr, af, tcn, purity, sex, c_normal)
  aff_cp <- cv$af*(cv$tcn+cv$c_normal*((1/cv$purity)-1))
  return(aff_cp)
  ## alternative: aff_cp = af*(tcn+c_normal/purity-c_normal)
}
#' predicts zygosity of a set of variants
#' @param purity purity of the sample (numeric value between 0 and 1 indicating 
#' the fraction of relevant sample with control/unrelevant tissue)
#' @param ploidy ploidy of the sample (numeric value)
#' @param sex sex of the sample (character: "male", "female", "m", "f")
#' @param somCna GRanges object containing all genomic regions with annotated 
#' total copynumber and cna_type as metadata columns. The total-copynumber 
#' column should be named "tcn" but also some other commonly used names. 
#' It should contain numeric values or characters that can be converted to 
#' numeric values. The cna_type column must contain the information about 
#' loss of heterozygosity (LOH). Therefore the term "LOH" must be explicitely 
#' mentioned in the column. If a genomic region is not present in the object, 
#' it will be taken as heterozygous with neutral TCN of 2. 
#' @param somSmallVars GRanges object containing all somatic small 
#' variants (SNV and INDEL).
#' Required metadata columns are reference base (ref/REF), 
#' alternative base (alt/ALT),
#' annotation of the gene name (gene/GENE) and the allele-frequency (af/AF). 
#' If the object is not provided the tool assumes there are no somatic small 
#' variants.
#' @param germSmallVars GRanges object containing all germline small 
#' variants (SNV and INDEL).
#' Required metadata columns are reference base (ref/REF), alternative 
#' base (alt/ALT),
#' annotation of the gene name (gene/GENE) and the allele-frequency (af/AF)
#' If the object is not provided the tool assumes there are no germline small 
#' variants.
#' @param geneModel GRanges object containing the gene-annoattion of 
#' the used reference genome with metadata column of the gene name (gene)
#' @param includeHomoDel default = TRUE; if FALSE homozygous deleteions are 
#' excluded
#' @param includeIncompleteDel default = TRUE; if FALSE heterzygous deleteions 
#' are excluded
#' @param assumeSomCnaGaps (logical, default=FALSE) Only required if the somCna
#' object lacks copy number information for genomic segments on which small 
#' variants are detected. By default, variants in such regions will be excluded 
#' from the analysis as required information about the copy number is missing. 
#' These variants will be attached to the final output list in a separate 
#' tibble. To include them, this flag must be set TRUE and the ground ploidy 
#' must be given as an input. This ground ploidy will then be taken as tcn in 
#' the missing regions. If no ploidy is given the tool will assume the ground 
#' ploidy of 2 when this flag is TRUE.
#' @param byTcn logical, default=TRUE; optional if includeHomoDel or 
#' includeIncompleteDelS is TRUE. If FALSE the tool will not use tcn as a 
#' criterion to assign large deletions. It will use the cna_type column and 
#' check for indicating strings like HOMDEL/HomoDel/DEL. Some commonly used 
#' strings are covered. It is recommended to leave this flag TRUE
#' @param colnameTcn character indicating the name of the metadata containing 
#' the tcn information in the somCna object. If not provided the tool tries to 
#' detect the column according to default names
#' @param colnameCnaType character indicating the name of the metadata 
#' containing cna type information in the somCna object. 
#' If not provided the tool tries to detect the column according to default 
#' names
#' @param ZP_env internal variable... not recommened to be changed by user
#' @param verbose logical, default=FALSE; prints functions that are called
#' @return A list containing tibbles with all input variants
#' @importFrom IRanges subsetByOverlaps
#' @importFrom GenomicRanges seqnames
#' @export
#' @examples 
#' cnvs  = GenomicRanges::GRanges(
#'   dplyr::tibble(
#'     chr = "chr17",
#'     start = c(170060, 34520990),
#'     end = c(34520990, 83198614),
#'     tcn = c(2, 1),
#'     cna_type = c("neutral", "LOH")
#'   )
#' )
#' somatic_vars = GenomicRanges::GRanges(
#'   dplyr::tibble(
#'     chr="chr17",
#'     start = 7675088,
#'     end = 7675088,
#'     ref = "C",
#'     alt = "T",
#'     af = 0.65,
#'     gene = "TP53" 
#'   )
#' )
#' germline_vars = GenomicRanges::GRanges(
#'   dplyr::tibble(
#'     chr="chr17",
#'     start = 41771694,
#'     end = 41771694,
#'     ref = "GTGT",
#'     alt = "G",
#'     af = 0.95,
#'     gene = "JUP" 
#'   )
#' )
#' reference = GenomicRanges::GRanges(
#'   dplyr::tibble(
#'     chr = "chr17",
#'     start = c(7661778, 41754603),
#'     end = c(7687538, 41786931),
#'     gene = c("TP53", "JUP")
#'   )
#' )
#' sex = "female"
#' purity = 0.9
#' predict_per_variant(purity = purity, sex = sex, 
#'   somCna = cnvs,
#'   somSmallVars = somatic_vars,
#'   germSmallVars = germline_vars,
#'   geneModel = reference
#' )
predict_per_variant <- function(purity, 
                                sex,
                                somCna, 
                                geneModel=NULL,
                                somSmallVars=NULL, 
                                germSmallVars=NULL,
                                ploidy=NULL, 
                                colnameTcn=NULL,
                                colnameCnaType=NULL,
                                includeHomoDel=TRUE,
                                includeIncompleteDel=TRUE,
                                assumeSomCnaGaps=FALSE,
                                byTcn=TRUE,
                                ZP_env=NULL,
                                verbose=FALSE
){
  status <- info <- wt_cp <- . <- df_homdels <- df_all_mutations <- gene <-
    final_phasing_info <- combined_read_details <-  final_output <-
    uncovered_som <- uncovered_germ <- gr_germ_cov <- gr_som_cov <- 
    som_covered <- germ_covered <- final_phasing_info <- 
    combined_snp_phasing <- NULL
  if(is.null(ZP_env)){
    ZP_env <- new.env()
    is_pre_eval <- TRUE
    set_global_variables(FALSE, verbose, FALSE, ZP_env)
    somCna <- check_somCna(somCna, geneModel, sex, ploidy, 
                           assumeSomCnaGaps, colnameTcn, 
                           colnameCnaType, ZP_env)  
    purity <- check_purity(purity)
    sex <- check_sex(sex)
    somSmallVars <- check_gr_small_vars(somSmallVars, "somatic", ZP_env)
    germSmallVars <- check_gr_small_vars(germSmallVars, "germline", ZP_env)
    ploidy <- check_ploidy(ploidy)
    geneModel <- check_gr_gene_model(geneModel, ZP_env, is_pre_eval)
  } else {
    is_pre_eval <- FALSE
  }
  ## for parent function this must be here after gobal variable edfinition
  #print("xes i am here")
  func_start(ZP_env)
  ## check input for valid format    
  includeIncompleteDel <- check_opt_incdel(includeIncompleteDel, ploidy)
  includeHomoDel <- check_opt_incdel(includeHomoDel, ploidy)
  assumeSomCnaGaps <- check_opt_assgap(assumeSomCnaGaps, ploidy)
  if(is.null(geneModel)){
    ## only small variants can be evaulated
    if(includeIncompleteDel==TRUE|includeHomoDel==TRUE){
      includeIncompleteDel <- FALSE
      includeHomoDel <- FALSE
      warning("To include large deletions the geneModel must be provided.",
              "IncludeHomoDel and IncludeIncompleteDel will be FALSE")
    }
    templateGenes <- c(somSmallVars$gene, germSmallVars$gene)
  } else {
    templateGenes <- geneModel$gene
  }
  ## get variants not covered by CNV input
  if(!is.null(somSmallVars)){
    if(length(intersect(levels(seqnames(somCna)), levels(seqnames(somSmallVars))))>0){
      gr_som_cov <- subsetByOverlaps(somSmallVars, somCna) 
      som_covered <- gr_som_cov$mid     
    } else {
      stop("objects: somCna and somSmallVars have no seqnames in common... this ",
           "might be due to different reference genomes used for the objects or ",
           "simply deviating annotation formats of chromosomes: chr1...chrX or 1...X")
    }
 
  }
  if(!is.null(germSmallVars)){
    if(length(intersect(levels(seqnames(somCna)), levels(seqnames(germSmallVars))))>0){
    gr_germ_cov <- subsetByOverlaps(germSmallVars, somCna) 
    germ_covered <- gr_germ_cov$mid
    } else {
      stop("objects: somCna and germSmallVars have no seqnames in common... this ",
           "might be due to different reference genomes used for the objects or ",
           "simply deviating annotation formats of chromosomes: chr1...chrX or 1...X")
    }
  }
  combined_uncovered <- combine_uncovered_input_variants(somSmallVars, 
                                                         germSmallVars,
                                                         som_covered,
                                                         germ_covered,
                                                         templateGenes, 
                                                         ZP_env)
  ## preapre input data for prediction (pre-evaluation)
  df_germ <- prepare_germline_variants(
    gr_germ_cov, 
    somCna, purity, sex,
    ZP_env)
  df_som <- prepare_somatic_variant_table(
    gr_som_cov, 
    templateGenes, 
    somCna, purity, sex,
    ZP_env)
  df_homdels <- extract_all_dels_of_sample(somCna, geneModel, 
                                           "homdel", byTcn, sex, ploidy,
                                           includeHomoDel, ZP_env)
  df_incompletedels <- extract_all_dels_of_sample(somCna, geneModel, 
                                                  "incompletedel", TRUE, sex, 
                                                  ploidy, includeIncompleteDel,
                                                  ZP_env)
  
  #print(df_germ)
  #print(df_som)
  #print()
  
  ## predict zygosity for each gene
  if(!is.null(df_germ)|!is.null(df_som)|!is.null(df_homdels)){
    #print(1)
    df_all_mutations <- combine_main_variant_tables(df_germ, df_som, 
                                                    df_homdels,
                                                    templateGenes, purity,
                                                    ZP_env)
    #print(2)
  }
  if(is_pre_eval){
    
    full_eval <- list(evaluation_per_variant= 
                  bind_incdel_to_pre_eval(df_incompletedels, df_all_mutations,
                                          ZP_env),
                combined_uncovered=combined_uncovered)
  } else {
    full_eval <- list(evaluation_per_variant=df_all_mutations,
                df_incompletedels=df_incompletedels,
                combined_uncovered=combined_uncovered)   
  }
  func_end(ZP_env)
  return(full_eval)
}
#' predicts zygosity of a set of genes of a sample
#' @param purity purity of the sample (numeric value between 0 and 1 indicating 
#' the fraction of relevant sample with control/unrelevant tissue)
#' @param ploidy ploidy of the sample (numeric value)
#' @param sex sex of the sample (character: "male", "female", "m", "f")
#' @param somCna GRanges object containing all genomic regions with annotated 
#' total copynumber and cna_type as metadata columns. The total-copynumber 
#' column should be named "tcn" but also some other commonly used names. 
#' It should contain numeric values or characters that can be converted to 
#' numeric values. The cna_type column must contain the information about 
#' loss of heterozygosity (LOH). Therefore the term "LOH" must be explicitely 
#' mentioned in the column. If a genomic region is not present in the object, 
#' it will be taken as heterozygous with neutral TCN of 2. 
#' @param somSmallVars GRanges object containing all somatic small 
#' variants (SNV and INDEL).
#' Required metadata columns are reference base (ref/REF), 
#' alternative base (alt/ALT),
#' annotation of the gene name (gene/GENE) and the allele-frequency (af/AF). 
#' If the object is not provided the tool assumes there are no somatic small 
#' variants.
#' @param germSmallVars GRanges object containing all germline small 
#' variants (SNV and INDEL).
#' Required metadata columns are reference base (ref/REF), alternative 
#' base (alt/ALT),
#' annotation of the gene name (gene/GENE) and the allele-frequency (af/AF)
#' If the object is not provided the tool assumes there are no germline small 
#' variants.
#' @param geneModel GRanges object containing the gene-annoattion of 
#' the used reference genome with metadata column of the gene name (gene)
#' @param bamDna path to bam-file
#' @param bamRna optional; path to rna file (bam format)
#' @param includeHomoDel default = TRUE; if FALSE homozygous deleteions are 
#' excluded
#' @param includeIncompleteDel default = TRUE; if FALSE heterzygous deleteions 
#' are excluded
#' @param AllelicImbalancePhasing logical, default=FALSE. Enables alleleic imbalance phasing if TRUE
#' @param printLog default = FALSE; if TRUE the gene which is evaluated is 
#' printed in console, 
#' containing the query-name of each read which was used to perform 
#' haplotype-phasing and the info into which class it was assigned.
#' @param assumeSomCnaGaps (logical, default=FALSE) Only required if the somCna
#' object lacks copy number information for genomic segments on which small 
#' variants are detected. By default, variants in such regions will be excluded 
#' from the analysis as required information about the copy number is missing. 
#' These variants will be attached to the final output list in a separate 
#' tibble. To include them, this flag must be set TRUE and the ground ploidy 
#' must be given as an input. This ground ploidy will then be taken as tcn in 
#' the missing regions. If no ploidy is given the tool will assume the ground 
#' ploidy of 2 when this flag is TRUE.
#' @param byTcn logical, default=TRUE; optional if includeHomoDel or 
#' includeIncompleteDelS is TRUE. If FALSE the tool will not use tcn as a 
#' criterion to assign large deletions. It will use the cna_type column and 
#' check for indicating strings like HOMDEL/HomoDel/DEL. Some commonly used 
#' strings are covered. It is recommended to leave this flag TRUE
#' @param colnameTcn character indicating the name of the metadata containing 
#' the tcn information in the somCna object. If not provided the tool tries to 
#' detect the column according to default names
#' @param colnameCnaType character indicating the name of the metadata 
#' containing cna type information in the somCna object. 
#' If not provided the tool tries to detect the column according to default 
#' names
#' @param vcf character; path to variant call file (.vcf.gz format). 
#' Will be used (if provided)
#' for extended SNP phasing if variants on the same gene are too far away from
#' each other for direct haplotype phasing
#' @param distCutOff numeric, default=5000; if input vcf is provided and SNP
#' phasing is performed, this will limt the distance at which the SNP phasing
#' should not be tried anymore. As the probability of finding overlapping reads
#' at such a long distance is very low and the runtime will increase
#' exponentially.
#' @param debug logical, default=FALSE; prints output for debugging
#' @param verbose logical, default=FALSE; prints functions that are called
#' @param haploBlocks GRanges object containing haploblocks. Haploblocks are
#' defined as genomic regions in which SNPs are phased to a specific allele.
#' For example a haploblock could be chr1:1000-10000. This would mean that every
#' genotype annotation in the format "1|0" or "0|1" of a SNP in this region will 
#' be used to phase somatic variants and define their genotype
#' @param logDir character; path to directory where logfiles and detailed infos 
#' of the run can be stored, if not given, no details will be stored or printed     
#' @param snpQualityCutOff numeric, default=1; Cutoff to filter for SNPS that 
#' can be used for phasing                     
#' @param phasingMode character, default="fast"; if set to full. Even if high 
#' confidence phasing result could be achieved, following phasing approaches 
#' will be carried out
#' @param showReadDetail default = FALSE; if TRUE a table is added to the 
#' output, containing all used reads/rea-pairs with anntated read classification 
#' (mut1, mut2, both, none, skipped, dev_var)
#' 
#' 
#' @return A list of dataframes. Those are the evaluation per variant, 
#' the evaluation per gene and, if performed, the info about the 
#' haplotype-phasing.
#' @examples
#' cnvs  = GenomicRanges::GRanges(
#'   dplyr::tibble(
#'     chr = "chr17",
#'     start = c(170060, 34520990),
#'     end = c(34520990, 83198614),
#'     tcn = c(2, 1),
#'     cna_type = c("neutral", "LOH")
#'   )
#' )
#' somatic_vars = GenomicRanges::GRanges(
#'   dplyr::tibble(
#'     chr="chr17",
#'     start = 7675088,
#'     end = 7675088,
#'     ref = "C",
#'     alt = "T",
#'     af = 0.65,
#'     gene = "TP53" 
#'   )
#' )
#' germline_vars = GenomicRanges::GRanges(
#'   dplyr::tibble(
#'     chr="chr17",
#'     start = 41771694,
#'     end = 41771694,
#'     ref = "GTGT",
#'     alt = "G",
#'     af = 0.95,
#'     gene = "JUP" 
#'   )
#' )
#' reference = GenomicRanges::GRanges(
#'   dplyr::tibble(
#'     chr = "chr17",
#'     start = c(7661778, 41754603),
#'     end = c(7687538, 41786931),
#'     gene = c("TP53", "JUP")
#'   )
#' )
#' sex = "female"
#' purity = 0.9
#' bamfile <- system.file("extdata", "ZP_example.bam", 
#'   package = "ZygosityPredictor")
#' predict_zygosity(purity = purity, sex = sex, 
#'   somCna = cnvs,
#'   somSmallVars = somatic_vars,
#'   germSmallVars = germline_vars,
#'   geneModel = reference,
#'   bamDna = bamfile
#' )
#' @importFrom stringr %>%
#' @importFrom IRanges subsetByOverlaps
#' @importFrom purrr compact
#' @importFrom dplyr bind_rows nth select tibble
#' @export
predict_zygosity <- function(purity, 
                             sex,
                             somCna, 
                             geneModel,
                             bamDna,
                             somSmallVars=NULL, 
                             germSmallVars=NULL,
                             bamRna=NULL,
                             ploidy=NULL, 
                             colnameTcn=NULL,
                             colnameCnaType=NULL,
                             includeHomoDel=TRUE,
                             includeIncompleteDel=TRUE,
                             showReadDetail=FALSE,
                             printLog=FALSE,
                             assumeSomCnaGaps=FALSE,
                             byTcn=TRUE,
                             vcf=NULL,
                             haploBlocks=NULL,
                             distCutOff=5000,
                             verbose=FALSE,
                             debug=FALSE,
                             logDir=NULL,
                             snpQualityCutOff=1, 
                             phasingMode="fast",
                             AllelicImbalancePhasing=FALSE
){
  status <- info <- wt_cp <- . <- df_homdels <- evaluation_per_variant <- 
    gene <- final_phasing_info <- combined_read_details <-  final_output <-
    uncovered_som <- uncovered_germ <- gr_germ_cov <- gr_som_cov <- 
    som_covered <- germ_covered <- final_phasing_info <- detailed_RLP_info <- 
    detailed_AIP_info <- combined_eval_per_gene <- full_phasing_info <- 
    combined_snp_phasing <- evaluation_per_gene <- log_list_per_gene <- 
    detailed_phasing_info <- comb_mat_phased <- comb_mat_info <- NULL
  ## define global debugging variable
  ZP_env <- new.env()
  set_global_variables(debug, verbose, printLog, ZP_env)
  ## for parent function this must be here after gobal variable edfinition
  func_start(ZP_env)
  ploidy <- check_ploidy(ploidy)
  #print(ZP_env$test)
  sex <- check_sex(sex)
  #print(ZP_env$test)
  assumeSomCnaGaps <- check_opt_assgap(assumeSomCnaGaps, ploidy)
  geneModel <- check_gr_gene_model(geneModel, ZP_env)
  somCna <- check_somCna(somCna, geneModel, sex, ploidy, 
                          assumeSomCnaGaps, colnameTcn, 
                          colnameCnaType, ZP_env)
  purity <- check_purity(purity)
  somSmallVars <- check_gr_small_vars(somSmallVars, "somatic", ZP_env)
  germSmallVars <- check_gr_small_vars(germSmallVars, "germline", ZP_env)
  
  evaluation_per_variant_pre <- predict_per_variant(purity, 
                                                    sex,
                                                    somCna, 
                                                    geneModel,
                                                    somSmallVars, 
                                                    germSmallVars,
                                                    ploidy, 
                                                    colnameTcn,
                                                    colnameCnaType,
                                                    includeHomoDel,
                                                    includeIncompleteDel,
                                                    assumeSomCnaGaps,
                                                    byTcn, 
                                                    ZP_env)
  evaluation_per_variant <- evaluation_per_variant_pre$evaluation_per_variant
  if(!is.null(evaluation_per_variant)){
    if(!is.null(geneModel)){
      if(!nrow(evaluation_per_variant)==0){
        bamDna <- check_bam(bamDna)
        bamRna <- check_rna(bamRna)
        logDir <- check_logDir(logDir)
        haploBlocks <- check_haploblocks(haploBlocks, ZP_env)
        vcf <- check_vcf(vcf)
        per_gene <- lapply(
          unique(evaluation_per_variant$gene), 
          predict_zygosity_genewise, 
          evaluation_per_variant, 
          bamDna,
          bamRna,
          showReadDetail,
          printLog,
          purity,
          sex,
          haploBlocks,
          vcf,
          distCutOff, 
          logDir,
          somCna,
          snpQualityCutOff, 
          phasingMode,
          AllelicImbalancePhasing,
          ZP_env)
        full_eval_per_gene <- lapply(per_gene, nth, n=2) %>% compact()
        log_list_per_gene <- lapply(per_gene, nth, n=3)
        if(length(full_eval_per_gene)!=0){
          full_phasing_info <- lapply(full_eval_per_gene, nth, n=2) %>% 
            compact() %>% 
            bind_rows() 
          detailed_RLP_info <- lapply(full_eval_per_gene, nth, n=3) %>% 
            compact() %>% 
            bind_rows() 
          detailed_AIP_info <- lapply(full_eval_per_gene, nth, n=6) %>% 
            compact() %>% 
            bind_rows() 
          comb_mat_phased <- lapply(full_eval_per_gene, nth, n=4) %>%
            compact() %>%
            Reduce(function(x,y)append(x,y),.)
          comb_mat_info <- lapply(full_eval_per_gene, nth, n=5) %>%
            compact() %>%
            Reduce(function(x,y)append(x,y),.)
          combined_eval_per_gene <- lapply(full_eval_per_gene, nth, n=1) %>% 
            bind_rows() %>%
            select(gene, n_mut, status, conf, eval_by, wt_cp, 
                   warning, wt_cp_range, info,  phasing, eval_time_s)
        } 
      }    
      evaluation_per_gene <- bind_incdel_to_final_eval(
        evaluation_per_variant_pre$df_incompletedels,
        combined_eval_per_gene, ZP_env) 
    }
  }
  result_list <- list(
    eval_per_variant=evaluation_per_variant,
    eval_per_gene=evaluation_per_gene,
    phasing_info=full_phasing_info,
    uncovered_input=evaluation_per_variant_pre$combined_uncovered,
    log_list_per_gene=log_list_per_gene,
    detailed_RLP_info=detailed_RLP_info,
    detailed_AIP_info=detailed_AIP_info,
    mat_phased=comb_mat_phased,
    mat_info=comb_mat_info
  ) %>%
    compact()
  #remove_global_vars(ZP_env)
  func_end(ZP_env)
  return(result_list)
} 

#' @keywords internal
#' description follows
#' @importFrom stringr str_match
#' @importFrom dplyr case_when
#' @importFrom GenomicRanges elementMetadata elementMetadata<-
assign_correct_colnames <- function(obj, type, ZP_env){
  func_start(ZP_env)
  #vm(as.character(sys.call()[1]), 1)
  . <- NULL
  if(type=="scna"){
    
    ## define correct tcn column
    col_tcn <- str_match(
      nm_md(obj), paste(allowed_inputs("colnames_tcn", regex=TRUE), collapse="|")) %>%
      .[which(!is.na(.))]
    elementMetadata(obj)[,"tcn"] <- 
      as.numeric(elementMetadata(obj)[,col_tcn]) 
    
    
    
    ## define correct cna_type column
    if("LOH" %in% nm_md(obj)){
      obj$cna_type <- case_when(obj$LOH==TRUE ~ "LOH",
                                TRUE ~ "HZ")
    } else {
      col_cna_type <- str_match(
        nm_md(obj), paste(allowed_inputs("colnames_cna_type", regex=TRUE), 
                          collapse="|")) %>%
        .[which(!is.na(.))]   
      elementMetadata(obj)[,"cna_type"] <- 
        elementMetadata(obj)[,col_cna_type]
      obj$LOH <- case_when(str_detect(obj$cna_type, "LOH") ~ TRUE,
                           TRUE ~ FALSE)
    }
    ## define allelic imbalance column (optional input)
    ## check if column provided in input
    col_all_imb <- str_match(
      nm_md(obj), paste(allowed_inputs("colnames_all_imb", regex=TRUE), 
                        collapse="|")) %>%
      .[which(!is.na(.))] 
    if(length(col_all_imb)>0){
    ## if colum already there, rename it
      elementMetadata(obj)[,"all_imb"] <- 
        elementMetadata(obj)[,col_all_imb]
    } else {
    ## if not yet there, calculate it for segments where it is possible
      obj$all_imb <- case_when(
        obj$LOH==FALSE&round(as.numeric(obj$tcn))%%2!=0&as.numeric(obj$tcn)>=2.5 ~ TRUE,
        TRUE ~ FALSE
      )
      #obj$gt_cna <- "1:2"
    }
    ## if allelic imbalance was detected somewhere, also the genotype needs to be annotated 
    ## except if tcn ==3
    col_genotype <- str_match(
      nm_md(obj), paste(allowed_inputs("colnames_genotype", regex=TRUE), 
                        collapse="|")) %>%
      .[which(!is.na(.))]
    if(length(col_genotype)>0){
      ## if colum already there, rename it
      ## and check if it has only entries that are actually imbalanced... remove
      ## something like 2:2 or 3:3
      input_gt_col <- as.character(elementMetadata(obj)[,col_genotype]) %>%
        lapply(.,function(GT){
          uniqued_gts <- str_split(GT, ":") %>% unlist() %>% unique()
          if(length(uniqued_gts)==1){
            warning("column gt_cna contains annotations of balanced segments. They are removed from imbalance phasing if enabled")
            return(NA)
          } else {
            return(GT)
          }
        }) %>%
        unlist()
      elementMetadata(obj)[,"gt_cna"] <- input_gt_col
    } else {
      ## if not yet there, calculate it for segments where it is possible
      obj$gt_cna <- case_when(
        round(obj$tcn)==3 ~ "1:2",
        TRUE ~ NA
      )
    }
  } else {
    col_gene <- str_match(
      nm_md(obj), 
      paste(allowed_inputs("colnames_gene", regex=TRUE), collapse="|")) %>%
      .[which(!is.na(.))]
    elementMetadata(obj)[,"gene"] <- 
      elementMetadata(obj)[,col_gene]  
    if(type=="small_vars"){
      col_af <- str_match(
        nm_md(obj), paste(allowed_inputs("colnames_af", regex=TRUE), collapse="|")) %>%
        .[which(!is.na(.))]
      col_ref <- str_match(
        nm_md(obj), paste(allowed_inputs("colnames_ref", regex=TRUE), collapse="|")) %>%
        .[which(!is.na(.))]
      col_alt <- str_match(
        nm_md(obj), paste(allowed_inputs("colnames_alt", regex=TRUE), collapse="|")) %>%
        .[which(!is.na(.))]
      elementMetadata(obj)[,"af"] <- 
        elementMetadata(obj)[,col_af]
      elementMetadata(obj)[,"ref"] <- 
        elementMetadata(obj)[,col_ref]
      elementMetadata(obj)[,"alt"] <- 
        elementMetadata(obj)[,col_alt]
      elementMetadata(obj)["mid"] <- 
        seq_len(length(obj))    
    }    
  }
  func_end(ZP_env)
  return(obj)
}
#' @keywords internal
#' no dependencies
formula_checks <- function(chr, af, tcn, purity, sex, c_normal, af_normal=0.5){
  purity <- check_purity(purity)
  af <- check_af(af)
  af_normal <- check_af(af_normal)
  tcn <- check_tcn(tcn)
  if(is.null(c_normal)){
    sex <- check_sex(sex)
    chr <- check_chr(chr)
    if((sex=="male"&(chr=="X"|chr=="chrX"))|(chr=="Y"|chr=="chrY")){
      c_normal <- 1
      af_normal <- 1
    } else {
      c_normal <- 2
      af_normal <- 0.5
    }
  } else {
    c_normal <- check_ploidy(c_normal)
  }
  return(list(af=af, tcn=tcn, purity=purity, c_normal=c_normal, 
              af_normal=af_normal))
}
#' @keywords internal
check_opt_assgap <- function(assumeSomCnaGaps, ploidy){
  if(assumeSomCnaGaps==TRUE&is.null(ploidy)){
    warning("somatic CNA gaps can only be assumed if input ploidy is",
            "provided. Provide ploidy=2 to assume diploid case")
    return(FALSE)
  } else {
    return(assumeSomCnaGaps)
  }
}
#' @keywords internal
check_opt_incdel <- function(includeIncompleteDel, ploidy){
  if(includeIncompleteDel==TRUE&is.null(ploidy)){
    warning("Large scale deletions cannot be included without ploidy",
            "Please provide input ploidy. Provide ploidy=2",
            "to assume diploid case")
    return(FALSE)
  } else {
    return(includeIncompleteDel)
  }
}
#' @keywords internal
#' returns names of metadata columns from GRanges object
#' @importFrom GenomicRanges elementMetadata
nm_md <- function(obj){
  return(names(elementMetadata(obj)))
}
#' @keywords internal
#' description follows
#' @importFrom dplyr between
check_af <- function(af){
  num_af <- as.numeric(af)
  if(is.na(num_af)){
    stop("input allele frequency (af) is not numeric")
  } else if(!between(num_af, 0, 1)){
    stop("input allele frequency (af) must be between 0 and 1")
  } else {
    return(num_af)
  }
}
#' @keywords internal
#' description follows
check_tcn <- function(tcn){
  num_tcn <- as.numeric(tcn)
  if(is.na(num_tcn)){
    stop("input total copynumber (tcn) is not numeric")
  } else {
    return(num_tcn)
  }
}
#' @keywords internal
check_chr <- function(chr){
  if(chr %in% allowed_inputs("chrom_names")){
    return(as.character(chr))
  } else {
    stop("input chromosome (chr) must be either a number between ",
         "1 and 23 or X or Y, or in the chr1 format")
  }
}
#' @keywords internal
#' description follows
#' @importFrom stringr %>% str_detect
#' @importFrom GenomicRanges elementMetadata elementMetadata<- seqnames
#' @importFrom methods is
check_somCna <- function(somCna, geneModel, sex, ploidy,
                         assumeSomCnaGaps, colnameTcn, 
                         colnameCnaType, ZP_env){
  func_start(ZP_env)
  . <- NULL
  ## geneModel can be null if predict_per_variant function is used alone
  if(!is.null(geneModel)){
    if(length(intersect(levels(seqnames(somCna)), levels(seqnames(geneModel))))==0){    
      
      stop("objects: somCna and geneModel have no seqnames in common... this ",
           "might be due to different reference genomes used for the objects or ",
           "simply deviating annotation formats of chromosomes: chr1...chrX or 1...X")
    }    
  }

  somCna <- general_gr_checks(somCna, "scna", "somCna", ZP_env)
  somCna$tcn_assumed <- FALSE
  ## check if there are segments without cna_type annotation
  if(sum(is.na(elementMetadata(somCna)[,"cna_type"]))>0){
    warning("cna_type column of input somCna contains", 
            sum(is.na(elementMetadata(somCna)[,"cna_type"])),
            "NA values;\n  they will be taken as hetero-zygous",
            "/hemizygous for gonosomes in male samples\n")
    elementMetadata(somCna)[,"cna_type"][
      which(
        is.na(elementMetadata(somCna)[,"cna_type"]))] <- NA
  }
  ## for male samples the gonosomes are hemizygous, we treat them like autosomal
  ## region with LOH, therefore we annotate LOH to the cna_type column
  if(sex=="male"&str_detect(
    paste(as.character(seqnames(somCna)), collapse=" "), "X|Y")){
    ## if true, the sample is male and has Gonosomal regions
    somCna[which(
      as.character(seqnames(somCna)) %in% c("X", "Y"))]$cna_type <-
      paste0(
        somCna[which(
          as.character(seqnames(somCna)) %in% c("X", "Y"))]$cna_type,
        ";LOH")
  }
  ## if CNA gaps should be assumed to ploidy
  if(assumeSomCnaGaps==TRUE){
    if(sum(is.na(elementMetadata(somCna)[,"tcn"]))>0){
      warning("tcn column of input somCna contains", 
              sum(
                is.na(elementMetadata(somCna)[,"tcn"])),
              "NA values;\n  they will be taken as ground ploidy:") 
      elementMetadata(somCna)[,"tcn_assumed"][
        which(is.na(elementMetadata(somCna)[,"tcn"]))] <- TRUE
      elementMetadata(somCna)[,"tcn"][
        which(is.na(elementMetadata(somCna)[,"tcn"]))] <- 
        ploidy
    }
    new_somCna <- insert_missing_cnv_regions(somCna, geneModel, 
                                             sex, ploidy, ZP_env)
  } else {
  ## if cna gaps should not be assumed, missing tcn segments will be excluded
    if(sum(is.na(elementMetadata(somCna)[,"tcn"]))>0){
      warning("tcn column of input somCna contains", 
              sum(
                is.na(elementMetadata(somCna)[,"tcn"])),
              "NA values;\n  they will be excluded from the analysis.",
              "Use assumeSomCnaGaps=TRUE to take ground ploidy",
              "(ploidy) as tcn"
      )
      new_somCna <- somCna[which(!is.na(
        elementMetadata(somCna)[,"tcn"]))]
    } else {
      new_somCna <- somCna
    }
  }
  new_somCna$seg_id <- seq(1, length(new_somCna))
  func_end(ZP_env)
  return(new_somCna)
}
#' @keywords internal
#' description follows
check_name_presence <- function(obj, type, ZP_env){
  func_start(ZP_env)
  if(type=="scna"){
    if(
      any(allowed_inputs("colnames_tcn") %in% nm_md(obj))&
      (any(allowed_inputs("colnames_cna_type") %in% nm_md(obj))|
       "LOH" %in% nm_md(obj))
    ){ 
      res <- TRUE
    } else {
      res <- FALSE
    } 
  } else {
    if(
      (any(allowed_inputs("colnames_gene") %in% nm_md(obj))&
        (type=="gene_model"|
         (
           any(allowed_inputs("colnames_af") %in% nm_md(obj))&
           any(allowed_inputs("colnames_ref") %in% nm_md(obj))&
           any(allowed_inputs("colnames_alt") %in% nm_md(obj))
         )
        )
      )
    ){
      res <- TRUE
    } else {
      res <- FALSE
    }    
  }
  func_end(ZP_env)
  return(res)
}
check_logDir <- function(logDir){
  if(!is.null(logDir)){
    if(file.exists(logDir)){
      return(logDir)
    } else {
      warning("logDir does not exist, no log will be stored")
      return(NULL)
    }
  } else {
    return(NULL)
  }
}
general_gr_checks <- function(obj, type, lab, ZP_env){
  func_start(ZP_env)
  if(!is(obj, "GRanges")){
    stop("input ", lab, " must be a GRanges object;",
         "given input appears to be:", 
         class(obj))
  } else if(type=="haploBlocks"){
  ## haploblocks do not need any metadata columns  
    res <- obj
  } else if(!check_name_presence(obj, type, ZP_env)){
    if(type=="scna"){
      stop(
        "input somCna requires the following metadata columns: ",
        "\'tcn\' and \'cna_type\'")
    } else if(type=="small_vars"){
      stop("input ", lab, " requires the following metadata columns:",
           "\'gene\'/\'GENE\', \'ref\'/\'REF\',",
           " \'alt\'/\'ALT\' and \'af\'/\'AF\'")      
    } else {
      stop(
        "input geneModel requires the following metadata columns: ",
        "\'gene\'/\'GENE\'")
    }
  } else {
    res <- assign_correct_colnames(obj, type, ZP_env)
  }
  func_end(ZP_env)
  return(res)
}
#' @keywords internal
#' description follows
check_gr_gene_model <- function(geneModel, ZP_env, is_pre_eval=FALSE){
  func_start(ZP_env)
  . <- warn <-res <- NULL

  if(is.null(geneModel)){
    warn <- "input geneModel is NULL -> Doing nothing"
  } else {
    res <- general_gr_checks(geneModel, "gene_model", "geneModel", ZP_env)
  }
  if(length(res)==0){
    warn <- "input gene model contains no gene ranges -> Doing nothing"
    res <- NULL
  }
  if(is_pre_eval==FALSE){
    warning(warn)
  }
  func_end(ZP_env)
  return(res)
}
#' @keywords internal
#' description follows
check_gr_small_vars <- function(obj, origin, ZP_env){
  func_start(ZP_env)
  . <- NULL
  lab <- ifelse(origin=="somatic",
                "somSmallVars",
                "germSmallVars")
  if(is.null(obj)){
    warning("Input ", lab, " empty/does not contain variants. ",
            "Assuming there are no ", origin, " small variants")
    res <- NULL
  } else if(length(obj)==0){
    warning("Input ", lab, " empty/does not contain variants. ",
            "Assuming there are no ", origin, " small variants")
    res <- NULL
  } else {
    res <- general_gr_checks(obj, "small_vars", lab, ZP_env)
  }
  func_end(ZP_env)
  return(res)
}
#' @keywords internal
#' description follows 
#' @importFrom dplyr between
check_purity <- function(purity){
  if(is.na(as.numeric(purity))){
    stop("input purity must be numeric or a character that can",
         "be converted to numeric;\n  ", purity, 
         "can not be converted to numeric")
  } else if(!between(as.numeric(purity), 0, 1)){
    stop("input purity must be a numeric value between 0 and 1")
  } else {
    return(as.numeric(purity))
  }
}
#' @keywords internal
#' description follows
check_ploidy <- function(ploidy){
  #ZP_env$test <- "this is a test"
  #func_start(ZP_env)
  if(is.null(ploidy)){
    res <- NULL
  } else {
    if(is.na(as.numeric(ploidy))){
      stop("input ploidy/c_normal must be numeric or a character that can",
           "be converted to numeric;\n  ", ploidy, 
           "can not be converted to numeric")
    } else {
      res <- as.numeric(ploidy)
    }
  }
  #func_end(ZP_env)
  return(res)
}
#' @keywords internal
#' description follows
#' @importFrom stringr str_detect
check_sex <- function(sex){
  . <- NULL
  #ZP_env$test <- "bit this is a second test"
  allowed_sex <- allowed_inputs("sex_names")
  if(!sex %in% allowed_sex){
    allowed_sex_to_print <- paste(allowed_sex, collapse = "\', \'") %>% 
      paste0("\'", ., "\'")
    stop("input sex must be one of ", allowed_sex_to_print)
  } else {
    sex <- tolower(sex)
    if(str_detect(sex, "f")){
      return("female")
    } else {
      return("male")
    }
  }
}
#' @keywords internal
#' description follows 
check_bam <- function(bamDna){
  if(file.exists(bamDna)){
    return(bamDna)
  } else {
    stop("input bamDna does not exist")
  }
}
#' @keywords internal
#' description follows
#' @importFrom Rsamtools TabixFile
check_vcf <- function(vcf){
  if(is.null(vcf)){
    return(NULL)
  } else if(!sum(unlist(lapply(vcf, file.exists)))==length(vcf)){
    stop("one of input vcfs does not exist")
  } else {
    processed_vcf <- lapply(vcf, function(VCF){
      if(file.exists(paste0(VCF, ".tbi"))){
        return(Rsamtools::TabixFile(VCF))
      } else {
        return(VCF)
      }
    })
    return(processed_vcf)
  }
}
#' @keywords internal
#' description follows
check_rna <- function(bamRna){
  if(is.null(bamRna)){
    message("no RNA file provided: Analysis will be done without RNA reads")
    return(NULL)
  } else if(file.exists(bamRna)){
    return(bamRna)   
  } else {
    warning("input bamRna does not exist:", bamRna,
            "\nanalysis will be done without RNA reads\n")
    return(NULL)
  }
}
check_haploblocks <- function(haploBlocks, ZP_env){
  if(is.null(haploBlocks)){
    return(NULL)
  } else {
    haploBlocks <- general_gr_checks(haploBlocks, "haploBlocks", "haploBlocks", 
                                     ZP_env)
    haploBlocks$hap_id <- seq(1, length(haploBlocks))
    return(haploBlocks)    
  }
}
#' @keywords internal
allowed_inputs <- function(which_one, regex=FALSE){
  . <- NULL
  type_list <- list(
    cna_homdel_annotation = paste(c("HOMDEL", "HomoDel", "HomoDEL", "HOMODEL", 
                                    "HomDel", "homdel"),collapse = "|"),
    cna_incompletedel_annotation = c("DEL", "del", "Del"),
    chrom_names = c(
      c(seq_len(23), "X", "Y"),
      paste0("chr", c(seq_len(23), "X", "Y"))
    ),
    colnames_gene = c("GENE", "gene", "Gene", 
                      "Gene_name", "GENE_NAME", "gene_name"),
    colnames_af = c("AF", "af", "Af"),
    colnames_ref = c("REF", "ref", "Ref"),
    colnames_alt = c("ALT", "alt","Alt"),
    colnames_tcn = c("TCN", "tcn", "Tcn"),
    colnames_cna_type = c("cna_type", "CNA_type", "Cna_Type", "CNA_Type"),
    colnames_all_imb = c("allelic_imbalance", "all_imb", "AI"),
    colnames_genotype = c("gt_cna", "genotype", "GT", "gt"),
    sex_names = c("male", "m", "female", "f") %>% c(.,toupper(.))
  )
  res <- type_list[[which_one]]
  if(regex){
    res_regex <- paste0("^", res, "$")
    return(res_regex)
  } else {
    return(res)
  }
  
}

library(testthat)
library(ZygosityPredictor)
#library(GenomicRanges)
## prepare inputs for predict_zygosity() unit test

FILE_BAM <- "inst/extdata/ZP_example.bam"#
PURITY <- 0.98
PLOIDY <- 1.57
SEX <- "female"

GR_SCNA  = GenomicRanges::GRanges(
  dplyr::tibble(
    chr = "chr17",
    start = c(170060, 34520990),
    end = c(34520990, 83198614),
    tcn = c(2, 1),
    cna_type = c("neutral", "LOH")
  )
)
GR_SOM_SMALL_VARS = GenomicRanges::GRanges(
  dplyr::tibble(
    chr="chr17",
    start = 7675088,
    end = 7675088,
    ref = "C",
    alt = "T",
    af = 0.65,
    gene = "TP53"
  )
)
GR_GERM_SMALL_VARS = GenomicRanges::GRanges(
  dplyr::tibble(
    chr="chr17",
    start = 41771694,
    end = 41771694,
    ref = "GTGT",
    alt = "G",
    af = 0.95,
    gene = "JUP"
  )
)
GR_GENE_MODEL = GenomicRanges::GRanges(
  dplyr::tibble(
    chr = "chr17",
    start = c(7661778, 41754603),
    end = c(7687538, 41786931),
    gene = c("TP53", "JUP")
  )
)

# predict_zygosity(purity = purity, sex = sex,
#   somCna = cnvs,
#   somSmallVars = somatic_vars,
#   germSmallVars = germline_vars,
#   geneModel = reference,
#   bamDna = bamfile
# )


test_that(
  "Test ZygosityPredictor exportet functions",
  {
    expect_equal(aff_som_copies(3, 0.5, 2, 1, "f"), 1)
    expect_equal(aff_som_copies("X", 0.5, 2, 0.5, "m"), 1.5)
    expect_equal(aff_som_copies("X", 0.5, 2, 0.5, "f"), 2)

    expect_equal(aff_germ_copies("X", 0.5, 2, 0.5, "f"), 1)
    expect_equal(aff_germ_copies("X", 0.5, 2, 0.5, "m"), 0.5)

    expect_type(predict_zygosity(
      purity = PURITY,
      ploidy = PLOIDY,
      sex = SEX,
      somCna = GR_SCNA,
      somSmallVars = GR_SOM_SMALL_VARS,
      germSmallVars = GR_GERM_SMALL_VARS,
      geneModel = GR_GENE_MODEL,
      bamDna = FILE_BAM
    ), "list")

    expect_message(predict_zygosity(
      purity = PURITY,
      ploidy = PLOIDY,
      sex = SEX,
      somCna = GR_SCNA,
      geneModel = GR_GENE_MODEL,
      bamDna = FILE_BAM
    )
    )

    expect_error(predict_zygosity(
      purity = PURITY,
      ploidy = PLOIDY,
      sex = SEX,
      somCna = GR_SCNA[,1],
      somSmallVars = GR_SOM_SMALL_VARS,
      germSmallVars = GR_GERM_SMALL_VARS,
      geneModel = GR_GENE_MODEL,
      bamDna = FILE_BAM
    ))
  }
)

test_that(
  "Test ZygosityPredictor internal functions",
  {
    expect_equal(ZygosityPredictor:::check_tcn(2), 2)
    expect_error(ZygosityPredictor:::check_tcn("two"))
    expect_equal(ZygosityPredictor:::check_chr(22), "22")
    expect_equal(ZygosityPredictor:::check_chr("chr22"),"chr22")
    expect_error(ZygosityPredictor:::check_chr("chr44"))
    expect_error(ZygosityPredictor:::check_somCna("I am a sCNA result"))
    expect_error(ZygosityPredictor:::check_gr_gene_model("I am a gene model"))
    expect_null(ZygosityPredictor:::check_gr_small_vars(NULL, origin="somatic"))
    expect_equal(ZygosityPredictor:::check_purity("0.1"), 0.1)
    expect_error(ZygosityPredictor:::check_purity(4))
    expect_null(ZygosityPredictor:::check_ploidy(NULL))
    expect_equal(ZygosityPredictor:::check_ploidy("10"), 10)
    expect_equal(ZygosityPredictor:::check_sex("F"), "female")

  }
)

#' copynumber object
#' @return Object containing somatic copy number abberations (sCNAs) of SeqC2 example case
#' @usage data(GR_SCNA)
#' @format ## `GR_SCNA`
#' GRanges object
"GR_SCNA"
#' somatic small variant object
#' @return Object containing somatic Indels and SNVs of SeqC2 example case
#' @usage data(GR_GERM_SMALL_VARS)
#' @format ## `GR_GERM_SMALL_VARS`
#' GRanges object
"GR_SOM_SMALL_VARS"
#' germline small variant object
#' @return Object containing germline Indels and SNVs of SeqC2 example case
#' @usage data(GR_SOM_SMALL_VARS)
#' @format ## `GR_SOM_SMALL_VARS`
#' GRanges object
"GR_GERM_SMALL_VARS"
#' germline small variant object
#' @return Object containing gene model of hg38
#' @usage data(GR_GENE_MODEL)
#' @format ## `GR_GENE_MODEL`
#' GRanges object
"GR_GENE_MODEL"
#' haploblocks 
#' @return Object containing haploblock annotations
#' @usage data(GR_HAPLOBLOCKS)
#' @format ## `GR_HAPLOBLOCKS`
#' GRanges object
"GR_HAPLOBLOCKS"

get_single_index <- function(x,y,nr){
  nr*(x-1)+y
}
get_xy_index <- function(inp, nr){
  lapply(inp, function(i){
    y <- i%%nr
    if(y==0){
      y_ret <- nr
      x <- (i-y)/nr 
    } else {
      y_ret <- y
      x <- (i-y)/nr+1
    }
    return(c(x=x, y=y_ret))  
  }) %>% Reduce(function(x,y)rbind(x,y),.)
  
}
get_main_mut_pos <- function(ZP_env){
  lm <- length(ZP_env$global_ZygosityPredictor_variable_main_muts)
  lapply(1:lm, function(i){
    s <- (i-1)*nrow(ZP_env$global_ZygosityPredictor_variable_mat_phased)+1
    e <- (i-1)*nrow(ZP_env$global_ZygosityPredictor_variable_mat_phased)+lm
    return(seq(s,e,1))
  }) %>% unlist() %>%
    return()
}
unphased_main <- function(ZP_env){
  lm <- seq(1,length(ZP_env$global_ZygosityPredictor_variable_main_muts),1)
  which(is.na(ZP_env$global_ZygosityPredictor_variable_mat_phased[lm, lm]))
}
unknown_main <- function(ZP_env){
  all_unknown <- which(upper.tri(ZP_env$global_ZygosityPredictor_variable_mat_phased)&ZP_env$global_ZygosityPredictor_variable_mat_phased==0)
  intersect(all_unknown, get_main_mut_pos(ZP_env))
}
#' @importFrom stringr %>%
#' @importFrom dplyr select filter mutate bind_rows
#' @importFrom tibble as_tibble
add_snps_to_matrices <- function(snps, ZP_env){
  func_start(ZP_env)
  snps_hap <- snps[which(!is.na(snps$block_final)),] %>%
    select(mut_id, block_final, gt=gt_final)
  known_combs <- lapply(unique(snps_hap$block_final) %>% .[which(.!=0)], function(HAP_ID){
    
    snps_in_hap <- snps_hap %>%
      filter(block_final==HAP_ID) %>%
      mutate(id=as.numeric(str_match(mut_id, "\\d+")))
    if(nrow(snps_in_hap)>1){
      ## at least two snps in haploblock required to draw any concludios
      res_list <- list()
      for (i in 1:(nrow(snps_in_hap)-1)) {
        result_matrix <- matrix(0, nrow = nrow(snps_in_hap)-i, ncol = 2)
        for (j in (i+1):nrow(snps_in_hap)) {
          comparison_result <- ifelse(snps_in_hap$gt[i] == snps_in_hap$gt[j], 1, 2)
          # Store the results in the result_matrix
          result_matrix[nrow(snps_in_hap)-j+1, 1] <- get_single_index(snps_in_hap$id[j], snps_in_hap$id[i], nrow(ZP_env$global_ZygosityPredictor_variable_mat_phased))
          result_matrix[nrow(snps_in_hap)-j+1, 2] <- comparison_result
        }
        res_list[[i]] <- result_matrix
      }
      to_add <- Reduce(function(x,y)rbind(x,y),res_list)
      colnames(to_add) <- c("comb", "nconst") 
      new <- to_add %>%
        as_tibble() %>%
        mutate(conf=5, phasing=HAP_ID, hap_id=HAP_ID)     
    } else {
      ## if haploblock has only one snp, no concludions can be drawn
      new <- tibble()
    }
    return(new)
  })  %>% bind_rows()
  append_matrices(known_combs, ZP_env, iterate=FALSE)
  func_end(ZP_env)
}
eval_rare_case <- function(all_comb){
  diff_combs <- all_comb[which(all_comb$nconst==2),]
  if(nrow(diff_combs)>2){
    ## more than 2 combinatiosna re diff
    ## check for all posiible master combinations
    mut_occurence <- unlist(str_split(diff_combs$comb, "-")) %>% table() %>%
      .[which(.>=2)]
    if(length(mut_occurence)>=3){
      rare_case <- " rare_case_detected"
    } else {
      rare_case <- ""
    }
  } else {
    rare_case <- ""
  }
  return(rare_case)
}
#' @keywords internal
make_dist_matrix <- function(v, all_variants, distCutOff){
  n <- length(v)
  mat <- matrix(0, nrow = n, ncol = n)
  for (i in 1:n) {
    mat[i, ] <- abs(v - v[i])
  }
  rownames(mat) <- all_variants
  colnames(mat) <- all_variants
  mat[mat>distCutOff] <- 0
  mat[lower.tri(mat)] <- 0
  return(mat) 
}
create_phasing_matrices <- function(all_variants, all_pos, distCutOff, ZP_env){
  ZP_env$global_ZygosityPredictor_variable_mat_dist <- make_dist_matrix(all_pos, all_variants, distCutOff) 
  ZP_env$global_ZygosityPredictor_variable_main_muts <- all_variants
  ZP_env$global_ZygosityPredictor_variable_main_pos <- all_pos
  ZP_env$global_ZygosityPredictor_variable_mat_phased <- ZP_env$global_ZygosityPredictor_variable_mat_dist
  ZP_env$global_ZygosityPredictor_variable_mat_phased[ZP_env$global_ZygosityPredictor_variable_mat_phased!=0] <- NA  
  ZP_env$global_ZygosityPredictor_variable_mat_info <- ZP_env$global_ZygosityPredictor_variable_mat_phased
}
append_phasing_matrices <- function(all_variants, all_pos, distCutOff, ZP_env){
  func_start(ZP_env)
  ZP_env$global_ZygosityPredictor_variable_mat_dist <- make_dist_matrix(c(ZP_env$global_ZygosityPredictor_variable_main_pos, all_pos), 
                                c(ZP_env$global_ZygosityPredictor_variable_main_muts, all_variants), 
                                distCutOff) 
  mat_phased_main <- ZP_env$global_ZygosityPredictor_variable_mat_phased
  mat_info_main <- ZP_env$global_ZygosityPredictor_variable_mat_info
  ZP_env$global_ZygosityPredictor_variable_mat_phased <- ZP_env$global_ZygosityPredictor_variable_mat_dist
  ZP_env$global_ZygosityPredictor_variable_mat_phased[ZP_env$global_ZygosityPredictor_variable_mat_phased!=0] <- NA 
  ZP_env$global_ZygosityPredictor_variable_mat_info <- ZP_env$global_ZygosityPredictor_variable_mat_phased
  nm <- length(ZP_env$global_ZygosityPredictor_variable_main_muts)
  ZP_env$global_ZygosityPredictor_variable_mat_phased[c(1:nm), c(1:nm)] <- mat_phased_main
  ZP_env$global_ZygosityPredictor_variable_mat_info[c(1:nm), c(1:nm)] <- mat_info_main
  func_end(ZP_env)
}
#' @importFrom magrittr %>%
#' @importFrom purrr set_names
get_main_mut_conns <- function(ZP_env){
  lapply(ZP_env$global_ZygosityPredictor_variable_main_muts, function(M){
    length(c(which(ZP_env$global_ZygosityPredictor_variable_mat_phased[M,]!=0|is.na(ZP_env$global_ZygosityPredictor_variable_mat_phased[M,])),
             which(ZP_env$global_ZygosityPredictor_variable_mat_phased[,M]!=0|is.na(ZP_env$global_ZygosityPredictor_variable_mat_phased[,M]))
    ))
  }) %>% unlist() %>%
    set_names(ZP_env$global_ZygosityPredictor_variable_main_muts)
}
#' @importFrom magrittr %>%
#' @importFrom tibble as_tibble
#' @importFrom purrr set_names
#' @importFrom igraph graph_from_data_frame all_shortest_paths
get_next_path <- function(comb, ZP_env){
  func_start(ZP_env)
      
  mains <- as.character(sort(get_xy_index(as.numeric(comb), nrow(ZP_env$global_ZygosityPredictor_variable_mat_phased))))

  ldc <- 1000
        open_conns <- which((is.na(ZP_env$global_ZygosityPredictor_variable_mat_phased)|ZP_env$global_ZygosityPredictor_variable_mat_phased>0)&ZP_env$global_ZygosityPredictor_variable_mat_dist<ldc) %>%
          get_xy_index(.,nrow(ZP_env$global_ZygosityPredictor_variable_mat_phased))
        if(length(dim(open_conns))!=2){
          connections <- as_tibble(t(open_conns)) %>%
            set_names(c("mut_id1", "mut_id2"))
        } else {
          colnames(open_conns) <- c("mut_id1", "mut_id2")
          connections <- as_tibble(open_conns)
        }
  if(length(intersect(as.numeric(mains), 
                      c(connections$mut_id1, connections$mut_id2)))==2){
    graph <- igraph::graph_from_data_frame(connections, directed = FALSE)
    all_paths <- igraph::all_shortest_paths(graph, from = mains[1], to = mains[2], mode = "all")$res  
    if(length(all_paths)>0){
        poss <- list()
        for (PATH in all_paths){
          P <- as.numeric(names(PATH))
          res <- c()
          for(i in c(1:(length(P)-1))){
            mv <- sort(c(P[i], P[i+1]))
            conn <- get_single_index(mv[2], mv[1], nrow(ZP_env$global_ZygosityPredictor_variable_mat_phased))
            res[i] <- c(res, ZP_env$global_ZygosityPredictor_variable_mat_phased[conn])
            if(is.na(ZP_env$global_ZygosityPredictor_variable_mat_phased[conn])){
              poss <- c(poss, conn)
            }
          }
        } 
        np <-unlist(poss)[1]
      } else {
        np <- NULL
      } 
  } else {
    np <- NULL
  }
  func_end(ZP_env)
  return(np)
}
#' @importFrom purrr set_names
#' @importFrom stringr %>%
#' @importFrom dplyr left_join
prioritize_combination <- function(ZP_env){
  func_start(ZP_env)
  unrel_rows <- rownames(ZP_env$global_ZygosityPredictor_variable_mat_phased) %>% 
    .[which(!. %in% colnames(ZP_env$global_ZygosityPredictor_variable_mat_phased))]
  mat_tmp <- ZP_env$global_ZygosityPredictor_variable_mat_dist
  mat_tmp[unrel_rows,] <- NA
  mat_tmp[,unrel_rows] <- NA
  mat_tmp[c((length(ZP_env$global_ZygosityPredictor_variable_main_muts)+1):nrow(ZP_env$global_ZygosityPredictor_variable_mat_phased)),] <- NA
  unphased <- which(is.na(ZP_env$global_ZygosityPredictor_variable_mat_phased))
  relevant_dists <- mat_tmp[unphased] 
  relevant_dists[which(relevant_dists==0)] <- NA
  ## to suppress warning: Warning: no non-missing arguments to min; returning Inf
  suppressWarnings(
    shortest <- unphased[which(relevant_dists==min(relevant_dists, na.rm=TRUE))] 
  )
  ## pick first one, if two have the same distance
  if(length(shortest)==0){
    main_mut_conns <- get_main_mut_conns(ZP_env)
    df_mmc <- tibble(m=names(main_mut_conns),
                     n=main_mut_conns)
    uc <- unknown_main(ZP_env)
    if(length(uc)>0){
      df_unknown_pre <- lapply(uc, function(C){
        sort(get_xy_index(C, nrow(ZP_env$global_ZygosityPredictor_variable_mat_phased))) %>%
          paste0("m",.) %>%
          set_names(c("m1", "m2")) %>%
          c(.,comb=C)
      }) %>%
        bind_rows() 
      df_unknown <- df_unknown_pre %>%
        left_join(df_mmc %>% select(m, n1=n), by=c("m1"="m")) %>%
        left_join(df_mmc %>% select(m, n2=n), by=c("m2"="m")) %>%
        filter(n1>0&n2>0)
      if(nrow(df_unknown)>0){
        ## now check available paths
        next_path <- get_next_path(df_unknown[1,]$comb, ZP_env)
        if(!is.null(next_path)){
          next_comb <- get_xy_index(next_path, nrow(ZP_env$global_ZygosityPredictor_variable_mat_phased))
        } else {
          next_comb <- NULL
        }      
      } else {
        next_comb <- NULL
      }      
    } else {
      next_comb <- NULL
    }
  } else {
    next_comb <- get_xy_index(shortest[1], nrow(ZP_env$global_ZygosityPredictor_variable_mat_phased))
  }
  func_end(ZP_env)
  return(next_comb)
}
#' @importFrom magrittr %>%
#' @importFrom tibble tibble
#' @importFrom purrr set_names
#' @importFrom dplyr bind_rows left_join select filter
aggregate_phasing <- function(all_combs, df_gene, read_level_phasing_info, copy_number_phasing=NULL, ZP_env){
  func_start(ZP_env)
  phasing_all_combs <- lapply(all_combs, function(mmp){
    comb_vec <- paste0("m",sort(get_xy_index(mmp, nrow(ZP_env$global_ZygosityPredictor_variable_mat_phased))))
    comb <- comb_vec %>% paste(collapse="-")    
    df_gene_relcomb <- df_gene %>%
        filter(mut_id %in% comb_vec)
    min_tcn <- min(df_gene_relcomb$tcn)
    nconst <- ZP_env$global_ZygosityPredictor_variable_mat_phased[mmp]
    const <- get_string_const(nconst)      
    min_poss_wt_cp <- calc_left_wt_copies(min_tcn,
                                         2,
                                         df_gene_relcomb$aff_cp[1],
                                         df_gene_relcomb$aff_cp[2])
    max_poss_wt_cp <- calc_left_wt_copies(min_tcn,
                                         1,
                                         df_gene_relcomb$aff_cp[1],
                                         df_gene_relcomb$aff_cp[2])
    ## gescheit machen irgendwann
    if(is.na(nconst)){
      nconst <- 0
    }
    
    if(nconst>0){
      ## status was defined
      ## calculate confidence
      used_combs <- ZP_env$global_ZygosityPredictor_variable_mat_info[mmp] %>% 
        str_split("-") %>% 
        unlist() %>%
        as.numeric()
      ## check if copynumber phasing was applied
      if(!str_detect(ZP_env$global_ZygosityPredictor_variable_mat_info[mmp], "som")){
        extracted_combs <- read_level_phasing_info %>% filter(comb %in% used_combs) 
        conf <- extracted_combs %>%
          mutate(ep=ifelse(nconst==2, 1-p_same, 1-p_diff)) %>%
          pull(ep) %>%
          prod()
        unplausible <- paste(as.numeric(extracted_combs$unplausible), collapse="-")
        subclonal <- paste(as.numeric(extracted_combs$subclonal), collapse="-")
        
        phasing <- case_when(
          str_detect(ZP_env$global_ZygosityPredictor_variable_mat_info[mmp], "h") ~ "haploblock-phasing",
          #str_detect(ZP_env$global_ZygosityPredictor_variable_mat_info[mmp], "s") ~ "imbalance",
          str_detect(ZP_env$global_ZygosityPredictor_variable_mat_info[mmp], "-") ~ "indirect-phasing",
          TRUE ~ "direct-phasing"
        )
      } else {
        phasing <- "imbalance-phasing"
        #conf <- paste(copy_number_phasing$conf, collapse = ";")
        conf <- prod(copy_number_phasing$conf)
        unplausible <- NA
        subclonal <- NA
      }
      via <- as.character(ZP_env$global_ZygosityPredictor_variable_mat_info[mmp])        
      wt_cp <- calc_left_wt_copies(min_tcn,
                                  nconst,
                                  df_gene_relcomb$aff_cp[1],
                                  df_gene_relcomb$aff_cp[2])
      score <- case_when(
          nconst==2&wt_cp<0.5 ~ 2,
          TRUE ~ 1)
    } else {
      ## status could not be defined
      ## calculate maximum possible affected copies
      conf <- 0
      unplausible <- NA
      subclonal <- NA
      via <- NA
      wt_cp <- NA
      score <- case_when(
        min_poss_wt_cp>0.5 ~ 1,
        TRUE ~ 0
      )
      phasing <- case_when(
        score==1 ~ "insufficient",
        TRUE ~ NA
      )
    }
    phasing_status <- tibble(comb=comb, nconst=nconst, const=const, 
                             phasing=phasing, via=via, conf=conf, 
                             unplausible=unplausible,
                             subclonal=subclonal, wt_cp=wt_cp, 
                             min_poss_wt_cp=min_poss_wt_cp,
                             max_poss_wt_cp=max_poss_wt_cp,
                             score=score)
    return(phasing_status)
  }) %>%
    bind_rows()  %>%
    select(comb, nconst, const, phasing, via, conf, unplausible, subclonal, 
           wt_cp, min_poss_wt_cp, max_poss_wt_cp, score)
  func_end(ZP_env)
  return(phasing_all_combs)
}
#' @keywords internal
#' @importFrom magrittr %>%
#' @importFrom tibble tibble column_to_rownames as_tibble
#' @importFrom dplyr group_by tally mutate ungroup rowwise pull filter summarize left_join select
#' @importFrom stats chisq.test
classify_combination <- function(classified_reads, ref_class1, ref_class2, 
                                 purity, printLog, ZP_env){
  func_start(ZP_env)
  result <- . <- fac <- NULL
  all_possible_results <- c('both', 'mut1', 'mut2', 'none', 
                            "dev_var", 'read_in_read', 
                            'skipped')
  ## calculate expected counts according to aff copies for both cases:
  ## same and diff
  pseudo_count <- 0.0000001
  exp_diff <- tibble(fac=all_possible_results[1:3],
                     exp=c(0,1,1)+pseudo_count)
  exp_same <- tibble(fac=all_possible_results[1:3],
                     exp=c(1,0,0)+pseudo_count)
  number <- classified_reads %>%
    mutate(fac=factor(result, levels = all_possible_results)) %>%
    group_by(fac, .drop = FALSE) %>%
    tally() %>% 
    column_to_rownames(var='fac') %>%
    t() %>%
    as_tibble()
  none_raw <- number[['none']]
  append_loglist(print_tibble(number), ZP_env=ZP_env)
  ## calculate confidence from basecalls and mapping quality of read
  ## and aggregate them per classification result
  cr_conf <- classified_reads %>%
    rowwise() %>%
    mutate(pb1=ascii_to_dec(baseq1),
           pb2=ascii_to_dec(baseq2)) %>%
    ungroup() %>%
    mutate(
      mq1=10^(as.numeric(mapq1)/(-10)),
      mq2=10^(as.numeric(mapq2)/(-10)),
      ppos1=(1-pb1)*(1-mq1),
      ppos2=(1-pb2)*(1-mq2),
      p_result=ppos1*ppos2
    ) %>%
    mutate(fac=factor(result, levels = all_possible_results)) %>%
    group_by(fac, .drop = FALSE) %>%
    summarize(n=length(fac),
              prob_sum=sum(p_result))
  relevant_for_decision <- cr_conf %>%
    filter(fac %in% c("both","mut1", "mut2")) %>%
    select(fac, prob_sum)
  both <- relevant_for_decision %>%
    filter(fac=="both") %>%
    pull(prob_sum)
  mut1 <- relevant_for_decision %>%
    filter(fac=="mut1") %>%
    pull(prob_sum)
  mut2 <- relevant_for_decision %>%
    filter(fac=="mut2") %>%
    pull(prob_sum)
  append_loglist(print_tibble(relevant_for_decision), ZP_env=ZP_env)
  ## predefine null result if no evidence
  const <- "null"
  p <- 1
  xsq_same <- xsq_diff <- p_diff <- p_same <- v_same <- v_diff <- NA
  nconst <- evidence <- certainty <- confidence <- conf_log <- 0
  subclonal <- unplausible <- FALSE
  ## get total number of relevant classifications
  sum_rel <- sum(relevant_for_decision$prob_sum)
  if(sum_rel==0){
    append_loglist("no evidence for any classification", ZP_env=ZP_env)
  } else if(both==0&mut1==0&ref_class2=="snp"){
    append_loglist("only mut2 detected, which is SNP", ZP_env=ZP_env)
  } else if(both==0&mut2==0&ref_class1=="snp"){
    append_loglist("only mut1 detected, which is SNP", ZP_env=ZP_env)
  } else {
    ## check for similarity of numbers in both, mut1 and mut2 by chi-squared
    suppressWarnings(
      sim_diff <- chisq.test(t(left_join(exp_diff, relevant_for_decision,
                                         by="fac") %>% 
                                 select(exp, prob_sum)))
    )
    suppressWarnings(
      sim_same <- chisq.test(t(left_join(exp_same, relevant_for_decision,
                                         by="fac") %>% 
                                 select(exp, prob_sum)))  
    )
    xsq_diff <- sim_diff$statistic[[1]]
    xsq_same <- sim_same$statistic[[1]]
    p_diff <- sim_diff$p.value
    p_same <- sim_same$p.value
    ## calculate cramers V... as we have always 2x3 table, m=1
    v_diff <- sqrt(xsq_diff/sum(sum_rel, exp_diff$exp))
    v_same <- sqrt(xsq_same/sum(sum_rel, exp_same$exp))
    if(p_diff>p_same){
      const <- "diff"
      nconst <- 2
    } else {
      const <- "same"
      nconst <- 1
    }  
  }
  ## check if evidence for subclonal sample is given  
  if(both>0&((mut1==0&mut2>0)|(mut1>0&mut2==0))){
    if(!"snp" %in% c(ref_class1, ref_class2)){
      warning("evidence fur subclonality found during phasing!")
      subclonal <- TRUE      
    }
  }
  ## check for unplausible case
  if(both>0&mut1>0&mut2>0){
    warning("unplausible phasing result!")
    unplausible <- TRUE
  }
  status_table <- tibble(
    both=both,
    mut1=mut1,  
    mut2=mut2,
    dev_var=number[['dev_var']],
    skipped=sum(as.numeric(number[['read_in_read']]), 
                   as.numeric(number[['skipped']])),
    const=const,
    nconst=nconst,
    xsq_same=xsq_same,
    xsq_diff=xsq_diff,
    p_same=p_same,
    p_diff=p_diff,
    v_same=v_same,
    v_diff=v_diff,
    none_raw=none_raw,
    DNA_rds=nrow(classified_reads %>% filter(origin=="DNA")),
    RNA_rds=nrow(classified_reads %>% filter(origin=="RNA")),
    subclonal=subclonal,
    unplausible=unplausible
  )
  func_end(ZP_env)
  return(status_table)
}  
#' @keywords internal
#' @importFrom stringr %>%
#' @importFrom GenomicRanges GRanges elementMetadata
#' @importFrom GenomicAlignments readGAlignmentPairs first last
#' @importFrom Rsamtools ScanBamParam
#' @importFrom IRanges subsetByOverlaps
#' @importFrom dplyr tibble
prepare_raw_bam_file <- function(bamDna, chr1, chr2, pos1, pos2, ZP_env){
  func_start(ZP_env)
  qname.first <- . <- NULL
  ## importFrom dplyr tibble filter
  ref_pos1 <- as.numeric(pos1)
  ref_pos2 <- as.numeric(pos2)
  ref_chr1 <- as.character(chr1)
  ref_chr2 <- as.character(chr2)
  if(pos1>pos2){
    #stop("avoid changing position order")
    ## exchange does not matter here
    ref_pos1 <- as.numeric(pos2)
    ref_pos2 <- as.numeric(pos1)
    ref_chr1 <- as.character(chr2)
    ref_chr2 <- as.character(chr1)
  }
  ref_gr1 <- GRanges(seqnames = ref_chr1, 
                     ranges = ref_pos1)
  ref_gr2 <- GRanges(seqnames = ref_chr2, 
                     ranges = ref_pos2)
  ## now load all reads/read-pairs that cover the position of the first variant
  all_covering_read_pairs <-GenomicAlignments:: readGAlignmentPairs(
    bamDna,
    param=Rsamtools::ScanBamParam(
      which=ref_gr1,
      what=c("qname","seq", "cigar", "mapq", "qual")
    )) 
  if(length(all_covering_read_pairs)==0){
    filtered_reads <- tibble()
    #print(3)
  } else {
    ## combine all ranges and check for ref_pos2
    ## from here for development
    all_reads <- c(
      GenomicAlignments::first(all_covering_read_pairs) %>%
        GRanges(),
      GenomicAlignments::last(all_covering_read_pairs) %>%
        GRanges()
    )
    shared_read_pairs <- all_reads %>%
      subsetByOverlaps(.,ref_gr2) %>%
      elementMetadata(.) %>%
      .[["qname"]] %>%
      unique()
    
    filtered_reads <- all_reads[
      which(all_reads$qname %in% shared_read_pairs)
    ]
    #print(6)
  }
  func_end(ZP_env)
  return(filtered_reads)
}
#' @keywords internal
#' description follows
check_for_overlapping_reads <- function(bamDna, bamRna,
                                        ref_chr1, 
                                        ref_chr2, 
                                        ref_pos1, 
                                        ref_pos2, ZP_env){
  func_start(ZP_env)
  dna_bam <- prepare_raw_bam_file(bamDna, 
                                  ref_chr1, 
                                  ref_chr2, 
                                  ref_pos1, 
                                  ref_pos2, ZP_env) 
  if(length(dna_bam)!=0){
    dna_bam$origin <- "DNA"
  }
  if(!is.null(bamRna)){
    rna_bam <- prepare_raw_bam_file(bamRna, 
                                    ref_chr1, 
                                    ref_chr2, 
                                    ref_pos1, 
                                    ref_pos2, ZP_env)
    if(length(rna_bam)==0){
      rna_bam <- NULL
    } else {
      rna_bam$origin <- "RNA"
    }
  } else {
    rna_bam <- NULL
  }
  func_end(ZP_env)
  return(c(dna_bam, rna_bam))
}
#' @keywords internal
#' @importFrom magrittr %>%
#' @importFrom dplyr tibble bind_rows 
classify_reads <- function(ref_pos1,
                           ref_pos2,
                           ref_chr1,
                           ref_chr2,
                           ref_alt1,
                           ref_alt2,
                           ref_ref1,
                           ref_ref2,
                           ref_class1,
                           ref_class2, bamDna, bamRna, ZP_env){
  vm(as.character(sys.call()[1]),  1, ZP_env=ZP_env)
  bam <- check_for_overlapping_reads(bamDna,
                                     bamRna,
                                     ref_chr1,
                                     ref_chr2,
                                     ref_pos1,
                                     ref_pos2, ZP_env)
  if(!length(bam)==0){  
    classified_reads <- lapply(unique(bam$qname),
                               core_tool,
                               bam, 
                               ref_pos1,
                               ref_pos2,
                               ref_alt1,
                               ref_alt2,
                               ref_ref1,
                               ref_ref2,
                               ref_class1,
                               ref_class2) %>%
      bind_rows() 
  } else {
    classified_reads <- tibble()
  } 
  func_end(ZP_env)
  return(classified_reads)
}
#' @importFrom magrittr %>%
#' @importFrom tibble tibble
#' @importFrom dplyr mutate
phase_combination <- function(mat_gene_relcomb, comb, bamDna, bamRna,  
                              geneDir, phasing_type, showReadDetail, ZP_env){
  func_start(ZP_env)
  append_loglist("Phasing:", comb, ZP_env=ZP_env)
  ## predefine empty output
  classified_main_comb <- tibble(comb=comb, const="null", nconst=0, conf=0,
                                 phasing=comb)
  ## define relevant properties for mut1 and mut2 globally for the phasing process
  mut1 <- 2
  mut2 <- 1
  
  ref_pos1 <- as.numeric(mat_gene_relcomb[,"pos"][[mut1]])  
  ref_pos2 <- as.numeric(mat_gene_relcomb[,"pos"][[mut2]])
  ## must be character to work for annotations like: chrX
  ref_chr1 <- as.character(mat_gene_relcomb[,"chr"][[mut1]])  
  ref_chr2 <- as.character(mat_gene_relcomb[,"chr"][[mut2]])
  
  ref_alt1 <- as.character(mat_gene_relcomb[,"alt"][[mut1]])
  ref_alt2 <- as.character(mat_gene_relcomb[,"alt"][[mut2]]) 
  
  ref_ref1 <- as.character(mat_gene_relcomb[,"ref"][[mut1]])
  ref_ref2 <- as.character(mat_gene_relcomb[,"ref"][[mut2]])
  
  ref_class1 <- as.character(mat_gene_relcomb[,"class"][[mut1]])
  ref_class2 <- as.character(mat_gene_relcomb[,"class"][[mut2]])
  
  main_classified_reads <- classify_reads(ref_pos1,
                                          ref_pos2,
                                          ref_chr1,
                                          ref_chr2,
                                          ref_alt1,
                                          ref_alt2,
                                          ref_ref1,
                                          ref_ref2,
                                          ref_class1,
                                          ref_class2, 
                                          bamDna, 
                                          bamRna,
                                          ZP_env=ZP_env)
  append_loglist(nrow(main_classified_reads), 
                 "reads / read-pairs covering both positions", ZP_env=ZP_env)
  if(nrow(main_classified_reads)!=0){
    if(showReadDetail==TRUE){
      store_log(geneDir, main_classified_reads %>%
                  mutate(comb=comb),
                paste0("classified_reads_", comb,".tsv"))
    }
    classified_main_comb <- classify_combination(main_classified_reads,
                                                 ref_class1, ref_class2,
                                                 purity,
                                                 printLog, ZP_env
                                                 
    ) %>%  
      mutate(
        dist=ZP_env$global_ZygosityPredictor_variable_mat_dist[comb],
        class_comb=paste(mat_gene_relcomb[,"class"][[2]], 
                         mat_gene_relcomb[,"class"][[1]], 
                         sep="-"),
        comb=comb,
        phasing=comb
      ) 
  }   
  func_end(ZP_env)
  return(classified_main_comb)
}
#' @importFrom magrittr %>%
append_matrices <- function(classified_main_comb, ZP_env, iterate=TRUE){
  func_start(ZP_env)
  if(nrow(classified_main_comb)>0){
    mat_old <- ZP_env$global_ZygosityPredictor_variable_mat_phased
    lapply(seq(1,nrow(classified_main_comb)), function(C){
      ZP_env$global_ZygosityPredictor_variable_mat_phased[classified_main_comb[C,]$comb] <- 
        classified_main_comb[C,]$nconst
      ZP_env$global_ZygosityPredictor_variable_mat_info[classified_main_comb[C,]$comb] <- 
        classified_main_comb[C,]$phasing
      return()
    })
    if(iterate==TRUE){
      something_changed <- sum(mat_old, na.rm = TRUE)!=sum(ZP_env$global_ZygosityPredictor_variable_mat_phased, na.rm = TRUE)
      mat_new <- ZP_env$global_ZygosityPredictor_variable_mat_phased
      mat_new[is.na(mat_new)] <- 0
      mat_old[is.na(mat_old)] <- 0
      changes <- which(!mat_new==mat_old) 
      i <- 1
      while(i<=length(changes)){
        m12 <- get_xy_index(changes[i], nrow(mat_new))
        status <- mat_new[changes[i]]
        ## get third mut/snp id that also has a non-null connection to one of the first ones 
        m1_same_conn <- c(m12[1], which(mat_new[m12[1],]==1),
                          which(mat_new[,m12[1]]==1)) %>% sort()
        m2_same_conn <- c(m12[2],
                          which(mat_new[m12[2],]==1),
                          which(mat_new[,m12[2]]==1)) %>% sort()
        if(sum(c(length(m1_same_conn), length(m2_same_conn)))>2){
          if(status==1){
            conns <- sort(unique(c(m1_same_conn, m2_same_conn)))
            for(x in 1:(length(conns)-1)){
              for(j in (x+1):length(conns)){
                p1 <- conns[x]
                p2 <- conns[j]
                notxf <- m12[which(!m12 %in% c(p1,p2))]
                if(length(notxf)>=1){
                  notx <- min(notxf)
                  if(is.na(ZP_env$global_ZygosityPredictor_variable_mat_phased[p1,p2])|ZP_env$global_ZygosityPredictor_variable_mat_phased[p1,p2]==0){
                    combined_info <- paste(
                      c(ZP_env$global_ZygosityPredictor_variable_mat_info[
                        min(c(notx, p1)),
                        max(c(notx, p1))
                      ], ZP_env$global_ZygosityPredictor_variable_mat_info[
                        min(c(notx, p2)),
                        max(c(notx, p2))                   
                      ]), 
                      collapse="-")
                    ZP_env$global_ZygosityPredictor_variable_mat_info[p1,p2] <- combined_info   
                  }
                } 
                if(is.na(ZP_env$global_ZygosityPredictor_variable_mat_phased[p1,p2])|ZP_env$global_ZygosityPredictor_variable_mat_phased[p1,p2]==0){
                  ZP_env$global_ZygosityPredictor_variable_mat_phased[p1,p2] <- 1
                }
              }
            }
          } else {
            for(x in m1_same_conn){
              for(j in m2_same_conn){
                sv <- sort(c(x,j))
                p1 <- sv[1]
                p2 <- sv[2]
                notxf <- m12[which(!m12 %in% c(x,j))]
                if(length(notxf)>=1){
                  notx <- min(notxf)
                  if(is.na(ZP_env$global_ZygosityPredictor_variable_mat_phased[p1,p2])|ZP_env$global_ZygosityPredictor_variable_mat_phased[p1,p2]==0){
                    ZP_env$global_ZygosityPredictor_variable_mat_info[p1, p2] <- paste(
                      c(ZP_env$global_ZygosityPredictor_variable_mat_info[
                        min(c(notx, x)),
                        max(c(notx, x))
                      ], ZP_env$global_ZygosityPredictor_variable_mat_info[
                        min(c(notx, j)),
                        max(c(notx, j))                   
                      ]), 
                      collapse="-")
                  }
                }
                if(is.na(ZP_env$global_ZygosityPredictor_variable_mat_phased[p1, p2])|ZP_env$global_ZygosityPredictor_variable_mat_phased[p1, p2]==0){
                  ZP_env$global_ZygosityPredictor_variable_mat_phased[p1, p2] <- 2
                }        
              }
            }
          }      
        }
        i <- i+1 
      } 
    }
  }
  func_end(ZP_env)
}
#' @importFrom tibble tibble
#' @importFrom dplyr  bind_rows
perform_direct_phasing <- function(df_gene, bamDna, bamRna, 
                                    printLog, showReadDetail, geneDir, ZP_env){
  func_start(ZP_env)
  mat_gene <- as.matrix(df_gene)
  rownames(mat_gene) <- mat_gene[,"mut_id"]
  phasing_type <- "direct"
  read_level_phasing_info <- tibble()
  unphased <- which(is.na(ZP_env$global_ZygosityPredictor_variable_mat_phased))
  i <- 1
  while(i <= length(unphased)){
    ## as long as there are unphased combinations, try to phase
    comb <- unphased[i]
    relcombxy <- get_xy_index(comb, nrow(ZP_env$global_ZygosityPredictor_variable_mat_phased))
    mat_gene_relcomb <- mat_gene[relcombxy,]
    
    classified_main_comb <- phase_combination(mat_gene_relcomb, comb, bamDna, bamRna, 
                                               geneDir, phasing_type, showReadDetail, ZP_env)
    read_level_phasing_info <- bind_rows(read_level_phasing_info,
                              classified_main_comb)
    i <- i+1
  }
  append_matrices(read_level_phasing_info, ZP_env)
  func_end(ZP_env)
  return(read_level_phasing_info)
}
perform_indirect_phasing <- function(df_gene, vcf, bamDna, bamRna, haploBlocks,  
                                     distCutOff, somCna, snpQualityCutOff, 
                                     purity, sex, geneDir, read_level_phasing_info, 
                                     showReadDetail, ZP_env){
  func_start(ZP_env)
  if(length(unknown_main(ZP_env))>0&!is.null(vcf)){
    append_loglist("unphased combinations left --> Initialize SNP phasing", ZP_env=ZP_env)
    ## missing combinations in main muts --> start secondary phasing approaches
    lsnps <- load_snps(df_gene, vcf, haploBlocks,  distCutOff,  somCna, snpQualityCutOff, ZP_env)
    store_log(geneDir, lsnps, "lsnps.tsv")
    if(!is.null(lsnps)){
      snps <- lsnps %>%
        ## filter low quality
        rowwise() %>%
        mutate(
          aff_cp=ifelse(is.na(af)|is.na(tcn),
                        NA,
                        aff_germ_copies(seqnames, af, tcn, purity, sex))
        ) %>%
        ungroup() %>%
        mutate(
          ## i currently disabled SNP imbalance phasing as genotype likelihood needs to be implemented here as well
          gt_seg=NA,
          gt_final=case_when(
            gt=="1|1"|gt=="0|0" ~ NA,
            !is.na(hap_id)&str_detect(gt, "\\|") ~ gt,
            !is.na(gt_seg) ~ gt_seg,
            TRUE ~ NA
          ),
          block_final=case_when(
            gt=="1|1"|gt=="0|0" ~ NA,
            !is.na(hap_id)&str_detect(gt, "\\|") ~ paste0("h", hap_id),
            !is.na(gt_seg) ~ paste0("s", seg_id),
            TRUE ~ NA
          )
        )
      append_phasing_matrices(snps$mut_id, snps$start, distCutOff, ZP_env)
      add_snps_to_matrices(snps, ZP_env)
      ## build main daatframe of muts and snps and annotate haploblocks and segments of allelic imbalance
      df <- bind_rows(
        df_gene %>% select(chr, pos, ref, alt, class, af, mut_id),
        snps %>% select(chr=seqnames, pos=start, ref=REF, alt=ALT, af, mut_id, gt=gt_final) %>% mutate(class="snp")
      )
      to_phase <- prioritize_combination(ZP_env)
      i <- 1
      while(!is.null(to_phase)&length(unknown_main(ZP_env))>0){
        append_loglist("Phasing combination:", paste(paste0(to_phase), collapse="-"), ZP_env=ZP_env)
        
        mat_gene_relcomb <- df[to_phase,] %>% as.matrix()
        comb <- get_single_index(to_phase[1],
                                 to_phase[2],
                                 nrow(ZP_env$global_ZygosityPredictor_variable_mat_phased))
        classified_main_comb <- phase_combination(mat_gene_relcomb, comb,
                                                  bamDna, bamRna,  
                                                  geneDir, comb, showReadDetail, ZP_env)
        append_matrices(classified_main_comb, ZP_env)
        read_level_phasing_info <- bind_rows(read_level_phasing_info,
                                  classified_main_comb)
        to_phase <- prioritize_combination(ZP_env)
        i <- i +1
      }
      store_log(geneDir, df, "df_snps_muts.tsv")
      store_log(geneDir, read_level_phasing_info, "all_indirect_phasing_combinations.tsv")
    }## no snps found
  } 
  func_end(ZP_env)
  return(read_level_phasing_info)
}
#' @importFrom GenomicAlignments readGAlignments granges
#' @importFrom Rsamtools ScanBamParam scanBamFlag
custom_readGalign <- function(bamDna, ref_gr, first, ZP_env){
  func_start(ZP_env)
  loaded <- GenomicAlignments::readGAlignments(
    bamDna,
    param=Rsamtools::ScanBamParam(
      flag=Rsamtools::scanBamFlag(isFirstMateRead=first),
      which=ref_gr,
      what=c("qname","seq", "cigar", "mapq", "qual")
    ))    %>%
    GenomicAlignments::granges(use.mcols = TRUE)
  func_end(ZP_env)
  return(loaded)
}
load_covering_reads <- function(ref_gr, rel_mut, bamDna, ZP_env){
  func_start(ZP_env)
  all_covering_reads_firstmate <- custom_readGalign(bamDna, ref_gr, TRUE, ZP_env)
  all_covering_reads_secondmate <- custom_readGalign(bamDna, ref_gr, FALSE, ZP_env)
  ## remove overlapping reads from same pair... so that its not counted two times
  all_covering_reads_rem_dup <- c(
    all_covering_reads_firstmate,
    all_covering_reads_secondmate[which(!all_covering_reads_secondmate$qname %in% all_covering_reads_firstmate$qname)]
  )
  if(length(all_covering_reads_rem_dup)>0){
    all_covering_reads_rem_dup$origin <- "DNA"
    all_reads <- lapply(all_covering_reads_rem_dup$qname, function(QNAME){
      parsed_read <- parse_cigar(all_covering_reads_rem_dup, QNAME, paired=FALSE)
      base_info1 <- extract_base_at_refpos(parsed_read, 
                                           rel_mut$pos, 
                                           rel_mut$class, 
                                           rel_mut$alt, 
                                           rel_mut$ref)  
    }) %>%
      bind_rows()    
  } else {
    all_reads <- NULL
  }

  func_end(ZP_env)
  return(all_reads)
}
formula_genotype_likelihood <- function(gtl_g, 
                                        gtl_eps_l,
                                        gtl_eps_v,
                                        gtl_m,
                                        gtl_k){
    gtl_prod_ref <- unlist(lapply(gtl_eps_l, function(E){
      (gtl_m-gtl_g)*E+gtl_g*(1-E)
    })) %>%
      prod()
    gtl_prod_alt <- unlist(lapply(gtl_eps_v, function(E){
      (gtl_m-gtl_g)*(1-E)+gtl_g*E
    })) %>%
      prod()
    ## full formula
    genotype_likelihood <- (1/(gtl_m^gtl_k))*gtl_prod_ref*gtl_prod_alt  
    ## formula without primary factor
    return(c(gt=gtl_g, lklhd=genotype_likelihood, prod_ref=gtl_prod_ref, prod_alt=gtl_prod_alt))
}

adjust_n_reference_reads_somatic <- function(a, c_tum, ac_tum){
  
  n_ref_adj <- a*((c_tum/ac_tum)-1)
  
  rounded <- round(n_ref_adj)
  return(rounded)
}

adjust_n_reference_reads_germline <- function(r, p, c_tum, c_norm, VAF_norm, VAF_tum){
  
  
  n_ref_adj <- r*(-((p-1)/c_tum))/((((p/c_norm)-((p-1)/c_tum))*VAF_norm*VAF_tum)/(VAF_norm-1)    +  (((p-1)/c_tum)-(p/c_norm))*VAF_tum)
  return(round(n_ref_adj))
}



calc_genotype_likelihood_per_mut <- function(variants_in_segment, 
                                             allele_specific_genotype, bamDna, 
                                             purity, sex, ZP_env){
  func_start(ZP_env)
  gt_lk_per_mut <- lapply(variants_in_segment$mut_id, function(MUT){
    rel_mut <- variants_in_segment[which(variants_in_segment$mut_id==MUT),]
    ref_gr <- GRanges(seqnames = rel_mut$chr, 
                      ranges = rel_mut$pos)
    ## now load all reads/read-pairs that cover the position of the first variant
    all_reads <- load_covering_reads(ref_gr, rel_mut, bamDna, ZP_env)%>%
      mutate(processed_mapq=10^(as.numeric(mapq)/(-10)))
    gtl_m <- rel_mut$tcn
    gtl_k <- nrow(all_reads)
    
    if(!is.null(all_reads)){
      if(rel_mut$class=="snv"){
        gtl_eps_v_raw <- all_reads %>%
          .[which(.$base==rel_mut$alt),] %>%
          rowwise() %>%
          mutate(processed_bq=ascii_to_dec(qual),
            aggregated_qual=1-(1-processed_bq)*(1-processed_mapq)) %>%
          pull(aggregated_qual) 
        ## epsylon l has all reference read basecall qualities
        ## it must be adjusted to only the somatic ones
        gtl_eps_l_raw <- all_reads %>%
          .[which(.$base==rel_mut$ref),] %>%
          rowwise() %>%
          mutate(processed_bq=ascii_to_dec(qual),
                 aggregated_qual=1-(1-processed_bq)*(1-processed_mapq)) %>%
          pull(aggregated_qual)              
      } else {
        gtl_eps_v_raw <- all_reads %>%
          .[which(.$exp_indel==TRUE),] %>%
          .$processed_mapq 
        gtl_eps_l_raw <- all_reads %>%
          .[which(.$exp_indel==FALSE),] %>%
          .$processed_mapq 
      }      
      ## we defined v as k-l from genotype likelihood function
      
      ## as we dont expect purity of 100 %, we need to adjust for that
      
      if(rel_mut$origin=="germline"){
        
        cv <- formula_checks(chr=rel_mut$chr, 
                       af=rel_mut$af, 
                       tcn=rel_mut$tcn, 
                       purity=purity, 
                       sex=sex, 
                       c_normal=NULL)
        
        n_reference_reads_in_control <- adjust_n_reference_reads_germline(
          r=length(gtl_eps_l_raw), 
          p=purity, 
          c_tum=rel_mut$tcn, 
          c_norm=cv$c_normal,
          VAF_norm=cv$af_normal, 
          VAF_tum=rel_mut$af
          )
        
        mn_n_ref <- min(0, n_reference_reads_in_control)
        
        ## so far, we do not include other genomic constellatons than HZ
        ## so the number of alt reads to remove is the same a ref reads
        n_alternative_reads_in_control <- n_reference_reads_in_control
        mn_n_alt <- min(0, n_alternative_reads_in_control)
        
        to_take_ref <- length(gtl_eps_l_raw)-mn_n_ref
        to_take_alt <- length(gtl_eps_v_raw)-mn_n_alt
        
        gtl_eps_l <- sort(gtl_eps_l_raw, decreasing = FALSE)[seq(1,to_take_ref,1)] 
        gtl_eps_v <- sort(gtl_eps_v_raw, decreasing = FALSE)[seq(1,to_take_alt,1)] 
        
      } else {
        n_reference_reads_to_use <- adjust_n_reference_reads_somatic(a=length(gtl_eps_v_raw), 
                                                                     c_tum=rel_mut$tcn, 
                                                                     ac_tum=rel_mut$aff_cp)
        
        
        mx_n <- min(n_reference_reads_to_use, length(gtl_eps_l_raw))
        ## to stay most restrictive, we only use the n_reference reads wich have the lowest basecall qualities
        gtl_eps_l <- sort(gtl_eps_l_raw, decreasing = FALSE)[seq(1,mx_n,1)]  
        gtl_eps_v <- gtl_eps_v_raw
      }
      
      prob_per_gt <- lapply(allele_specific_genotype, function(gtl_g){
        formula_genotype_likelihood(gtl_g,
                                    gtl_eps_l,
                                    gtl_eps_v,
                                    gtl_m,
                                    (length(gtl_eps_l)+length(gtl_eps_v))) %>%
          bind_rows() %>%
          return()
      }) %>%
        bind_rows() %>%
        mutate(mut_id=MUT) 
      return(prob_per_gt)
    } else {
      warning("Genotype-likelihood can not be determined as provided bam file has no reads in the area of the variant")
      return(NULL)
    }
    
    }) %>%
      bind_rows() 
    
  func_end(ZP_env)
  return(gt_lk_per_mut)
}
make_mut_template <- function(variants_in_segment, ZP_env){
  func_start(ZP_env)
  template <- expand_grid(variants_in_segment$mut_id, variants_in_segment$mut_id) %>%
    apply(.,1,function(muts){
      sorted <- as.character(sort(c(muts))) %>%
        unique()
      if(length(sorted)==1){
        return()
      } else {
        return(sorted)
      }
    }, simplify=FALSE) %>%
    compact() %>%
    unique() %>%
    lapply(.,set_names, nm=c("mut_id1", "mut_id2"))
  func_end(ZP_env)
  return(template)
}
aggregate_likelihoods <- function(mut_template, gt_lk_per_mut, allele_specific_genotype, ZP_env){
  func_start(ZP_env)
  #print(mut_template)
  aggregated_likelihoods <- mut_template %>%
    lapply(., function(comb){
      #print(comb)
      p_m1_gt1 <- gt_lk_per_mut %>%
        filter(mut_id==comb[["mut_id1"]]&gt==allele_specific_genotype[1]) %>%
        pull(lklhd)
      p_m1_gt2 <- gt_lk_per_mut %>%
        filter(mut_id==comb[["mut_id1"]]&gt==allele_specific_genotype[2])%>%
        pull(lklhd)
      p_m2_gt1 <- gt_lk_per_mut %>%
        filter(mut_id==comb[["mut_id2"]]&gt==allele_specific_genotype[1])%>%
        pull(lklhd)
      p_m2_gt2 <- gt_lk_per_mut %>%
        filter(mut_id==comb[["mut_id2"]]&gt==allele_specific_genotype[2]) %>%
        pull(lklhd)
      
      p_diff <- prod(p_m1_gt1*p_m2_gt2)+prod(p_m1_gt2*p_m2_gt1)
      p_same <- prod(p_m1_gt1*p_m2_gt1)+prod(p_m1_gt2*p_m2_gt2)
      
      LLR_d_s <- log10(p_diff/p_same)
      #LLR_s_d <- log10(p_same/p_diff)
      
      return(c(comb, p_same=p_same, p_diff=p_diff, llr=LLR_d_s#, llr_s_d=LLR_s_d
               ))
      
    }) %>%
    bind_rows() %>%
    mutate_at(.vars=c("p_same", "p_diff", "llr"),
              .funs = as.numeric) 
  func_end(ZP_env)
  return(aggregated_likelihoods)
}
perform_copy_number_phasing <- function(df_gene, AllelicImbalancePhasing, bamDna, purity, sex, ZP_env){
  func_start(ZP_env)
  if(AllelicImbalancePhasing){
  ## works only if both variants are in the same segment
    per_segment <- lapply(unique(df_gene$seg_id), function(SEG_ID){
      
      variants_in_segment <- df_gene %>%
        filter(seg_id==SEG_ID)
      if(nrow(variants_in_segment)>=2){
        if(unique(variants_in_segment$all_imb)==TRUE){
          if(!is.na(unique(variants_in_segment$gt_cna))){
            
            allele_specific_genotype <- unique(variants_in_segment$gt_cna) %>%
              str_split(":") %>% unlist() %>% as.numeric()
            
            
            gt_lk_per_mut <- calc_genotype_likelihood_per_mut(variants_in_segment, 
                                                              allele_specific_genotype, 
                                                              bamDna, purity, sex, ZP_env)
            
            if(nrow(gt_lk_per_mut)>0){
              ## can be zero if bam file does not cover variant  
              mut_template <- make_mut_template(variants_in_segment, ZP_env)
            
              aggregated_likelihoods <- aggregate_likelihoods(mut_template, 
                                                              gt_lk_per_mut, allele_specific_genotype, ZP_env)
              
              copy_number_phasing <- aggregated_likelihoods %>%
                rowwise() %>%
                mutate(
                  comb=get_single_index(as.numeric(unlist(str_replace(mut_id2, "m", ""))), 
                                        as.numeric(unlist(str_replace(mut_id1, "m", ""))), 
                                        nrow(ZP_env$global_ZygosityPredictor_variable_mat_phased)),
                  phasing=paste0("som-s", SEG_ID),
                  nconst=case_when(
                    as.numeric(p_same)>=as.numeric(p_diff) ~ 1,
                    as.numeric(p_diff)>as.numeric(p_same) ~ 2
                  ),
                  #conf=max(as.numeric(p_diff),as.numeric(p_same))
                  #conf=min(abs(llr_d_s), abs(llr_s_d))
                  conf=abs(llr)
                )   
            
            ## define inputs for genotype likelyhood
            ## m = ploidy, so probably tcn
            ## g = number of reference alleles
            ## epsylonJ = sequencing error: 1-base call accuracy looped via J
            ## k = number of reads
            ## l = reads supp reference
            ## k-l = reads supporting variant
            }else {
              ## bam file doesnot cover variants
              copy_number_phasing <- NULL
            }
          } else {
            copy_number_phasing <- NULL
          }
        } else {
          ## no alleleic imbalance in segment
          copy_number_phasing <- NULL
        }
      } else {
        ## only one variant in this segment
        copy_number_phasing <- NULL
      }
      return(copy_number_phasing)
    }) %>%
      bind_rows()
    append_matrices(per_segment, ZP_env)
  ## output needs this:
  ##  status nconst p_same p_diff subclonal unplausible  dist class_comb  comb phasing
  } else {
    per_segment <- tibble()
  }
  func_end(ZP_env)
  return(per_segment)
}
#' @keywords internal
#' description follows
#' @importFrom magrittr %>%
#' @importFrom stringr str_split str_detect
#' @importFrom dplyr case_when bind_rows mutate rowwise ungroup
#' @importFrom tibble as_tibble
phase <- function(df_gene,
                  somCna,  
                  bamDna, 
                  purity, 
                  sex, 
                  bamRna=NULL, 
                  haploBlocks=NULL, 
                  vcf=NULL,
                  distCutOff=5000, 
                  printLog=FALSE, 
                  geneDir=NULL, 
                  showReadDetail=FALSE, 
                  snpQualityCutOff=1, 
                  phasingMode="full",
                  AllelicImbalancePhasing=FALSE, 
                  #verbose=FALSE,
                  ZP_env){
  #vm(as.character(sys.call()[1]),  1)
  func_start(ZP_env)
  ## (1): define all combinations of variants to be phased
  create_phasing_matrices(df_gene$mut_id, df_gene$pos, distCutOff, ZP_env)
  unphased <- unphased_main(ZP_env)
  append_loglist(length(unphased), "main combinations to phase,", 
                 abs(length(unphased)-length(ZP_env$global_ZygosityPredictor_variable_mat_phased[upper.tri(ZP_env$global_ZygosityPredictor_variable_mat_phased)])), 
                 "are over distCutOff", ZP_env=ZP_env)
  ## (2): perform direct phasing between variants (read-based)
  direct_phasing <- perform_direct_phasing(df_gene, bamDna, bamRna, printLog, 
                                           showReadDetail, geneDir, ZP_env)
  read_level_phasing_info <- perform_indirect_phasing(df_gene, vcf, bamDna, bamRna, 
                                           haploBlocks,  distCutOff, somCna, 
                                           snpQualityCutOff, purity, sex, 
                                           geneDir, direct_phasing, 
                                           showReadDetail, ZP_env)
  copy_number_phasing <- perform_copy_number_phasing(df_gene, AllelicImbalancePhasing, bamDna, purity, sex, ZP_env)
  ## indirect phasing done
  append_loglist("finalizing phasing results", ZP_env=ZP_env)
  ## reconstruct phasing results
  ZP_env$global_ZygosityPredictor_variable_main_pos <- get_main_mut_pos(ZP_env)
  uppertri <- which(upper.tri(ZP_env$global_ZygosityPredictor_variable_mat_phased))
  all_combs <- intersect(ZP_env$global_ZygosityPredictor_variable_main_pos, uppertri)
 #print(1)
 #print(read_level_phasing_info)
  phasing_all_combs <- aggregate_phasing(all_combs, df_gene, read_level_phasing_info, copy_number_phasing, ZP_env)
 if(nrow(read_level_phasing_info)==0){
   read_level_phasing_info_export <- tibble()
 } else {
    read_level_phasing_info_export <- read_level_phasing_info %>%
      rowwise() %>%
      mutate(ncomb=comb,
             comb=paste(
               sort(
                 factor(
                  c(colnames(ZP_env$global_ZygosityPredictor_variable_mat_phased)[get_xy_index(ncomb, 
                                                      nrow(ZP_env$global_ZygosityPredictor_variable_mat_phased))["x"]], 
                     rownames(ZP_env$global_ZygosityPredictor_variable_mat_phased)[get_xy_index(ncomb, 
                                                       nrow(ZP_env$global_ZygosityPredictor_variable_mat_phased))["y"]]),
                  levels=c(paste0("m", seq(1,1000,1)),
                           paste0("s", seq(1,10000,1)))
                 )
               ),
               collapse="-"
               ))   
  }
  store_log(geneDir, as_tibble(ZP_env$global_ZygosityPredictor_variable_mat_phased), "mat_phased.tsv")
  store_log(geneDir, as_tibble(ZP_env$global_ZygosityPredictor_variable_mat_info), "mat_info.tsv")
 #print(phasing_all_combs)
  func_end(ZP_env)
  return(list(phasing_all_combs, 
              read_level_phasing_info_export, 
              ZP_env$global_ZygosityPredictor_variable_mat_phased, 
              ZP_env$global_ZygosityPredictor_variable_mat_info,
              copy_number_phasing))
}
#' @keywords internal
calc_left_wt_copies <- function(mtcn, nconst, aff_copies1, aff_copies2){
  left_wt_copies <- ifelse(nconst==2,
                           mtcn-sum(as.numeric(aff_copies1), 
                                    as.numeric(aff_copies2)),
                           mtcn-max(as.numeric(aff_copies1), 
                                    as.numeric(aff_copies2)))
  return(as.numeric(left_wt_copies))
}
#' @keywords internal
#' @importFrom GenomicRanges GRanges
#' @importFrom stringr str_detect str_replace_all
#' @importFrom dplyr mutate
#' @importFrom tibble as_tibble
make_both_annotations <- function(region_to_load_in){
  sec <- region_to_load_in %>%
    as_tibble()
  if(str_detect(sec$seqnames[1],"chr")){
    prim <- sec %>%
      mutate(seqnames=str_replace_all(seqnames, "chr", "")) %>%
      GRanges()
  } else {
    prim <- sec %>%
      mutate(seqnames=paste0("chr",seqnames)) %>%
      GRanges()
  }
  fin <- c(region_to_load_in, prim)
  return(fin)
}
#' @importFrom GenomicRanges GRanges
#' @importFrom DelayedArray rowRanges
#' @importFrom IRanges subsetByOverlaps
#' @importFrom stringr str_detect str_replace str_replace_all
#' @importFrom dplyr mutate pull nth 
#' @importFrom tibble as_tibble
#' @importFrom purrr compact
#' @importFrom VariantAnnotation readVcf geno info
#' @importFrom Rsamtools seqnamesTabix TabixFile
loadVcf <- function(vcf_in, chrom, region_to_load_in, ZP_env,  which="all",
                    colname_gt="GT", colname_af="AF", colname_dp4="DP4", 
                    dkfz=FALSE){
  func_start(ZP_env)
  ## first check which format input vcf has
  region_to_load <- region_to_load_in
  chr_anno <- as_tibble(region_to_load_in) %>%
    pull(seqnames) %>% .[1] %>% str_detect("chr")
  sec_chrom <- ifelse(chr_anno, str_replace(chrom, "chr", ""), 
                      paste0("chr", chrom))
  gr_list <- lapply(vcf_in, function(VCF){
    if(is(VCF, "TabixFile")){
      if(chrom %in% Rsamtools::seqnamesTabix(VCF)){
      #if(chrom %in% seqnamesTabix(VCF)){
        loadedVcf <- 
          VariantAnnotation::readVcf(VCF, 
                  param=region_to_load)
        combVcf <- rowRanges(loadedVcf) 
        if(length(combVcf)>0){
          combVcf$ALT <- unlist(lapply(combVcf$ALT, 
                                       function(x){as.character(x[[1]][1])}))
          gt <- VariantAnnotation::geno(loadedVcf)[[colname_gt]][,1] %>%
            as.character()
          af <- VariantAnnotation::info(loadedVcf)[[colname_af]]
          dp4 <- VariantAnnotation::info(loadedVcf)[[colname_dp4]]
          # gt <- geno(loadedVcf)[[colname_gt]][,1] %>% as.character()
          # af <- info(loadedVcf)[[colname_af]]
          # dp4 <- info(loadedVcf)[[colname_dp4]] 
          combVcf$gt <- gt
          combVcf$dp4 <- dp4
          combVcf$af <- af
          vcf_info <- "variants_detected"
        } else {
          combVcf <- NULL
          vcf_info <- "no snps in roi"
        }
      } else {
        combVcf <- NULL
        vcf_info <- "vcf does not cover chromosome"
      }
    } else {
      if(str_detect(VCF, paste0("chr", 
                                unlist(str_replace(chrom, "chr", "")), 
                                "[^0-9]"))){
        #message("loading")
        loadedVcf <- VariantAnnotation::readVcf(VCF)
        gt <- VariantAnnotation::geno(loadedVcf)[[colname_gt]] %>%
          as.character()
        af <- VariantAnnotation::info(loadedVcf)[[colname_af]]
        dp4 <- VariantAnnotation::info(loadedVcf)[[colname_dp4]]
        loadedVcf <- VariantAnnotation::readVcf(VCF)
        # gt <- geno(loadedVcf)[[colname_gt]] %>% as.character()
        # af <- info(loadedVcf)[[colname_af]]
        # dp4 <- info(loadedVcf)[[colname_dp4]]
        rangesVcf <- rowRanges(loadedVcf) 
        rangesVcf$gt <- gt
        rangesVcf$af <- af
        rangesVcf$dp4 <- dp4
        combVcf <-  subsetByOverlaps(rangesVcf, region_to_load) 
        if(length(combVcf)>0){
          combVcf$ALT <- unlist(lapply(combVcf$ALT, 
                                       function(x){as.character(x[[1]][1])}))
          combVcf <- as_tibble(combVcf) %>%
            mutate(seqnames=case_when(
              chr_anno==TRUE&!str_detect(seqnames, "chr") ~ paste0("chr",seqnames),
              chr_anno==FALSE&str_detect(seqnames, "chr") ~ str_replace_all(seqnames, "chr", ""),
              TRUE ~ seqnames
              
            )) %>%
            GRanges()
          vcf_info <- "variants_detected"
        } else {
          combVcf <- NULL
          vcf_info <- "no snps in roi"
        }
      } else {
        combVcf <- NULL
        vcf_info <- "no vcf file covering roi"
      }
    }
    return(list(combVcf, vcf_info))
  }) 
  final_vcf <- lapply(gr_list, nth, 1) %>%
    compact() %>%
    Reduce(function(x,y)c(x,y),.)
  if(which=="HZ"){
    filtered_vcf <- final_vcf[which(final_vcf$gt %in% c("1/0", "0/1", "1|0", "0|1"))]
  } else {
    filtered_vcf <- final_vcf
  }
  func_end(ZP_env)
  return(filtered_vcf)
}
#' @importFrom dplyr case_when
get_string_const <- function(nconst){
  case_when(nconst==2 ~ "diff",
            nconst==1 ~ "same",
            TRUE ~ "null"
  )
}
#' @importFrom IRanges mergeByOverlaps
#' @importFrom GenomicRanges GRanges
#' @importFrom dplyr filter select mutate left_join
#' @importFrom tibble as_tibble
load_snps <- function(df_gene, vcf, haploBlocks, distCutOff,  somCna, 
                      snpQualityCutOff, ZP_env, which="all"){
  func_start(ZP_env)
  chrom <- unique(df_gene$chr)
  region_to_load <- paste0(chrom, ":", min(df_gene$pos)-distCutOff,
                           "-", max(df_gene$pos)+distCutOff) %>%
    GRanges()  %>%
    split_genomic_range(.,df_gene$pos)
  loaded_vcf_hz <- loadVcf(vcf, chrom, region_to_load, ZP_env)
  if(length(loaded_vcf_hz)>0){
    fsnps <- loaded_vcf_hz %>% as_tibble() %>%
      filter(!is.na(QUAL)&QUAL>snpQualityCutOff) %>%
      filter(gt %in% c("1|0", "0|1", "1/0", "0/1"))
    if(nrow(fsnps)>0){
      snps <- fsnps %>%
        mutate(mut_id=paste0("s", c(1:nrow(.))+nrow(df_gene)))
      append_loglist(nrow(snps), "SNPs detected in distCutOff range", ZP_env=ZP_env)
      if(!"af" %in% names(snps)){
        if("dp4" %in% names(snps)){
          snps$af <- lapply(seq(1,nrow(snps)), function(i){
            sum(snps$dp4[[i]][c(3,4)])/sum(snps$dp4[[i]])
          }) %>% as.numeric()  
        } 
      }
      gr_snps <- GRanges(snps %>% select(1:3, mut_id))
      if(!is.null(haploBlocks)){
        merged_haploblocks <- mergeByOverlaps(gr_snps, haploBlocks) %>%
          as_tibble() %>%
          select(all_of(names(.) %>% .[which(!str_detect(.,"\\."))]))    
      } else {
        merged_haploblocks <- snps %>% select(mut_id) %>%
          mutate(hap_id=NA)
      }
      merged_somCna <- mergeByOverlaps(gr_snps, somCna) %>%
        as_tibble() %>%
        select(all_of(names(.) %>% .[which(!str_detect(.,"\\."))]))
      annotated_snps <- left_join(
        snps,
        merged_haploblocks,
        by="mut_id"
      ) %>%
        left_join(
          .,
          merged_somCna,
          by="mut_id"
        )
    } else {
      ## no high quality and heterozygous snps
      annotated_snps <- NULL
    }
  } else {
    ## no snps detected
    annotated_snps <- NULL
  }
  func_end(ZP_env)
  return(annotated_snps)
}
aggregate_probs <- function(ps){
  if(length(ps)>1){
    np <- ps[1]
    for (n in seq(1, length(ps)-1, 1)){
      np <- np+(1-np)*ps[n+1]
    }
    return(np)    
  } else {
    return(ps) 
  }
}
#' @importFrom stringr str_sub
#' @importFrom stats na.omit
#' @importFrom dplyr pull all_of
extract_subseq <- function(ref_pos, element_mut, parsed_read, 
                           length_indel, string){
  pos_in_seq <- ref_pos-element_mut$map_start+1
  subseq_in_element <- str_sub(element_mut[[string]], start=pos_in_seq, 
                               end=-1)
  subseq_in_following_elements <- 
    parsed_read[
      which(as.numeric(parsed_read$id)>as.numeric(element_mut$id)),] %>%
    pull(all_of(string)) %>%
    na.omit() %>%
    paste(collapse="")
  full_remaining <- paste0(subseq_in_element, subseq_in_following_elements)
  return(str_sub(full_remaining, start=1, end=length_indel))
}
#' @importFrom stringr str_sub
#' @importFrom tibble tibble
extract_snv <- function(ref_pos, element_mut){
  ## get cigar element in which reference position in located
  pos_in_seq <- ref_pos-element_mut$map_start+1
  base <- 
    str_sub(element_mut$seq,
            start=pos_in_seq,
            end=pos_in_seq) 
  qual <- 
    str_sub(element_mut$qual,
            start=pos_in_seq,
            end=pos_in_seq)  
  return(tibble(base=base, qual=qual, info="mapped", len_indel=NA,
                exp_indel=NA))
}
#' @importFrom stringr str_sub
#' @importFrom tibble tibble
extract_insertion <- function(ref_pos, parsed_read, element_mut, length_ins){
  ## insertion are always indicated by the I type. The reference position tells
  ## the position in front of the I segment, which means we are looking for
  ## the subsequent one type
  expected_indel_detected <- FALSE
  len <- NA
  if(as.numeric(element_mut$id)==nrow(parsed_read)){
    if(ref_pos==element_mut$map_end){
      info <- 
        "no insertion detected: ref pos is last mapped base of read" 
    } else {
      info <- 
        "no insertion detected: ref pos mapped in last element of parsed read"      
    }
    base <- extract_subseq(ref_pos, element_mut, parsed_read, 1, "seq")
    #qual <- extract_subseq(ref_pos, element_mut, parsed_read, 1, "qual")
  } else if("I" %in% parsed_read$type){
    ## insertion detected... check if it is the matching one
    if(ref_pos==element_mut$map_end){
      ## if insertion is at refrence position, the subsequent position is an a
      ## new elemnt labelled as I
      next_element <- parsed_read[as.numeric(element_mut$id)+1,]
      if(next_element$type=="I"){
        ## insertion detected at position.. extract inserted sequence
        base <- paste0(
          str_sub(element_mut$seq, start=-1, end=-1),
          next_element$seq
        )
        #  qual <- paste0(
        #   str_sub(element_mut$qual, start=-1, end=-1),
        #  next_element$qual
        #)
        info <- "insertion detected"
        len <- next_element$width
        expected_indel_detected <- TRUE
      } else {
        info <- paste(next_element$type, "detected - wrong class")
        base <- extract_subseq(ref_pos, element_mut, parsed_read, length_ins, 
                               "seq")
        #qual <- extract_subseq(ref_pos, element_mut, parsed_read, length_ins, 
        #                      "qual")
      }
    } else {
      info <- "no insertion detected: ref pos not mapped to end of element"
      base <- extract_subseq(ref_pos, element_mut, parsed_read, 1, "seq")
      # qual <- extract_subseq(ref_pos, element_mut, parsed_read, 1, "qual")
    }
  } else {
    ## no inserttion detected.. extract base at position to get base quality
    info <- "no insertion detected: no I in cigar"
    base <- extract_subseq(ref_pos, element_mut, parsed_read, 1, "seq")
    # qual <- extract_subseq(ref_pos, element_mut, parsed_read, 1, "qual")
  }
  ## qual is set to null, as we only take the mapping quality for indels
  return(tibble(base=base, qual=0, info=info, len_indel=len, 
                exp_indel=expected_indel_detected))
}
#' @importFrom tibble tibble
extract_deletion <- function(ref_pos, parsed_read, element_mut, length_del){
  expected_indel_detected <- FALSE
  len <- NA
  if(as.numeric(element_mut$id)==nrow(parsed_read)){
    if(ref_pos==element_mut$map_end){
      info <- 
        "no deletion detected: ref pos is last mapped base of read" 
    } else {
      info <- 
        "no delection detected: ref pos mapped in last element of parsed read"      
    }
    base <- extract_subseq(ref_pos, element_mut, parsed_read, 1, "seq")
    # qual <- extract_subseq(ref_pos, element_mut, parsed_read, 1, "qual")
  } else if("D" %in% parsed_read$type){
    ## insertion detected... check if it is the matching one
    if(ref_pos==element_mut$map_end){
      ## if insertion is at refrence position, the subsequent position is an a
      ## new elemnt labelled as D
      next_element <- parsed_read[as.numeric(element_mut$id)+1,]
      if(next_element$type=="D"){
        expected_indel_detected <- TRUE
        info <- "deletion detected"
        len <- next_element$width
      } else {
        info <- "no deletion detected: subsequent element not I"
      }
    } else {
      info <- "no insertion detected: ref pos not mapped to end of element"
    }      
    base <- extract_subseq(ref_pos, element_mut, parsed_read, length_del, 
                           "seq")
    #qual <- extract_subseq(ref_pos, element_mut, parsed_read, length_del, 
    #                      "qual")
  } else {
    ## no inserttion detected.. extract base at position to get base quality
    info <- "no insertion detected: no I in cigar"
    base <- extract_subseq(ref_pos, element_mut, parsed_read, length_del, 
                           "seq")
    #qual <- extract_subseq(ref_pos, element_mut, parsed_read, length_del, 
    #                      "qual")
  }
  ## qual is set to null, as we only take the mapping quality for indels
  return(tibble(base=base, qual=0, info=info, len_indel=len,
                exp_indel=expected_indel_detected))
}
#' @importFrom dplyr between
cigar_element <- function(parsed_read, ref_pos){
  #func_start(ZP_env)
  cigel <- parsed_read[which(between(rep(ref_pos, nrow(parsed_read)), 
                                     parsed_read$map_start, 
                                     parsed_read$map_end)),] %>%
    .[1,]
  #func_end(ZP_env)
  return(cigel)
}
#' @importFrom tibble tibble
extract_base_at_refpos <- function(parsed_read, ref_pos, class, ref_alt, 
                                   ref_ref){
  #func_start(ZP_env)
  ## extract elemnt containing reference position from parsed read
  element_mut <- cigar_element(parsed_read, ref_pos)
  ## if position is inside N or D type, the read does not cover the position
  ## N means for example skipped exon in RNA
  ## D wouldmean the position is deleted
  ## real deletions are still mapped in the M part
  if(element_mut$type %in% c("N", "D")){
    base_info <- tibble(base=NA, qual=NA, info="position skipped/deleted", 
                        len_indel=NA, exp_indel=NA)
  } else if(class %in% c("snv", "snp")){
    base_info <- extract_snv(ref_pos, element_mut)
  } else if(class=="ins"){
    base_info <- extract_insertion(ref_pos, parsed_read, element_mut, 
                                   nchar(ref_alt)-nchar(ref_ref))
  } else if(class=="del"){
    base_info <- extract_deletion(ref_pos, parsed_read, element_mut, 
                                  1+abs(nchar(ref_ref)-nchar(ref_alt)))
  } else {
    stop("class of variant needs to be provided")
  }
  #func_end(ZP_env)
  return(base_info %>% mutate(class=class, mapq=element_mut$mapq))
}
#' @importFrom stringr str_sub
#' @importFrom dplyr bind_rows mutate na_if
#' @importFrom GenomicAlignments start end width cigarRangesAlongQuerySpace cigarRangesAlongReferenceSpace
parse_cigar <- function(bam, qname, paired){
  #func_start(ZP_env)
  paired_reads <- bam[which(bam$qname==qname)] 
  read_start <- GenomicAlignments::start(paired_reads)
  #read_start <- start(paired_reads)
  seq <- as.character(paired_reads$seq)
  cigar <- paired_reads$cigar
  qual <- as.character(paired_reads$qual)
  if(paired){
    mate <- c(1,2)
  } else {
    mate <- 1
  }
  
  ## parse cigar string according to query
  ## from here for devel
  cigq <- GenomicAlignments::cigarRangesAlongQuerySpace(cigar, with.ops = TRUE) 
  cigr <- GenomicAlignments::cigarRangesAlongReferenceSpace(cigar, with.ops = FALSE) 
  ## from here for bioconductor
  #cigq <- cigarRangesAlongQuerySpace(cigar, with.ops = T) 
  #cigr <- cigarRangesAlongReferenceSpace(cigar, with.ops = F) 
  ## util here for bioconductor
  
  
  raw_cigs <- lapply(mate, function(i){
    cigq_start <- GenomicAlignments::start(cigq[[i]])
    #cigq_start <- start(cigq[[i]])
    cigq_end <- GenomicAlignments::end(cigq[[i]])
    #cigq_end <- end(cigq[[i]])
    cigq_len <- length(cigq_start)
    cigr_start <- GenomicAlignments::start(cigr[[i]])
    #cigr_start <- start(cigr[[i]])
    cigr_end <- GenomicAlignments::end(cigr[[i]])
    #cigr_end <- end(cigr[[i]])
    seqq <- str_sub(seq[i],
                    start=cigq_start,
                    end=cigq_end) %>% na_if("")
    qualq <- str_sub(qual[i],
                     start=cigq_start,
                     end=cigq_end) %>% na_if("")
    width <- pmax(GenomicAlignments::width(cigq[[i]]),
               GenomicAlignments::width(cigr[[i]]))
    # width=pmax(width(cigq[[i]]), 
    #            width(cigr[[i]]))
    map_start <- read_start[i]+cigr_end-width
    map_end <- read_start[i]+cigr_end-1
    raw_cigs_new <- data.frame(
      width=width,
      type=names(cigq[[i]]),
      seq=seqq,
      qual=qualq,
      map_start=map_start,
      map_end=map_end,
      start=cigr_start,
      end=cigr_end,
      mate=mate[i],
      mapq=paired_reads$mapq[i],
      origin=paired_reads$origin[i]
    )
  }) %>%
    bind_rows() %>%
    mutate(id=c(1:nrow(.)))
  #func_end(ZP_env)
  return(raw_cigs)
}
#' @importFrom dplyr case_when
evaluate_base <- function(base_info, ref_alt, ref_ref){
  if(base_info$class %in% c("snv", "snp")){
    detected <- case_when(
      base_info$base==ref_alt ~ 1,
      base_info$base==ref_ref ~ 0,
      TRUE ~ -2
    )
  } else if(base_info$class=="ins"){
    detected <- case_when(
      base_info$base==ref_alt ~ 1,
      base_info$base==ref_ref ~ 0,
      TRUE ~ -2
    )
  } else {
    detected <- case_when(
      base_info$len_indel==abs(nchar(ref_alt)-nchar(ref_ref)) ~ 1,
      is.na(base_info$len_indel) ~ 0,
      TRUE ~ -2
    )
  }
  return(detected)
}
#' @keywords internal
#' description follows
#' @importFrom dplyr case_when 
core_tool <- function(qname, bam,
                      ref_pos1, ref_pos2,
                      ref_alt1, ref_alt2,
                      ref_ref1, ref_ref2,
                      ref_class1, ref_class2, version="old"){
  #vm("core_tool",  1)
  . <- NULL
  ## parse read according to cigar string
  parsed_read <- parse_cigar(bam, qname, paired=TRUE)
  ## extract base at reference position
  base_info1 <- extract_base_at_refpos(parsed_read, ref_pos1, ref_class1, 
                                       ref_alt1, ref_ref1)
  base_info2 <- extract_base_at_refpos(parsed_read, ref_pos2, ref_class2, 
                                       ref_alt2, ref_ref2)
  ## assign final status to read
  if(is.na(base_info1$base)|is.na(base_info2$base)){
    final_assignment <- "skipped"
  } else {
    mut1_in_read <- evaluate_base(base_info1, ref_alt1, ref_ref1)
    mut2_in_read <- evaluate_base(base_info2, ref_alt2, ref_ref2)
    final_assignment <- case_when(
      sum(mut1_in_read, mut2_in_read)==2 ~ "both",
      sum(mut1_in_read, mut2_in_read)==0 ~ "none",
      mut1_in_read==1 ~ "mut1",
      mut2_in_read==1 ~ "mut2",
      TRUE ~ "dev_var"
    )    
  }
  return(c(qname=qname, result=final_assignment,
           origin=unique(parsed_read$origin),
           baseq1=base_info1$qual,
           mapq1=base_info1$mapq,
           baseq2=base_info2$qual,
           mapq2=base_info2$mapq
  )
  )
}

---
title: "Usage of ZygosityPredictor"
author: "Marco Rheinnecker"
output:
  BiocStyle::html_document:
    toc: true
    toc_depth: 2
vignette: >
  %\VignetteIndexEntry{Usage}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# Introduction

The software package ZygosityPredictor allows to predict how many copies of a 
gene are affected by mutations, in particular small variants (single nucleotide 
variants, SNVs, and small insertions and deletions, Indels). In addition to the 
basic calculations of the affected copy number of a variant, ZygosityPredictor 
can phase multiple variants in a gene and ultimately make a prediction if and 
how many wild-type copies of the gene are left. This information proves to be 
of particular use in the context of translational medicine. For example, in 
cancer genomes, ZygosityPredictor can address whether unmutated copies of 
tumor-suppressor genes are present. ZygosityPredictor was developed to handle 
somatic and germline small variants. In addition to the small variant context, 
it can assess larger deletions, which may cause losses of exons or whole genes.

# Important

To use ZygosityPredictor, the input data has to meet some requirements. The 
most important one is that all input variants, SNVs, Indels and sCNAs must 
originate from a fully clonal tumor. If indications are present that a sample
consists of several sub clones (e.g. in tumor samples), the results must be seen 
with caution as the implementation is founded on the assumption of clonality.
The second important requirement applies mainly to the second part of the tool
and will thereofre be explained in section "Predict Zygosity"

# Installation

The following code can be used to install ZygosityPredictor. The installation
needs to be done once. 

```{r, eval=FALSE}
if (!require("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install("ZygosityPredictor")
```


# load example data

To demonstrate the use of ZygosityPredictor, NGS data from the Seq-C2 project 
were used [1]. In the following chunk, all required datalayer of the WGS_LL_T_1 
sample are loaded. The variants are loaded as GRanges objects, one for somatic 
copy number alterations (GR_SCNA), one for germline- and one for somatic small 
variants (GR_GERM_SMALL_VARS and GR_SOM_SMALL_VARS). The input formats will be 
discussed in more detail in section 4. 

```{r}
library(ZygosityPredictor)
library(dplyr)
library(stringr)
library(GenomicRanges)

# file to sequence alignment 
FILE_BAM <- system.file("extdata", "ZP_example.bam", 
                        package = "ZygosityPredictor")
VCF <- system.file("extdata", "ZP_example_chr7.vcf.gz", 
                        package = "ZygosityPredictor")
# meta information of the sample
PURITY <- 0.98
PLOIDY <- 1.57
SEX <- "female"
# variants
data("GR_SCNA")
data("GR_GERM_SMALL_VARS")
data("GR_SOM_SMALL_VARS")
# used gene model
data("GR_GENE_MODEL")
data("GR_HAPLOBLOCKS")

```

# Calculation of affected copies of a variant

Two functions are provided to calculate how many copies are affected by single 
small variants, based on two formulas, one for germline variants and one for 
somatic variants.

## Germline variants

To calculate the affected copies for a germline variant by using 
```aff_germ_copies()```, the following inputs are required:

* **af**: numeric; between 0 and 1; calculated allele frequency of the variant 
in the tumor sample
* **tcn**: numeric; total copy number at the position of the variant
* **purity**: numeric; between 0 and 1; purity or tumor cell content of the 
tumor sample
* **c_normal**: numeric; expected copy number at the position of the variant 
in normal tissue, 1 for gonosomes in male samples, and 2 for male autosomes and 
all chromosomes in female samples. (The function can also assess the c_normal 
parameter by itself, but then the following two inputs must be provided: 
chr and sex)
* **chr**: (only if c_normal is not provided) character; can be either a single 
number or in the “chr1” format; chromosome of the variant
* **sex**: (only if c_normal is not provided) character; either “male” or 
“female” / “m” or “f”; sex of the sample
* **af_normal**: (default 0.5) numeric; allele-frequency of germline variant in 
normal tissue. 0.5 represents heterozygous variants in diploid genome, 1 would
be homozygous. Could be relevant if germline CNVs are present at the position.
Then also the c_normal parameter would have to be adjusted.


the output is a numeric value that indicates the affected copies.

```{r}
## as an example we take the first variant of our prepared input data and 
## extract the required information from different input data layer
## the allele frequency and the chromosome can be taken from the GRanges object

AF = elementMetadata(GR_GERM_SMALL_VARS[1])[["af"]]
CHR = seqnames(GR_GERM_SMALL_VARS[1]) %>%
  as.character()

## the total copy number (tcn) can be extracted from the CNV object by selecting
## the CNV from the position of the variant

TCN = elementMetadata(
  subsetByOverlaps(GR_SCNA, GR_GERM_SMALL_VARS[1])
  )[["tcn"]]

## purity and sex can be taken from the global variables of the sample
## with this function call the affected copies are calculated for the variant
aff_germ_copies(af=AF,
                tcn=TCN,
                purity=PURITY,
                chr=CHR,
                sex=SEX)
```

## Somatic variants

To calculate how many copies are affected by a somatic variant by 
```aff_som_copies()```, the same inputs are required, but a different formula 
is evaluated:
```{r}
## the function for somatic variants works the same way as the germline function

AF = elementMetadata(GR_SOM_SMALL_VARS[1])[["af"]]
CHR = seqnames(GR_SOM_SMALL_VARS[1]) %>%
  as.character()
TCN = elementMetadata(
  subsetByOverlaps(GR_SCNA, GR_SOM_SMALL_VARS[1])
  )[["tcn"]]

aff_som_copies(af=AF,
               chr=CHR,
               tcn=TCN,
               purity=PURITY,
               sex=SEX)
```



## Calculate affected copies of a set of variants

In order to apply the previously mentioned functions to a whole set of variants 
and calculate the affected copies, the following code can be used.
```{r}
## as an example we calculate the affected copies for the somatic variants:
GR_SOM_SMALL_VARS %>%
  ## cnv information for every variant is required.. merge with cnv object
  IRanges::mergeByOverlaps(GR_SCNA) %>% 
  as_tibble() %>%
  ## select relevant columns
  select(chr=1, pos=2, gene, af, tcn) %>%
  mutate_at(.vars=c("tcn", "af"), .funs=as.numeric) %>%
  rowwise() %>%
  mutate(
    aff_copies = aff_som_copies(chr, af, tcn, PURITY, SEX),
    wt_copies = tcn-aff_copies
  )
```

There s also the function ```predict_per_variant``` that basically does the same
with slightly adjusted inputs. 

```{r}
predict_per_variant(purity=PURITY, 
                    sex=SEX,
                    somCna=GR_SCNA,
                    somSmallVars=GR_SOM_SMALL_VARS)

```




# Predict Zygosity

In this section, we will use the WGS_LL_T_1 dataset from the Seq-C2 project as 
an example to investigate whether mutations in the following genes result in 
total absence of wildtype copies. The genes which were selected as an example 
for the analysis are shown below. The example data set was reduced to these 
genes.

* TP53
* BRCA1
* TRANK1
* TRIM3
* JUP
* CDYL
* SCRIB
* ELP2

## Format of input data

Some inputs are optional, while others are compulsory. The latter are labeled 
with “**”. Of note, ZygosityPredictor is applied downstream of variant calling, 
therefore the variant calls, including information on identified somatic copy 
number aberrations (sCNAs), have to be provided. The inputs can be divided into 
five classes:


* File paths:
  + **bamDna\*\* **: character; path to indexed alignment (.bam format)
  + **bamRna**: character; path to rna-sequencing data (.bam format). 
  + **vcf**: character, or character vector containing several vcf file paths; 
  path to variant call file (.vcf.gz format) or .vcf.
  for extended SNP phasing if variants on the same gene are too far away from
  each other for direct phasing
* Sample meta information:
  + **purity\*\* **: numeric; between 0 and 1; indicates purity or tumor cell 
  content of the sample
  + **ploidy**: numeric; ground ploidy of the sample
  + **sex\*\* **: character; “male” or “female” / “m” or “f”; sex of the 
  patient the sample was taken from
* Variants
  + **somCna\*\* **: GRanges object; containing all genomic segments (sCNA) 
  with annotated total copy number (default metadata column name *“tcn”*, 
  custom name can be provided by COLNAME_TCN) and information about LOH 
  (default column name *“cna_type”*, custom name can be provided by 
  COLNAME_CNA_TYPE). The cna_type column should contain the string “LOH” if 
  loss-of heterozygosity is present at the segment. If large deletions should 
  be included to the analysis the total copy number has to be decreased 
  accordingly. If the total copy number is smaller than 0.5, the tool will 
  assume a homozygous deletion. An incomplete deletion is assumed if at least 
  one copy is lost compared to the ploidy of the sample (works only if the
  ploidy is provided as an input)
  + **somSmallVars**: GRanges object; containing all somatic small variants. 
  Required metadata columns are: reference base (*“REF”/”ref”*), alternative 
  base (*“ALT”/”alt”*), allele frequency in the tumor sample (raw allele 
  frequency, i.e. as measured in the tumor sample; not the corrected allele 
  frequency in the supposedly pure tumor) (*“AF”/”af”*), gene (*“gene”/”GENE”*, 
  according to the used gene model (GENCODE39 in the example data) and the 
  annotation provided below). If no relevant somatic small variants are 
  present, can also be NULL or not provided.
  + **germSmallVars**: GRanges object; Analogous to GR_SOM_SMALL_VARS. If no 
  relevant germline small variants are present, can also be NULL or not 
  provided.
* Used Gene model
  + **geneModel\*\* **: GRanges object; containing the gene model for the used 
  reference genome. Required metadata columns are: *"gene"*. Artificially 
  restricting the gene model can be used to tell the tool which genes to 
  analyze. In the case of this vignette, the object contains only the genes we 
  selected.
* Haploblocks
  + **haploBlocks**: GRanges object; containing predefined haplock region, i.e.
  genomic regions in which SNPs could be phased to one specific haplotype. By 
  providing this haploblock together with the SNPs and its genotype annotation
  (1|0 or 0|1) via input **vcf**, haploblock phasing will be enabled.
* Options
  + **logDir**: character; If provided, detailed output will be stored in it
  + **includeIncompleteDel**: logical, default=TRUE; Should incomplete 
  deletions (monoallelic deletions in a diploid sample) be included in the 
  evaluation? Since these often span large parts of the chromosomes and will 
  lead to many affected genes, it can be advisable to include or exclude them, 
  depending on the research question.
  + **includeHomoDel**: logical, default=TRUE; Should homozygous deletions be 
  included in the evaluation?
  + **AllelicImbalancePhasing**: logical, default=FALSE; If TRUE, performs (if
  read-level phasing fails) allelic imbalance phasing. Results must be seen 
  with caution as for somatic variants incorrect constellations might be 
  determined.
  + **showReadDetail**: logical, default=FALSE; If this option is TRUE, another 
  table is added to the output that contains more detailed information about 
  the classification of the read pairs. More detailed information is provided 
  in section 4.3.4.
  + **assumeSomCnaGaps**: logical, default=FALSE; Only required if the somCna 
  object lacks copy number information for genomic segments on which small 
  variants are detected. By default, variants in such regions will be excluded 
  from the analysis as required information about the copy number is missing. 
  These variants will be attached to the final output list in a separate 
  tibble. To include them, this flag must be set TRUE and the ground ploidy 
  must be given as an input. This ground ploidy will then be taken as *tcn* 
  in the missing regions.
  + **byTcn**: logical, default=TRUE; optional if includeHomoDel or 
  includeIncompleteDel is TRUE. If FALSE the tool will not use tcn as a 
  criterion to assign large deletions. It will use the cna_type column and 
  check for indicating strings like HOMDEL/HomoDel/DEL. Some commonly used 
  strings are covered. It is recommended to leave this flag TRUE.
  + **printLog**:  logical, default=TRUE; If TRUE, the tool will print detailed
  information how the assessment is done for each gene.
  + **verbose**: logical, default=FALSE; prints debugging information
  + **colnameTcn**: character; indicating the name of the metadata column 
  containing the tcn information in the somCna object. If not provided 
  the tool tries to detect the column according to default names.
  + **colnameCnaType**: character; The same as for colnameTcn, but for 
  cna-type information.
  + **distCutOff**: numeric, default=5000; if input vcf is provided and SNP
  phasing is performed, this will limt the distance at which the SNP phasing
  should not be tried anymore. As the probability of finding overlapping reads
  at such a long distance is very low and the runtime will increase
  exponentially.
  + **snpQualityCutOff**
  + **refGen**: character, default="hg38"; Required if vcf files is provided. 
  Either "hg38" or "hg19". Relevant for VariantAnnotation vcf loading


## Predict zygosity for a set of genes in a sample

The prediction of zygosity for a set of genes in a sample can be assessed by 
the ```predict_zygosity()``` function.

**Important note**: The runtime of the analysis depends strongly on the number 
of genes to be assessed and on the number of input variants. It is therefore 
recommended to reduce the number of genes to the necessary ones. Also, 
depending on the research question to be addressed, the variants should be 
filtered to the most relevant ones, not only because of runtime considerations, 
but also to sharpen the final result. A large number of mutations in a gene, 
some of which are of little biological relevance or even SNPs, will inevitably 
reduce the validity of the results.

```{r, results = FALSE}
fp <- predict_zygosity(
  purity = PURITY, 
  ploidy = PLOIDY,
  sex = SEX,
  somCna = GR_SCNA, 
  somSmallVars = GR_SOM_SMALL_VARS, 
  germSmallVars = GR_GERM_SMALL_VARS, 
  geneModel = GR_GENE_MODEL,
  bamDna = FILE_BAM,
  vcf=VCF,
  haploBlocks = GR_HAPLOBLOCKS
)
```

## Interpretation of results

Of note, the results displayed here were chosen to explain and exemplify the 
functionality of the tool; biological and medical impact of the specific 
variants has not been a selection criterion.
The result which is returned by the function consists of a list of tibbles: 

* Evaluation per variant
* Evaluation per gene
* Main Phasing info
* detailed read-level phasing (RLP) info
* detailed alleleic imbalance phasing (AIP) info, if enabled
* Read pair info (only if showReadDetail=TRUE and logDir is provided)
* Variants not covered by somCna (only if present and no sCNA gap assumption 
was done)

One way of accessing the results is a simple extraction of the tibbles from the
list. In addition, two accessor function are implemented: ```ZP_ov()``` shows an 
overview of the full resultand and ```gene_ov()``` shows more detailed 
information about a selected gene.

```{r}
ZP_ov(fp)
```

The function provides an overview about the evaluations done by 
ZygosityPredictor. One can see that in our case 12 input small variants were 
used to predict gene status. Of those, two got the status all_copies_affected.


### Evaluation per variant

The first result of the function is the evaluation per variant. In this step 
all information required for subsequent steps is annotated and the affected 
copies per variant are calculated. For every variant, the function checks 
whether it already affects all copies of the gene. The format of the output is 
a tibble; the number of rows corresponds to the total 
number of input variants. The tool annotates a few self-explanatory columns 
such as the origin of the respective variant (germline/somatic) or the class 
(snv/ins/del). It also appends information from the sCNA results: the total 
copy number at the position of the variant and the information if a loss of 
heterozygosity is present (cna_type). Also, an ID is assigned to every small 
variant. Then, the genes are numbered consecutively in order to unambiguously 
assign variants to genes in the following analysis. The most important results 
of this step are the calculation of the affected and wildtype copies, as well
as, depending on the data, an initial check of whether a variant already affects 
all copies.


Of note, there can be situations in which left wildtype copies are below 0.5,  
but still this information is not sufficient to predict *“all_copies_affected”* 
without doubt. Depending on the origin of the variant, further criteria must 
be met (e.g., LOH). The procedure for this first check is shown in the pre_info 
column.


### Evaluation per gene + phasing info

By using  ```gene_ov()```, more detailed information can be viewed about how the
tool came to a gene status. What the accessor function does is basically 
filtering the output tibbles to the gene of relevance. Detailed explanation about all columns can be found below the next chunk.

```{r}
gene_ov(fp, OR2J1)
```

The first prompted tibble originates from the eval_per_gene tibble of the output. 
The tibble shown 
below originates from eval_per_variant and shows all input small variants of the
gene. The next tibble originates from main_phasing_info and shows the main
phasing combinations. Finally, the last tibble comes from detailed_RLP_info
and contains every phasing combination including the ones with and between SNPs.

What can be seen is that the final gene status of OR2J1 is wt_copies_left. The
tool predicted around 2.6 remaining wt copies with maximum confidence level of 
1. Below it can be seen that 3 small variants inside the gene were in the input
leading to 3 main phasing combinations. All of them were phased via direct read level phasing.
As all main combinations could 
be solved, the left wt copies could be accurately defined and the gene status 
assigned. In this particular case, the gene status could have been solved even 
without phasing all main combinations by the case we refer to as insufficient.
As the final gene status is defnied via the so called ```integrated_affected_copies```, 
which are defnied for diff constellation via:  ```aff_cp(m1)+aff_cp(m2) ``` and for the same constellation:
 ```max(aff_cp(m1), aff_cp(m2))```, we can pre-calculate the maximum affected
 copies in case of a diff constellation. If the difference of the total copy number 
 and the integrated affected copies leaves more than 0.5 wt copies, the status wt_cp_left can be concluded.
As can be seen in the main phasing combinations in column: min_poss_wt_cp, none 
of the main phasing combinations could lead to less than 0.5 wt copies even if 
the variants were found on different reads.

The outputs contain columns providing more detailed information about gene status definition
which will be explained in the following. 


#### All read-level phasing combinations 
(either visible via ```gene_ov()``` or accessable via ```fp$detailed_RLP_info```):

+ **both**: number of reads classified as both (both variants present), adjusted with basecall and maping quality
+ **mut1**: number of reads classified as mut1 (only the first variant of combination pßresent), adjusted with basecall and maping quality
+ **mut2**: number of reads classified as mut2 (only the second variant of combination pßresent), adjusted with basecall and maping quality
+ **dev_var**: number of reads having another variant at one of the positions of the expected variants
+ **skipped**: number of reads not mapping to the position of the variant (only expected with RNA reads)
+ **nstatus**: variant constellation in numeric representation (2 = all copies affected, 1 = wt copies left, 0 = undefined)
+ **status**: variant constellation in character representation
+ **nstatus**: numeric representation of constellation
+ **xsq_same/xsq_diff**: X-squared values of chi-squared tests. _same is the one against expected same result
+ **p_same/p_diff**: p-value of chi-squared tests
+ **v_same/v_diff**: Cramers V value of chi-squared tests (5)
+ **none**: number of reads classified as none (none of both variants found)
+ **DNA_rds/RNA_rds**: number of DNA/RNA reads used
+ **dist**: Distance between variants
+ **class_comb**: mutational classes corresponding to character combination identifier
+ **comb**: combination identifier in character representation (mX-mY). 
+ **ncomb**: combination identifier in numeric representation. Derived from position in phasing matrix
+ **subclonal**: TRUE if reads with classification both and either mut1 or mut2 were found. There are two explanantions why soimething like this could happen in a
cancer sample: First that the tumor is not fully clonal and a subclone is present carrying only the variant that happened earlier during tumor development. The second variant is only carried by a subclone that emerged later. The second explanation could be that the reads are actually from the same cells but a duplication of one allel happend between their occurence. We could imagine a scenario were the genotype of a gene is 1:2 and both variants are on the duplicated allele. If one of them happend before the duplication, we would expect reads carrying only the earlier variant but also reads carrying both of them.
+ **unplausible** TRUE if reads with classification both, mut1 and mut2 are found for the same variant combination. Compared to the "subclonal" case this is more difficult to explain and might also be an artifact. The only way this could happen in a cancer sample is that the same mutational event happens independently on both alleles and or in different subclones. This case is generally very unlikely and is therefore annotated as unplasuible. Such a setting is also reflected in the p-valkue and thereofre the confidence, as similary with both expected results is found.


#### Main phasing combinations (either visible via ```gene_ov()``` or accessable via ```fp$phasing_info```):

+ **phasing**: Info which phasing approach was used (direct, indirect, haploblock, imbalance, flagged)
+ **via**: combinations which were used to solve this combination. For direct phasing the number is the numeric representation of the combi9nation identifier.
            For indirect phasing it will look like this: 4-7-8 which measn thta the combinations 4, 7 and 8 were used to solve.
+ **conf**: aggregated confidence for the combination
+ **wt_cp**: exact left wt copies. If phasing of combination was not succesful, wt copies can not be calculated acurately which leads to NA
+ **min_poss_wt_cp**: Minimal possible wt copies if variants are on different copies according to affected copynumber of the two variants in the combination. If greater than 0.5, the constellation is unable to contributre to a status of all copies affrected for the gene.
+ **max_poss_wt_cp**: maximum possible wt copies if the two variants are on the same copy
+ **score**: contribution to gene status. (2 is a contribution of all copies affected by the combination, 1 has more than 0.5 wt copies left)
+ **gene**: Gene that w2as tried to solve with the combination





# References

1. Fang LT, Zhu B, Zhao Y, Chen W, Yang Z, Kerrigan L, Langenbach K, de Mars M, 
Lu C, Idler K, et al. Establishing community reference samples, data and call 
sets for benchmarking cancer mutation detection using whole-genome sequencing. 
Nature Biotechnology. 2021;39(9):1151-1160 / PMID:34504347
2. Lawrence M, Huber W, Pages H, Aboyoun P, Carlson M, et al. (2013) Software 
for Computing and Annotating Genomic Ranges. PLoS Comput Biol 9(8): e1003118. 
doi:10.1371/journal.pcbi.1003118*
3. Wickham H, François R, Henry L, Müller K (2022). _dplyr: A Grammar of Data 
Manipulation_. R package version 1.0.10, 
<https://CRAN.R-project.org/package=dplyr>.
4. Wickham H (2022). _stringr: Simple, Consistent Wrappers for Common String 
Operations_. 
https://stringr.tidyverse.org, https://github.com/tidyverse/stringr.
5. Cramér, Harald. 1946. Mathematical Methods of Statistics. Princeton: 
Princeton University Press, page 282 (Chapter 21. The two-dimensional case)

```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{GR_GENE_MODEL}
\alias{GR_GENE_MODEL}
\title{germline small variant object}
\format{
## `GR_GENE_MODEL`
GRanges object
}
\usage{
data(GR_GENE_MODEL)
}
\value{
Object containing gene model of hg38
}
\description{
germline small variant object
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{GR_GERM_SMALL_VARS}
\alias{GR_GERM_SMALL_VARS}
\title{germline small variant object}
\format{
## `GR_SOM_SMALL_VARS`
GRanges object
}
\usage{
data(GR_SOM_SMALL_VARS)
}
\value{
Object containing germline Indels and SNVs of SeqC2 example case
}
\description{
germline small variant object
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{GR_HAPLOBLOCKS}
\alias{GR_HAPLOBLOCKS}
\title{haploblocks}
\format{
## `GR_HAPLOBLOCKS`
GRanges object
}
\usage{
data(GR_HAPLOBLOCKS)
}
\value{
Object containing haploblock annotations
}
\description{
haploblocks
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{GR_SCNA}
\alias{GR_SCNA}
\title{copynumber object}
\format{
## `GR_SCNA`
GRanges object
}
\usage{
data(GR_SCNA)
}
\value{
Object containing somatic copy number abberations (sCNAs) of SeqC2 example case
}
\description{
copynumber object
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exported.R
\name{predict_zygosity}
\alias{predict_zygosity}
\title{predicts zygosity of a set of genes of a sample}
\usage{
predict_zygosity(
  purity,
  sex,
  somCna,
  geneModel,
  bamDna,
  somSmallVars = NULL,
  germSmallVars = NULL,
  bamRna = NULL,
  ploidy = NULL,
  colnameTcn = NULL,
  colnameCnaType = NULL,
  includeHomoDel = TRUE,
  includeIncompleteDel = TRUE,
  showReadDetail = FALSE,
  printLog = FALSE,
  assumeSomCnaGaps = FALSE,
  byTcn = TRUE,
  vcf = NULL,
  haploBlocks = NULL,
  distCutOff = 5000,
  verbose = FALSE,
  debug = FALSE,
  logDir = NULL,
  snpQualityCutOff = 1,
  phasingMode = "fast",
  AllelicImbalancePhasing = FALSE
)
}
\arguments{
\item{purity}{purity of the sample (numeric value between 0 and 1 indicating 
the fraction of relevant sample with control/unrelevant tissue)}

\item{sex}{sex of the sample (character: "male", "female", "m", "f")}

\item{somCna}{GRanges object containing all genomic regions with annotated 
total copynumber and cna_type as metadata columns. The total-copynumber 
column should be named "tcn" but also some other commonly used names. 
It should contain numeric values or characters that can be converted to 
numeric values. The cna_type column must contain the information about 
loss of heterozygosity (LOH). Therefore the term "LOH" must be explicitely 
mentioned in the column. If a genomic region is not present in the object, 
it will be taken as heterozygous with neutral TCN of 2.}

\item{geneModel}{GRanges object containing the gene-annoattion of 
the used reference genome with metadata column of the gene name (gene)}

\item{bamDna}{path to bam-file}

\item{somSmallVars}{GRanges object containing all somatic small 
variants (SNV and INDEL).
Required metadata columns are reference base (ref/REF), 
alternative base (alt/ALT),
annotation of the gene name (gene/GENE) and the allele-frequency (af/AF). 
If the object is not provided the tool assumes there are no somatic small 
variants.}

\item{germSmallVars}{GRanges object containing all germline small 
variants (SNV and INDEL).
Required metadata columns are reference base (ref/REF), alternative 
base (alt/ALT),
annotation of the gene name (gene/GENE) and the allele-frequency (af/AF)
If the object is not provided the tool assumes there are no germline small 
variants.}

\item{bamRna}{optional; path to rna file (bam format)}

\item{ploidy}{ploidy of the sample (numeric value)}

\item{colnameTcn}{character indicating the name of the metadata containing 
the tcn information in the somCna object. If not provided the tool tries to 
detect the column according to default names}

\item{colnameCnaType}{character indicating the name of the metadata 
containing cna type information in the somCna object. 
If not provided the tool tries to detect the column according to default 
names}

\item{includeHomoDel}{default = TRUE; if FALSE homozygous deleteions are 
excluded}

\item{includeIncompleteDel}{default = TRUE; if FALSE heterzygous deleteions 
are excluded}

\item{showReadDetail}{default = FALSE; if TRUE a table is added to the 
output, containing all used reads/rea-pairs with anntated read classification 
(mut1, mut2, both, none, skipped, dev_var)}

\item{printLog}{default = FALSE; if TRUE the gene which is evaluated is 
printed in console, 
containing the query-name of each read which was used to perform 
haplotype-phasing and the info into which class it was assigned.}

\item{assumeSomCnaGaps}{(logical, default=FALSE) Only required if the somCna
object lacks copy number information for genomic segments on which small 
variants are detected. By default, variants in such regions will be excluded 
from the analysis as required information about the copy number is missing. 
These variants will be attached to the final output list in a separate 
tibble. To include them, this flag must be set TRUE and the ground ploidy 
must be given as an input. This ground ploidy will then be taken as tcn in 
the missing regions. If no ploidy is given the tool will assume the ground 
ploidy of 2 when this flag is TRUE.}

\item{byTcn}{logical, default=TRUE; optional if includeHomoDel or 
includeIncompleteDelS is TRUE. If FALSE the tool will not use tcn as a 
criterion to assign large deletions. It will use the cna_type column and 
check for indicating strings like HOMDEL/HomoDel/DEL. Some commonly used 
strings are covered. It is recommended to leave this flag TRUE}

\item{vcf}{character; path to variant call file (.vcf.gz format). 
Will be used (if provided)
for extended SNP phasing if variants on the same gene are too far away from
each other for direct haplotype phasing}

\item{haploBlocks}{GRanges object containing haploblocks. Haploblocks are
defined as genomic regions in which SNPs are phased to a specific allele.
For example a haploblock could be chr1:1000-10000. This would mean that every
genotype annotation in the format "1|0" or "0|1" of a SNP in this region will 
be used to phase somatic variants and define their genotype}

\item{distCutOff}{numeric, default=5000; if input vcf is provided and SNP
phasing is performed, this will limt the distance at which the SNP phasing
should not be tried anymore. As the probability of finding overlapping reads
at such a long distance is very low and the runtime will increase
exponentially.}

\item{verbose}{logical, default=FALSE; prints functions that are called}

\item{debug}{logical, default=FALSE; prints output for debugging}

\item{logDir}{character; path to directory where logfiles and detailed infos 
of the run can be stored, if not given, no details will be stored or printed}

\item{snpQualityCutOff}{numeric, default=1; Cutoff to filter for SNPS that 
can be used for phasing}

\item{phasingMode}{character, default="fast"; if set to full. Even if high 
confidence phasing result could be achieved, following phasing approaches 
will be carried out}

\item{AllelicImbalancePhasing}{logical, default=FALSE. Enables alleleic imbalance phasing if TRUE}
}
\value{
A list of dataframes. Those are the evaluation per variant, 
the evaluation per gene and, if performed, the info about the 
haplotype-phasing.
}
\description{
predicts zygosity of a set of genes of a sample
}
\examples{
cnvs  = GenomicRanges::GRanges(
  dplyr::tibble(
    chr = "chr17",
    start = c(170060, 34520990),
    end = c(34520990, 83198614),
    tcn = c(2, 1),
    cna_type = c("neutral", "LOH")
  )
)
somatic_vars = GenomicRanges::GRanges(
  dplyr::tibble(
    chr="chr17",
    start = 7675088,
    end = 7675088,
    ref = "C",
    alt = "T",
    af = 0.65,
    gene = "TP53" 
  )
)
germline_vars = GenomicRanges::GRanges(
  dplyr::tibble(
    chr="chr17",
    start = 41771694,
    end = 41771694,
    ref = "GTGT",
    alt = "G",
    af = 0.95,
    gene = "JUP" 
  )
)
reference = GenomicRanges::GRanges(
  dplyr::tibble(
    chr = "chr17",
    start = c(7661778, 41754603),
    end = c(7687538, 41786931),
    gene = c("TP53", "JUP")
  )
)
sex = "female"
purity = 0.9
bamfile <- system.file("extdata", "ZP_example.bam", 
  package = "ZygosityPredictor")
predict_zygosity(purity = purity, sex = sex, 
  somCna = cnvs,
  somSmallVars = somatic_vars,
  germSmallVars = germline_vars,
  geneModel = reference,
  bamDna = bamfile
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exported.R
\name{predict_per_variant}
\alias{predict_per_variant}
\title{predicts zygosity of a set of variants}
\usage{
predict_per_variant(
  purity,
  sex,
  somCna,
  geneModel = NULL,
  somSmallVars = NULL,
  germSmallVars = NULL,
  ploidy = NULL,
  colnameTcn = NULL,
  colnameCnaType = NULL,
  includeHomoDel = TRUE,
  includeIncompleteDel = TRUE,
  assumeSomCnaGaps = FALSE,
  byTcn = TRUE,
  ZP_env = NULL,
  verbose = FALSE
)
}
\arguments{
\item{purity}{purity of the sample (numeric value between 0 and 1 indicating 
the fraction of relevant sample with control/unrelevant tissue)}

\item{sex}{sex of the sample (character: "male", "female", "m", "f")}

\item{somCna}{GRanges object containing all genomic regions with annotated 
total copynumber and cna_type as metadata columns. The total-copynumber 
column should be named "tcn" but also some other commonly used names. 
It should contain numeric values or characters that can be converted to 
numeric values. The cna_type column must contain the information about 
loss of heterozygosity (LOH). Therefore the term "LOH" must be explicitely 
mentioned in the column. If a genomic region is not present in the object, 
it will be taken as heterozygous with neutral TCN of 2.}

\item{geneModel}{GRanges object containing the gene-annoattion of 
the used reference genome with metadata column of the gene name (gene)}

\item{somSmallVars}{GRanges object containing all somatic small 
variants (SNV and INDEL).
Required metadata columns are reference base (ref/REF), 
alternative base (alt/ALT),
annotation of the gene name (gene/GENE) and the allele-frequency (af/AF). 
If the object is not provided the tool assumes there are no somatic small 
variants.}

\item{germSmallVars}{GRanges object containing all germline small 
variants (SNV and INDEL).
Required metadata columns are reference base (ref/REF), alternative 
base (alt/ALT),
annotation of the gene name (gene/GENE) and the allele-frequency (af/AF)
If the object is not provided the tool assumes there are no germline small 
variants.}

\item{ploidy}{ploidy of the sample (numeric value)}

\item{colnameTcn}{character indicating the name of the metadata containing 
the tcn information in the somCna object. If not provided the tool tries to 
detect the column according to default names}

\item{colnameCnaType}{character indicating the name of the metadata 
containing cna type information in the somCna object. 
If not provided the tool tries to detect the column according to default 
names}

\item{includeHomoDel}{default = TRUE; if FALSE homozygous deleteions are 
excluded}

\item{includeIncompleteDel}{default = TRUE; if FALSE heterzygous deleteions 
are excluded}

\item{assumeSomCnaGaps}{(logical, default=FALSE) Only required if the somCna
object lacks copy number information for genomic segments on which small 
variants are detected. By default, variants in such regions will be excluded 
from the analysis as required information about the copy number is missing. 
These variants will be attached to the final output list in a separate 
tibble. To include them, this flag must be set TRUE and the ground ploidy 
must be given as an input. This ground ploidy will then be taken as tcn in 
the missing regions. If no ploidy is given the tool will assume the ground 
ploidy of 2 when this flag is TRUE.}

\item{byTcn}{logical, default=TRUE; optional if includeHomoDel or 
includeIncompleteDelS is TRUE. If FALSE the tool will not use tcn as a 
criterion to assign large deletions. It will use the cna_type column and 
check for indicating strings like HOMDEL/HomoDel/DEL. Some commonly used 
strings are covered. It is recommended to leave this flag TRUE}

\item{ZP_env}{internal variable... not recommened to be changed by user}

\item{verbose}{logical, default=FALSE; prints functions that are called}
}
\value{
A list containing tibbles with all input variants
}
\description{
predicts zygosity of a set of variants
}
\examples{
cnvs  = GenomicRanges::GRanges(
  dplyr::tibble(
    chr = "chr17",
    start = c(170060, 34520990),
    end = c(34520990, 83198614),
    tcn = c(2, 1),
    cna_type = c("neutral", "LOH")
  )
)
somatic_vars = GenomicRanges::GRanges(
  dplyr::tibble(
    chr="chr17",
    start = 7675088,
    end = 7675088,
    ref = "C",
    alt = "T",
    af = 0.65,
    gene = "TP53" 
  )
)
germline_vars = GenomicRanges::GRanges(
  dplyr::tibble(
    chr="chr17",
    start = 41771694,
    end = 41771694,
    ref = "GTGT",
    alt = "G",
    af = 0.95,
    gene = "JUP" 
  )
)
reference = GenomicRanges::GRanges(
  dplyr::tibble(
    chr = "chr17",
    start = c(7661778, 41754603),
    end = c(7687538, 41786931),
    gene = c("TP53", "JUP")
  )
)
sex = "female"
purity = 0.9
predict_per_variant(purity = purity, sex = sex, 
  somCna = cnvs,
  somSmallVars = somatic_vars,
  germSmallVars = germline_vars,
  geneModel = reference
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exported.R
\name{aff_som_copies}
\alias{aff_som_copies}
\title{calculates how many copies are affected by a somatic small variant}
\usage{
aff_som_copies(chr, af, tcn, purity, sex, c_normal = NULL)
}
\arguments{
\item{chr}{chromosome of the variant (either format 1,2,..,X,Y or 
chr1,..,chrX)}

\item{af}{Allele-frequency of the variant (numeric value between 0 and 1)}

\item{tcn}{total-copynumber at position of the variant (numeric value >0)}

\item{purity}{purity of the sample (numeric value between 0 and 1 indicating 
the fraction of relevant sample with control/unrelevant tissue)}

\item{sex}{sex of the sample (character: "male", "female", "m", "f")}

\item{c_normal}{expected copy number at the position of the variant in 
normal tissue, 1 for gonosomes in male samples, and 2 for male autosomes and 
all chromosomes in female samples. (The function can also assess the 
c_normal parameter by itself, but then the following two inputs must be 
provided: chr and sex)}
}
\value{
A numeric value indicating the affecting copies for the variant
}
\description{
calculates how many copies are affected by a somatic small variant
}
\examples{
library(dplyr)
library(purrr)
library(stringr)
aff_som_copies(chr="chrX", af=0.67, tcn=2, purity=0.9, sex="female")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exported.R
\name{gene_ov}
\alias{gene_ov}
\title{accesor for gene predictions printing detailed info about how a gene status 
was assigned}
\usage{
gene_ov(fp, inp_gene, n = 20)
}
\arguments{
\item{fp}{full prediction (output of predict_zygoisty())}

\item{inp_gene}{name of gene that should be printed with detailed information}

\item{n}{max number of rows to print, as some gene status depend on loads of
phasing results#'}
}
\value{
prints overview about run from function predict_zygoisty() with specific information about provided gene
}
\description{
accesor for gene predictions printing detailed info about how a gene status 
was assigned
}
\examples{
cnvs  = GenomicRanges::GRanges(
  dplyr::tibble(
    chr = "chr17",
    start = c(170060, 34520990),
    end = c(34520990, 83198614),
    tcn = c(2, 1),
    cna_type = c("neutral", "LOH")
  )
)
somatic_vars = GenomicRanges::GRanges(
  dplyr::tibble(
    chr="chr17",
    start = 7675088,
    end = 7675088,
    ref = "C",
    alt = "T",
    af = 0.65,
    gene = "TP53" 
  )
)
germline_vars = GenomicRanges::GRanges(
  dplyr::tibble(
    chr="chr17",
    start = 41771694,
    end = 41771694,
    ref = "GTGT",
    alt = "G",
    af = 0.95,
    gene = "JUP" 
  )
)
reference = GenomicRanges::GRanges(
  dplyr::tibble(
    chr = "chr17",
    start = c(7661778, 41754603),
    end = c(7687538, 41786931),
    gene = c("TP53", "JUP")
  )
)
sex = "female"
purity = 0.9
bamfile <- system.file("extdata", "ZP_example.bam", 
  package = "ZygosityPredictor")
fp <- predict_zygosity(purity = purity, sex = sex, 
  somCna = cnvs,
  somSmallVars = somatic_vars,
  germSmallVars = germline_vars,
  geneModel = reference,
  bamDna = bamfile
)
gene_ov(fp, TP53)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{GR_SOM_SMALL_VARS}
\alias{GR_SOM_SMALL_VARS}
\title{somatic small variant object}
\format{
## `GR_GERM_SMALL_VARS`
GRanges object
}
\usage{
data(GR_GERM_SMALL_VARS)
}
\value{
Object containing somatic Indels and SNVs of SeqC2 example case
}
\description{
somatic small variant object
}
\keyword{datasets}
